from django.core.exceptions import ValidationError
from django.db import models
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class AbbreviationQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Abbreviation.get_all_abbreviations()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        """
        Return a new QuerySet instance with the args ANDed to the existing
        set.
        """
        if len(args) > 0:
            search_string = args[0].children[0].children[0][1]
            if search_string:
                abbreviations = []
                self._fetch_all()
                if self._result_cache is not None:
                    for abbreviation in self._result_cache:
                        if (search_string in abbreviation.abbreviation) or (search_string in abbreviation.explanation):
                            abbreviations.append(abbreviation)

                queryset = AbbreviationQuerySet(Abbreviation)
                queryset._result_cache = abbreviations
                return queryset

        return self


class AbbreviationManager(models.Manager):
    def get_queryset(self):
        queryset = AbbreviationQuerySet(Abbreviation)
        return queryset


class Abbreviation(models.Model):
    all_abbreviations = None
    abbreviation = models.CharField(max_length=256, null=False, blank=False, verbose_name="Förkortning", unique=True)
    explanation = models.CharField(max_length=2000, null=False, blank=False, verbose_name="Förklaring")
    comment = models.CharField(max_length=2000, null=False, blank=True, default="", verbose_name="Kommentar")
    is_currently_used = models.BooleanField(default=True, null=False, verbose_name="Används för närvarande")
    category = models.CharField(max_length=2000, null=False, blank=True, default="", verbose_name="Kategori")
    objects = AbbreviationManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "förkortning"
        verbose_name_plural = "förkortningar"

    def __str__(self):
        if (hasattr(self, 'id')) and (self.id is not None):
            return f'Förkortning: {self.abbreviation} (id={self.id})'
        else:
            return f'Förkortning: {self.abbreviation}'

    # Get all abbreviations from smp service.
    @classmethod
    def get_all_abbreviations(cls):
        if Abbreviation.all_abbreviations is None:
            smp_service = SmpWebService()
            abbreviations_json = smp_service.get_abbreviations()
            abbreviations = [Abbreviation.get_abbreviation_from_json(abbreviation_json) for abbreviation_json in
                             abbreviations_json]
            Abbreviation.all_abbreviations = abbreviations
        return Abbreviation.all_abbreviations

    @classmethod
    def create(cls, abbreviation):
        # Create abbreviation in SmpService.
        smp_service = SmpWebService()
        abbreviation_json = smp_service.create_abbreviation(abbreviation,
                                                            abbreviation.author.token,
                                                            abbreviation.csrftoken)
        Abbreviation.all_abbreviations = None
        return Abbreviation.get_abbreviation_from_json(abbreviation_json)

    def delete(self):
        smp_service = SmpWebService()
        Abbreviation.all_abbreviations = None
        ok = smp_service.delete_abbreviation(self, self.author.token, self.csrftoken)

    @classmethod
    def get_abbreviation_from_json(cls, abbreviation_json):
        abbreviation = Abbreviation()
        abbreviation.abbreviation = abbreviation_json['abbreviation']
        abbreviation.explanation = abbreviation_json['explanation']
        abbreviation.comment = abbreviation_json['comment']
        abbreviation.is_currently_used = abbreviation_json['is_currently_used']
        abbreviation.category = abbreviation_json['category']
        abbreviation.pk = abbreviation_json['id']
        abbreviation.id = abbreviation_json['id']
        return abbreviation

    def save(self, commit=True):
        # Save abbreviation in SmpService.
        smp_service = SmpWebService()
        smp_service.update_abbreviation(self, self.author.token, self.csrftoken)
        Abbreviation.all_abbreviations = None
        return self

    def validate_unique(self, exclude):
        """
        Check that unique fields are unique for this data type.
        """

        # Check that abbreviation is unique.
        abbreviations = Abbreviation.objects.all()
        for abbreviation in abbreviations:
            # Check that abbreviation is unique.
            if self.abbreviation == abbreviation.abbreviation:
                #Check that id is not None.
                if (hasattr(self, 'id')) or (self.id is None):
                    if self.id != abbreviation.id:
                        raise ValidationError(f'Det finns redan en förkortning "{self.abbreviation}".')
                else:
                    raise ValidationError(f'Det finns redan en förkortning "{self.abbreviation}".')
