import random
from datetime import datetime

from django.db import models

from ISOF.model.language_smp import Language
from ISOF.model.lexical_name import LexicalName
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService
# import dates specification
from ISOF.model.date_specification import DateSpecification


class ArchivePostQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if not hasattr(self, '_result_cache') or self._result_cache is None:
            self._result_cache = ArchivePost.get_all_archive_posts()
            self.create_dictionary(self._result_cache)
            return self._result_cache

    def filter(self, *args, **kwargs):
        if len(args) > 0:
            #search for archive post name in source or register
            search_string = args[0].children[0].children[0][1]
            if search_string:
                archive_posts = []
                self._fetch_all()
                if self._result_cache is not None:
                    for archive_post in self._result_cache:
                        if search_string.casefold() in archive_post.name_in_source.casefold() or search_string.casefold() in archive_post.register.casefold():
                            archive_posts.append(archive_post)
                queryset = ArchivePostQuerySet(ArchivePost)
                queryset._result_cache = archive_posts
                queryset.create_dictionary(queryset._result_cache)
                return queryset
        return self


class ArchivePostManager(models.Manager):
    def get_queryset(self):
        queryset = ArchivePostQuerySet(ArchivePost)
        return queryset


class ArchivePost(models.Model):
    all_archive_posts = None

    ARCHIVE_POST_TYPE = (
        ('PlaceArchivePost', 'Ortsbelägg'),
        ('PersonArchivePost', 'Personbelägg'),
        ('CommentaryArchivePost', 'Anmärkningsbelägg')
    )

    change_date = models.DateTimeField(null=False, verbose_name="Ändringsdatum")
    date_fake = models.ForeignKey(DateSpecification, null=True, blank=True, editable=True, on_delete=models.CASCADE,
                                  related_name='archive_post_date_fake', verbose_name="Felaktigt datum")
    date_original = models.ForeignKey(DateSpecification, null=True, blank=True, editable=True, on_delete=models.CASCADE,
                                      related_name='archive_post_date_original', verbose_name="Datum i originalet")
    date_transcription = models.ForeignKey(DateSpecification, null=True, blank=True, editable=True,
                                           on_delete=models.CASCADE,
                                           related_name='archive_post_date_transcription',
                                           verbose_name="Datum för avskrift")
    internal_note = models.CharField(max_length=4000, null=True, blank=True, verbose_name="Intern kommentar")
    is_in_book = models.BooleanField(default=True, null=False, verbose_name="Ingår i boken")
    is_seal = models.BooleanField(null=False, default=False, verbose_name="Sigill")
    name_in_source = models.CharField(null=False, blank=False, max_length=512, verbose_name="Beläggstext")
    register = models.TextField(max_length=4000, null=True, blank=True, verbose_name="Regest")
    type = models.CharField(max_length=25, null=False, blank=False,
                            choices=ARCHIVE_POST_TYPE, verbose_name="Typ av belägg")
    language = models.ForeignKey(Language, null=False, blank=False, editable=True, default=96,
                                on_delete=models.DO_NOTHING, verbose_name="Språk")
    objects = ArchivePostManager()
    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = 'belägg'
        verbose_name_plural = 'belägg'

    def __str__(self):
        return "Belägg: " + self.name_in_source

    @classmethod
    def get_all_archive_posts(cls):
        if ArchivePost.all_archive_posts is None:
            smp_service = SmpWebService()
            archive_posts_json = smp_service.get_archive_posts()
            ArchivePost.all_archive_posts = ArchivePost.get_all_archive_posts_from_json(archive_posts_json)
        return ArchivePost.all_archive_posts

    @classmethod
    def get_all_archive_posts_from_json(cls, archive_posts_json):
        archive_posts = []
        for archive_post_json in archive_posts_json:
            archive_post = ArchivePost.get_archive_post_from_json(archive_post_json)
            archive_posts.append(archive_post)
        return archive_posts

    @classmethod
    def get_archive_post_from_json(cls, archive_post_json):
        archive_post = ArchivePost()
        archive_post.id = archive_post_json['id']
        archive_post.pk = archive_post_json['id']
        archive_post.change_date = datetime.strptime(archive_post_json['change_date'], '%Y-%m-%dT%H:%M:%S.%fZ')
        if archive_post_json['date_fake'] is not None:
            archive_post.date_fake = DateSpecification.get_date_specification_from_json(archive_post_json['date_fake'])
        if archive_post_json['date_original'] is not None:
            archive_post.date_original = DateSpecification.get_date_specification_from_json(
                archive_post_json['date_original'])
        if archive_post_json['date_transcription'] is not None:
            archive_post.date_transcription = DateSpecification.get_date_specification_from_json(
                archive_post_json['date_transcription'])
        archive_post.internal_note = archive_post_json['internal_note']
        archive_post.is_in_book = archive_post_json['is_in_book']
        archive_post.is_seal = archive_post_json['is_seal']
        archive_post.register = archive_post_json['register']
        archive_post.type = archive_post_json['type']
        archive_post.name_in_source = archive_post_json['name_in_source']
        archive_post.language = Language.objects.get(id=archive_post_json['language'])
        return archive_post

    # create
    @classmethod
    def create(cls, archive_post):
        smp_service = SmpWebService()
        archive_post_json = smp_service.create_archive_post(archive_post, archive_post.author.token,
                                                            archive_post.csrftoken)
        archive_post = ArchivePost.get_archive_post_from_json(archive_post_json)
        ArchivePost.all_archive_posts = None
        return archive_post

    # delete
    def delete(self):
        smp_service = SmpWebService()
        archive_post_json = smp_service.delete_archive_post(self, self.author.token, self.csrftoken)
        ArchivePost.all_archive_posts = None
        return archive_post_json

    # save
    def save(self):
        smp_service = SmpWebService()
        archive_post_json = smp_service.update_archive_post(self, self.author.token, self.csrftoken)
        ArchivePost.all_archive_posts = None
        return archive_post_json
