from django.core.exceptions import ValidationError
from django.db import models
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class CaseQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Case.get_all_cases()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        return self


class CaseManager(models.Manager):
    def get_queryset(self):
        queryset = CaseQuerySet(Case)
        return queryset


class Case(models.Model):
    all_cases = None
    abbreviation = models.CharField(max_length=20, null=False, blank=False, verbose_name="Förkortning", unique=True)
    case = models.CharField(max_length=50, null=False, blank=False, verbose_name="Kasus", unique=True)
    objects = CaseManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "kasus"
        verbose_name_plural = "kasus"

    def __str__(self):
        if (hasattr(self, 'id')) and (self.id is not None):
            return f'{self.case} (id={self.id})'
        else:
            return f'{self.case}'

    @classmethod
    def get_all_cases(cls):
        if Case.all_cases is None:
            smp_service = SmpWebService()
            cases_json = smp_service.get_cases()
            cases_list = Case.get_cases_from_json(cases_json)
            Case.all_cases = cases_list
        return Case.all_cases

    @classmethod
    def get_cases_from_json(cls, cases_json):
        cases = []
        for case_json in cases_json:
            case = Case.get_case_from_json(case_json)
            cases.append(case)
        return cases

    @classmethod
    def get_case_from_json(cls, case_json):
        case = Case()
        case.abbreviation = case_json['abbreviation']
        case.case = case_json['case']
        case.id = case_json['id']
        case.pk = case_json['id']
        return case

    @classmethod
    def create(cls, case):
        smp_service = SmpWebService()
        case_json = smp_service.create_case(case, case.author.token, case.csrftoken)
        new_case = Case.get_case_from_json(case_json)
        Case.all_cases = None
        return new_case

    def save(self, commit=True):
        smp_service = SmpWebService()
        case_json = smp_service.update_case(self, self.author.token, self.csrftoken)
        Case.all_cases = None
        return case_json

    def delete(self):
        smp_service = SmpWebService()
        case_json = smp_service.delete_case(self, self.author.token, self.csrftoken)
        Case.all_cases = None
        return case_json

    def validate_unique(self, exclude):
        """
        Check that unique fields are unique for this data type.
        """

        # Check that abbreviation is unique.
        cases = Case.objects.all()
        for case in cases:
            # Check that abbreviation is unique.
            if self.abbreviation == case.abbreviation:
                # Check that id is not None.
                if (hasattr(self, 'id')) or (self.id is None):
                    if self.id != case.id:
                        raise ValidationError(f'Det finns redan ett kasus med förkortningen "{self.abbreviation}".')
                else:
                    raise ValidationError(f'Det finns redan ett kasus med förkortningen "{self.abbreviation}".')

            # Check that case is unique.
            if self.case == case.case:
                # Check that id is not None.
                if (hasattr(self, 'id')) or (self.id is None):
                    if self.id != case.id:
                        raise ValidationError(f'Det finns redan ett kasus "{self.case}".')
                else:
                    raise ValidationError(f'Det finns redan ett kasus "{self.case}".')
