# import Castle
from django.core.exceptions import ValidationError
from django.db import models

from ISOF.model.city import City
from ISOF.model.province_smp import Province
# caste from ISOF models
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class CastleQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Castle.get_all_castles()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        if len(args) > 0:
            args_part = args[0].children[0]
            if (isinstance(args_part, tuple) and isinstance(args_part[0], str) and (args_part[0] == 'province__in')):
                # TODO: Handle the entire list of provinces.
                province = args_part[1][0]
                castles = []
                self._fetch_all()
                if (self._result_cache is not None):
                    for castle in self._result_cache:
                        if (castle.province_id == province.id):
                            castles.append(castle)
                queryset = CastleQuerySet(Castle)
                queryset._result_cache = castles
                queryset.create_dictionary(queryset._result_cache)
                return queryset
            else:
                search_string = args[0].children[0].children[0][1]
                if search_string:
                    castles = []
                    self._fetch_all()
                    if self._result_cache is not None:
                        for castle in self._result_cache:
                            if search_string.casefold() in castle.castle_name.casefold():
                                castles.append(castle)
                    queryset = CastleQuerySet(Castle)
                    queryset._result_cache = castles
                    queryset.create_dictionary(queryset._result_cache)
                    return queryset

        return self


class CastleManager(models.Manager):
    def get_queryset(self):
        queryset = CastleQuerySet(Castle)
        return queryset


class Castle(models.Model):
    castle_name = models.CharField(max_length=512, null=False, blank=False, verbose_name="Slott")
    province = models.ForeignKey(Province, on_delete=models.CASCADE, null=False, blank=False, verbose_name="Landskap")
    abbreviation = models.CharField(max_length=512, null=False, blank=False, verbose_name="Förkortning")
    comment = models.CharField(max_length=512, null=True, blank=True, verbose_name="Kommentar")
    objects = CastleManager()
    all_castles = None

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "slott"
        verbose_name_plural = "slott"

    def __str__(self):
        return self.castle_name

    # get all castles class method
    @classmethod
    def get_all_castles(cls):
        if Castle.all_castles is None:
            smp_service = SmpWebService()
            # get castles from smp service
            castles_json = smp_service.get_castles()
            castles = Castle.get_castles_from_json(castles_json)
            Castle.all_castles = castles
        return Castle.all_castles

    @classmethod
    def get_castles_from_json(cls, castles_json):
        castles = []
        provinces = Province.objects.all()
        for castle_json in castles_json:
            castle = Castle()
            castle.castle_name = castle_json['castle_name']
            castle.province = provinces.get(id=castle_json['province'])
            castle.abbreviation = castle_json['abbreviation']
            castle.comment = castle_json['comment']
            castle.id = castle_json['id']
            castle.pk = castle_json['id']
            castles.append(castle)
        return castles

    @classmethod
    def get_castle_from_json(cls, castle_json):
        castle = Castle()
        provinces = Province.objects.all()
        castle.castle_name = castle_json['castle_name']
        castle.province = provinces.get(id=castle_json['province'])
        castle.abbreviation = castle_json['abbreviation']
        castle.comment = castle_json['comment']
        castle.id = castle_json['id']
        castle.pk = castle_json['id']
        return castle

    @classmethod
    def create(cls, castle):
        smp_service = SmpWebService()
        castle_json = smp_service.create_castle(castle, castle.author.token, castle.csrftoken)
        Castle.all_castles = None
        return Castle.get_castle_from_json(castle_json)

    def save(self, commit=True):
        # save in web service
        smp_service = SmpWebService()
        smp_service.update_castle(self, self.author.token, self.csrftoken)
        Castle.all_castles = None
        return self

    def delete(self):
        smp_service = SmpWebService()
        Castle.all_castles = None
        return smp_service.delete_castle(self, self.author.token, self.csrftoken)

    def validate_unique(self, exclude):
        castles = Castle.objects.all()
        for castle in castles:
            if self.castle_name == castle.castle_name and self.province == castle.province:
                if (hasattr(self, 'id')) or (self.id is not None):
                    if self.id != castle.id:
                        raise ValidationError('Slottet ' + self.castle_name + ' finns redan i landskapet '+ self.province.name)
                else:
                    raise ValidationError("Kunde inte spara slott")
