from django.core.exceptions import ValidationError
from django.db import models

from ISOF.model.province_smp import Province
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService
class CityQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = City.get_all_cities()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        if len(args) > 0:
            args_part = args[0].children[0]
            if (isinstance(args_part, tuple) and isinstance(args_part[0], str) and (args_part[0] == 'province__in')):
                # TODO: Handle the entire list of provinces.
                province = args_part[1][0]
                cities = []
                self._fetch_all()
                if (self._result_cache is not None):
                    for city in self._result_cache:
                        if (city.province_id == province.id):
                            cities.append(city)
                queryset = CityQuerySet(City)
                queryset._result_cache = cities
                queryset.create_dictionary(queryset._result_cache)
                return queryset
            elif 'children' in args[0].children[0]:
                search_string = args[0].children[0].children[0][1]
                if search_string:
                    cities = []
                    self._fetch_all()
                    if self._result_cache is not None:
                        for city in self._result_cache:
                            if search_string.casefold() in city.name.casefold():
                                cities.append(city)
                    queryset = CityQuerySet(City)
                    queryset._result_cache = cities
                    queryset.create_dictionary(queryset._result_cache)
                    return queryset
            else:
                raise ValidationError('Kunde inte söka efter stad')

        return self


class CityManager(models.Manager):
    def get_queryset(self):
        queryset = CityQuerySet(City)
        return queryset


class City(models.Model):
    all_cities = None
    abbreviation = models.CharField(max_length=128, null=True, blank=True, verbose_name="Förkortning")
    name = models.CharField(max_length=512, null=False, blank=False, verbose_name="Stad")
    province = models.ForeignKey('Province', on_delete=models.CASCADE, null=False, blank=False, verbose_name="Landskap")
    objects = CityManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "stad"
        verbose_name_plural = "städer"

    def __str__(self):
        return self.name

    @classmethod
    def get_all_cities(cls):
        if City.all_cities is None:
            smp_service = SmpWebService()
            # get cities from smp service
            cities_json = smp_service.get_cities()
            cities = City.get_cities_from_json(cities_json)
            City.all_cities = cities
        return City.all_cities

    @classmethod
    def get_cities_from_json(cls, cities_json):
        cities = []
        provinces = Province.objects.all()
        for city_json in cities_json:
            city = City()
            city.abbreviation = city_json['abbreviation']
            city.name = city_json['name']
            city.province = provinces.get(id=city_json['province'])
            city.id = city_json['id']
            city.pk = city_json['id']
            cities.append(city)
        return cities

    @classmethod
    def get_city_from_json(cls, city_json):
        city = City()
        city.abbreviation = city_json['abbreviation']
        city.name = city_json['name']
        city.province = Province.objects.get(id=city_json['province'])
        city.id = city_json['id']
        city.pk = city_json['id']
        return city

    # create
    @classmethod
    def create(cls, city):
        smp_service = SmpWebService()
        city_json = smp_service.create_city(city, city.author.token, city.csrftoken)
        city = City.get_city_from_json(city_json)
        City.all_cities = None
        return city

    # update in smp service
    def save(self, commit=True):
        smp_service = SmpWebService()
        city_json = smp_service.update_city(self, self.author.token, self.csrftoken)
        city = City.get_city_from_json(city_json)
        City.all_cities = None
        return city

    # delete in smp service
    def delete(self):
        smp_service = SmpWebService()
        ok = smp_service.delete_city(self, self.author.token, self.csrftoken)
        City.all_cities = None
        return ok

    def validate_unique(self, exclude):
        cities = City.objects.all()
        for city in cities:
            if (hasattr(self, 'id')) or (self.id is not None):
                if city.province == self.province and city.name == self.name and city.id != self.id:
                    raise ValidationError('Stad med samma namn finns redan i landskapet')
            else:
                raise ValidationError('Kunde inte spara stad')
