from django.db import models

from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class DateSpecificationQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if not hasattr(self, '_result_cache') or self._result_cache is None:
            self._result_cache = DateSpecification.get_all_date_specifications()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        date_specification_id = self.get_id(*args, **kwargs)
        if (date_specification_id):
            date_specification = self.get_date_specification(date_specification_id)
            date_specifications = []
            date_specifications.append(date_specification)
            queryset = DateSpecificationQuerySet(DateSpecification)
            queryset._result_cache = date_specifications
            return queryset
        else:
            self._fetch_all()
            return self

    def get_date_specification(self, date_specification_id):
        smp_service = SmpWebService()
        date_specification_json = smp_service.get_date_specification(date_specification_id)
        if (date_specification_json):
            return DateSpecification.get_date_specification_from_json(date_specification_json)
        else:
            return None


class DateSpecificationManager(models.Manager):
    def get_queryset(self):
        queryset = DateSpecificationQuerySet(DateSpecification)
        return queryset


class DateSpecification(models.Model):
    all_date_specifications = None
    DATE_OCCURRENCE = (
        ('HasDate', 'Daterad'),
        ('NoDate', 'Odaterad'),
        ('NoDay', 'Utan dag'),
        ('NoYear', 'Utan år'),
        ('UncertainDay', 'Oviss dag'),
        ('UncertainYear', 'Ovisst år')
    )

    DATE_SPECIFICATION_TYPE = (
        ('Standard', 'Standard'),
        ('Before', 'Före'),  # Abbreviation: 'f.'
        ('Beginning', 'Början av'),  # Abbreviation: 'b.'
        ('FirstHalf', 'Förra hälften av'),  # Abbreviation: 'f.h.'
        ('After', 'Efter'),  # Abbreviation: 'e.'
        ('Middle', 'Mitten'),  # Abbreviation: 'm.'
        ('Around', 'Omkring'),  # Abbreviation: 'omkr.'
        ('LatterHalf', 'Senare hälften av'),  # Abbreviation: 's.h.'
        ('EndOf', 'Slutet av'),  # Abbreviation: 'sl.'
        ('EndOfMiddleAges', 'Slutet av medeltiden'),
        ('MiddleAges', 'Medeltiden'),
        ('AfterMiddleAges', 'Efter medeltiden'),
        ('Winter', 'Vintern'),
        ('SpringWinter', 'Vårvintern'),
        ('Lent', 'Fastan'),
        ('Spring', 'Våren'),
        ('Summer', 'Sommaren'),
        ('Autumn', 'Hösten')
    )

    date_occurrence = models.CharField(max_length=64, null=False, blank=False, default='HasDate',
                                       choices=DATE_OCCURRENCE, verbose_name="Förekomst av datum")
    date_sort_order = models.DateTimeField(null=True, verbose_name="Sorteringsordning", blank=True, auto_now=True,
                                           editable=False)
    day_end = models.IntegerField(null=True, blank=True, verbose_name="Dag för slutdatum")
    day_start = models.IntegerField(null=True, blank=True, verbose_name="Dag för startdatum")
    external_note = models.TextField(max_length=4000, null=True, blank=True, verbose_name="Extern kommentar")
    is_period = models.BooleanField(null=False, default=False, verbose_name="Är en tidsperiod")
    month_end = models.IntegerField(null=True, blank=True, verbose_name="Månad för slutdatum")
    month_start = models.IntegerField(null=True, blank=True, verbose_name="Månad för startdatum")
    type_end_date = models.CharField(max_length=64, null=True, blank=True, choices=DATE_SPECIFICATION_TYPE,
                                     verbose_name="Typ av datumangivelse för slut datumet")
    type_start_date = models.CharField(max_length=64, null=False, blank=False, choices=DATE_SPECIFICATION_TYPE,
                                       default='Standard', verbose_name="Typ av datumangivelse för start datumet")
    year_end = models.IntegerField(null=True, blank=True, verbose_name="År för slutdatum")
    year_start = models.IntegerField(null=True, blank=True, verbose_name="År för startdatum")


    objects = DateSpecificationManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "datumangivelse"
        verbose_name_plural = "datumangivelser"

    def __str__(self):
        date_spec_str = ""
        if self.year_start is not None:
            date_spec_str = date_spec_str + str(self.year_start) + "-"
        if self.year_end is not None:
            date_spec_str = date_spec_str + str(self.year_end)

        return date_spec_str + " (id=" + str(self.id) + ")"


    @classmethod
    def get_all_date_specifications(cls):
        if DateSpecification.all_date_specifications is None:
            smp_service = SmpWebService()
            date_specifications_json = smp_service.get_date_specifications()
            DateSpecification.all_date_specifications = DateSpecification.get_all_date_specifications_from_json(
                date_specifications_json)
        return DateSpecification.all_date_specifications

    @classmethod
    def get_all_date_specifications_from_json(cls, date_specifications_json):
        date_specifications = []
        for date_specification_json in date_specifications_json:
            date_specification = DateSpecification.get_date_specification_from_json(date_specification_json)
            date_specifications.append(date_specification)
        return date_specifications

    @classmethod
    def get_date_specification_from_json(cls, date_specification_json):
        date_specification = DateSpecification()
        date_specification.date_occurrence = date_specification_json['date_occurrence']
        date_specification.date_sort_order = date_specification_json['date_sort_order']
        date_specification.day_end = date_specification_json['day_end']
        date_specification.day_start = date_specification_json['day_start']
        date_specification.external_note = date_specification_json['external_note']
        date_specification.is_period = date_specification_json['is_period']
        date_specification.month_end = date_specification_json['month_end']
        date_specification.month_start = date_specification_json['month_start']
        date_specification.type_start_date = date_specification_json['type_start_date']
        date_specification.type_end_date = date_specification_json['type_end_date']
        date_specification.year_end = date_specification_json['year_end']
        date_specification.year_start = date_specification_json['year_start']
        date_specification.id = date_specification_json['id']
        date_specification.pk = date_specification_json['id']
        return date_specification

    @classmethod
    def create(cls, date_specification):
        smp_service = SmpWebService()
        date_specification_json = smp_service.create_date_specification(date_specification,
                                                                        date_specification.author.token,
                                                                        date_specification.csrftoken)
        new_date_specification = DateSpecification.get_date_specification_from_json(date_specification_json)
        DateSpecification.all_date_specifications = None
        return new_date_specification

    def delete(self):
        smp_service = SmpWebService()
        delete_ok = smp_service.delete_date_specification(self, self.author.token, self.csrftoken)
        DateSpecification.all_date_specifications = None
        return delete_ok

    def save(self):
        if self.id is None:
            new_date_specification = DateSpecification.create(self)
            self.id = new_date_specification.id
            self.pk = new_date_specification.id
            return new_date_specification
        else:
            smp_service = SmpWebService()
            date_specification_json = smp_service.update_date_specification(self, self.author.token, self.csrftoken)
            updated_date_specification = DateSpecification.get_date_specification_from_json(date_specification_json)
            DateSpecification.all_date_specifications = None
            return updated_date_specification


    def validate_unique(self, exclude=None):
        pass
