# import django models
from django.core.exceptions import ValidationError
from django.db import models

from ISOF.model.province_smp import Province
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


# inherit from simple model query set
class DistrictQuerySet(SimpleModelQuerySet):
    # fetch all districts from smp service
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = District.get_all_districts()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        if (len(args) > 0):
            search_string = args[0].children[0].children[0][1]
            if (search_string):
                districts = []
                self._fetch_all()
                if (self._result_cache is not None):
                    for district in self._result_cache:
                        if (search_string.casefold() in district.name.casefold()):
                            districts.append(district)
                queryset = DistrictQuerySet(District)
                queryset._result_cache = districts
                queryset.create_dictionary(queryset._result_cache)
                return queryset
        return self

    def get_districts_by_province(self, province):
        districts = []
        self._fetch_all()
        if (self._result_cache is not None):
            for district in self._result_cache:
                if (district.province == province):
                    districts.append(district)
        queryset = DistrictQuerySet(District)
        queryset._result_cache = districts
        return queryset

    def get_districts_by_province_id(self, province_id):
        districts = []
        self._fetch_all()
        if (self._result_cache is not None):
            for district in self._result_cache:
                if (district.province.id == province_id):
                    districts.append(district)
        queryset = DistrictQuerySet(District)
        queryset._result_cache = districts
        return queryset


# district manager
class DistrictManager(models.Manager):
    def get_queryset(self):
        queryset = DistrictQuerySet(District)
        return queryset


class District(models.Model):
    all_districts = None
    name = models.CharField(max_length=512, null=False, blank=False, verbose_name="Härad")
    province = models.ForeignKey(Province, on_delete=models.DO_NOTHING, null=False, default=1, verbose_name="Landskap", limit_choices_to={'country': 'Sverige'})
    harad_id_isof = models.IntegerField(null=False, blank=False, default=0, verbose_name="ISOF-id")
    objects = DistrictManager()


    def __str__(self):
        return self.name

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "härad"
        verbose_name_plural = "härad"
        ordering = ['name']

    # get all districts from smp service
    @classmethod
    def get_all_districts(cls):
        if District.all_districts is None:
            smp_service = SmpWebService()
            districts_json = smp_service.get_districts()
            districts = District.get_districts_from_json(districts_json)
            District.all_districts = districts
        return District.all_districts

    @classmethod
    def get_district_from_json(cls, district_json):
        district = District()
        district.name = district_json['name']
        district.province = Province.objects.get(id=district_json['province'])
        district.harad_id_isof = district_json['harad_id_isof']
        district.id = district_json['id']
        district.pk = district_json['id']
        return district

    @classmethod
    def get_districts_from_json(cls, districts_json):
        districts = []
        provinces = Province.objects.all()
        for district_json in districts_json:
            district = District()
            district.name = district_json['name']
            district.province = provinces.get(id=district_json['province'])
            district.harad_id_isof = district_json['harad_id_isof']
            district.id = district_json['id']
            district.pk = district_json['id']
            districts.append(district)
        return districts

    # class method create
    @classmethod
    def create(cls, district):
        smp_service = SmpWebService()
        district_json = smp_service.create_district(district,
                                                            district.author.token,
                                                            district.csrftoken)
        District.all_districts = None
        return District.get_district_from_json(district_json)

    def save(self, commit=True):
        smp_service = SmpWebService()
        smp_service.update_district(self, self.author.token, self.csrftoken)
        District.all_districts = None
        return self

    def delete(self):
        smp_service = SmpWebService()
        ok = smp_service.delete_district(self, self.author.token, self.csrftoken)
        District.all_districts = None
        return ok

    def validate_unique(self, exclude):
        districts = District.objects.all()
        for district in districts:
            # Check if district name is unique.
            if (self.name == district.name):
                if (hasattr(self, 'id')) or (self.id is not None):
                    if (self.id != district.id):
                        raise ValidationError(f'Det finns redan ett härad "{self.name}".')
                else:
                    raise ValidationError(f'Det finns redan ett härad "{self.name}".')
