# model class Language
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Exists

from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


# language query set
class LanguageQuerySet(SimpleModelQuerySet):
    # fetch all languages from smp service
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Language.get_all_languages()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        is_in_lexical_name = None
        if (len(args) > 0):
            if isinstance(args[0], Exists):
                is_in_lexical_name = True
            else:
                filter_tuple = args[0].children[0]
                if hasattr(filter_tuple, 'children'):
                    search_string = filter_tuple.children[0][1]
                    if (search_string):
                        languages = []
                        self._fetch_all()
                        if (self._result_cache is not None):
                            for language in self._result_cache:
                                if (search_string in language.language):
                                    languages.append(language)
                        queryset = LanguageQuerySet(Language)
                        queryset._result_cache = languages
                        queryset.create_dictionary(queryset._result_cache)
                        return queryset
                else:
                    is_in_lexical_name = filter_tuple[1]

        if is_in_lexical_name:
            languages = []
            self._fetch_all()
            if self._result_cache is not None:
                for language in self._result_cache:
                    if language.is_in_lexical_name == is_in_lexical_name:
                        languages.append(language)
            queryset = LanguageQuerySet(Language)
            queryset._result_cache = languages
            return queryset
        return self


class LanguageManager(models.Manager):
    def get_queryset(self):
        queryset = LanguageQuerySet(Language)
        return queryset

    # meta class
    class Meta:
        managed = True
        verbose_name = "språk"
        verbose_name_plural = "språk"

    def __str__(self):
        return self.language

    @classmethod
    def get_language_from_json(cls, language_json):
        language = Language()
        language.abbreviation = language_json['abbreviation']
        language.language = language_json['language']
        language.sort_order = language_json['sort_order']
        language.is_in_lexical_name = language_json['is_in_lexical_name']
        return language


class Language(models.Model):
    all_languages = None
    abbreviation = models.CharField(max_length=20, null=False, blank=False, verbose_name="Förkortning")
    language = models.CharField(max_length=50, null=False, blank=False, verbose_name="Språk")
    sort_order = models.IntegerField(verbose_name="Sorteringsordning")
    is_in_lexical_name = models.BooleanField(null=False, blank=False, verbose_name="Förekommer i lexikalt namn")
    objects = LanguageManager()

    # create
    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "språk"
        verbose_name_plural = "språk"
        ordering = ['sort_order']

    def __str__(self):
        return self.language

    @classmethod
    def get_all_languages(cls):
        if Language.all_languages is None:
            smp_service = SmpWebService()
            languages_json = smp_service.get_languages()
            Language.all_languages = Language.get_languages_from_json(languages_json)
        return Language.all_languages
    @classmethod
    def get_language_from_json(cls, language_json):
        language = Language()
        language.abbreviation = language_json['abbreviation']
        language.language = language_json['language']
        language.sort_order = language_json['sort_order']
        language.is_in_lexical_name = language_json['is_in_lexical_name']
        language.id = language_json['id']
        language.pk = language_json['id']
        return language

    # get all languages from json
    @classmethod
    def get_languages_from_json(cls, languages_json):
        languages = []
        for language_json in languages_json:
            language = Language()
            language.abbreviation = language_json['abbreviation']
            language.language = language_json['language']
            language.sort_order = language_json['sort_order']
            language.is_in_lexical_name = language_json['is_in_lexical_name']
            language.id = language_json['id']
            language.pk = language_json['id']
            languages.append(language)
        return languages

    @classmethod
    def create(cls, language):
        smp_service = SmpWebService()
        language_json = smp_service.create_language(language, language.author.token, language.csrftoken)
        new_language = Language.get_language_from_json(language_json)
        Language.all_languages = None
        return new_language

    def save(self, commit=True):
        smp_service = SmpWebService()
        language_json = smp_service.update_language(self, self.author.token, self.csrftoken)
        Language.all_languages = None
        return self

    # non class method delete
    def delete(self):
        smp_service = SmpWebService()
        ok = smp_service.delete_language(self, self.author.token, self.csrftoken)
        Language.all_languages = None
        return ok

    # validate unique
    def validate_unique(self, exclude):
        languages = Language.objects.all()
        for language in languages:
            if language.language == self.language:
                # check that id is not the same
                if language.id != self.id:
                    raise ValidationError(
                        ('Språk med detta namn existerar redan.'),
                        params={'language': self.language},
                    )

# language manager class
