from django.db import models

from ISOF.model.block import Block
from ISOF.model.castle import Castle
from ISOF.model.city import City
from ISOF.model.district_smp import District
from ISOF.model.locality_relation import LocalityRelation
from ISOF.model.monastery import Monastery
from ISOF.model.parish_smp import Parish
from ISOF.model.person import Person
from ISOF.model.province_smp import Province
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class LocalityQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Locality.get_all_localities()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        #if kwargs
        if kwargs and 'person' in kwargs:
            person = kwargs['person']
            localities = []
            self._fetch_all()
            if self._result_cache is not None:
                for locality in self._result_cache:
                    if locality.person == person:
                        localities.append(locality)
            queryset = LocalityQuerySet(Locality)
            queryset._result_cache = localities
            queryset.create_dictionary(queryset._result_cache)
            return queryset

        if len(args) > 0:
            search_string = args[0].children[0].children[0][1]
            if search_string:
                localities = []
                self._fetch_all()
                if self._result_cache is not None:
                    for locality in self._result_cache:
                        # use casefold () on both strings
                        if search_string.casefold() in locality.homestead.casefold():
                            localities.append(locality)
                queryset = LocalityQuerySet(Locality)
                queryset._result_cache = localities
                queryset.create_dictionary(queryset._result_cache)
                return queryset
        return self


class LocalityManager(models.Manager):
    def get_queryset(self):
        queryset = LocalityQuerySet(Locality)
        return queryset


class Locality(models.Model):
    all_localities = None

    block = models.ForeignKey(Block, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Kvarter")
    castle = models.ForeignKey(Castle, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Slott/Borg")
    city = models.ForeignKey(City, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Stad")
    district = models.ForeignKey(District, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Härad")
    end_date = models.DateField(null=True, blank=True, default=None, verbose_name="Slutdatum")
    external_note = models.TextField(max_length=4000, null=True, blank=True, verbose_name="Extern kommentar")
    homestead = models.CharField(max_length=512, null=True, blank=True, verbose_name="Gårdsnamn")
    medieval_district = models.CharField(max_length=256, null=True, blank=True, verbose_name="Medeltida härad")
    medieval_parish = models.CharField(max_length=256, null=True, blank=True, verbose_name="Medeltida socken")
    monastery = models.ForeignKey(Monastery, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Kloster")
    parish = models.ForeignKey(Parish, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Socken")
    person = models.ForeignKey(Person, null=False, blank=False, on_delete=models.DO_NOTHING, verbose_name="Person")
    province = models.ForeignKey(Province, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Landskap")
    start_date = models.DateField(null=True, blank=True, default=None, verbose_name="Startdatum")
    type = models.ForeignKey(LocalityRelation, null=True, blank=True, default=5,
                             on_delete=models.DO_NOTHING, verbose_name="Typ av lokalisering")

    objects = LocalityManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "lokalisering"
        verbose_name_plural = "lokaliseringar"

    def __str__(self):
        locality_string = ''
        if (self.person):
            locality_string = locality_string + self.person.name
        if (self.type):
            locality_string = locality_string + ', ' + self.type.locality_relation
        if (hasattr(self, 'id')):
            locality_string = locality_string + " (id=" + str(self.id) + ")"
        return locality_string

    # get all localities class method
    @classmethod
    def get_all_localities(cls):
        # get localities from webservice
        if Locality.all_localities is None:
            web_service = SmpWebService()
            localities_json = web_service.get_localities()
            localities = Locality.get_localities_from_json(localities_json)
            Locality.all_localities = localities
        return Locality.all_localities

    # get localities from json class method
    @classmethod
    def get_localities_from_json(cls, localities_json):
        localities = []
        blocks = Block.objects.all()
        castles = Castle.objects.all()
        cities = City.objects.all()
        districts = District.objects.all()
        locality_relations = LocalityRelation.objects.all()
        monasteries = Monastery.objects.all()
        parishes = Parish.objects.all()
        persons = Person.objects.all()
        provinces = Province.objects.all()

        for locality_json in localities_json:
            locality = Locality()
            locality.id = locality_json['id']
            locality.pk = locality_json['id']
            if locality_json['block'] is not None:
                locality.block = blocks.get(id=locality_json['block'])
            if locality_json['castle'] is not None:
                locality.castle = castles.get(id=locality_json['castle'])
            if locality_json['city'] is not None:
                locality.city = cities.get(id=locality_json['city'])
            if locality_json['district'] is not None:
                locality.district = districts.get(id=locality_json['district'])
            if locality_json['end_date'] is not None:
                locality.end_date = locality_json['end_date']
            if locality_json['external_note'] is not None:
                locality.external_note = locality_json['external_note']
            if locality_json['homestead'] is not None:
                locality.homestead = locality_json['homestead']
            if locality_json['medieval_district'] is not None:
                locality.medieval_district = locality_json['medieval_district']
            if locality_json['medieval_parish'] is not None:
                locality.medieval_parish = locality_json['medieval_parish']
            if locality_json['monastery'] is not None:
                locality.monastery = monasteries.get(id=locality_json['monastery'])
            if locality_json['parish'] is not None:
                locality.parish = parishes.get(id=locality_json['parish'])
            if locality_json['person'] is not None:
                locality.person = persons.get(id=locality_json['person'])
            if locality_json['province'] is not None:
                locality.province = provinces.get(id=locality_json['province'])
            if locality_json['start_date'] is not None:
                locality.start_date = locality_json['start_date']
            if locality_json['type'] is not None:
                locality.type = locality_relations.get(id=locality_json['type'])
            localities.append(locality)
        return localities

    # create locality class method
    @classmethod
    def create(cls, locality):
        web_service = SmpWebService()
        locality_json = web_service.create_locality(locality, locality.author.token, locality.csrftoken)
        locality = Locality.get_locality_from_json(locality_json)
        Locality.all_localities = None
        return locality

    # get locality from json class method
    @classmethod
    def get_locality_from_json(cls, locality_json):
        locality = Locality()
        locality.id = locality_json['id']
        if locality_json['block'] is not None:
            locality.block = Block.objects.get(id=locality_json['block'])
        if locality_json['castle'] is not None:
            locality.castle = Castle.objects.get(id=locality_json['castle'])
        if locality_json['city'] is not None:
            locality.city = City.objects.get(id=locality_json['city'])
        if locality_json['district'] is not None:
            locality.district = District.objects.get(id=locality_json['district'])
        if locality_json['end_date'] is not None:
            locality.end_date = locality_json['end_date']
        if locality_json['external_note'] is not None:
            locality.external_note = locality_json['external_note']
        if locality_json['homestead'] is not None:
            locality.homestead = locality_json['homestead']
        if locality_json['medieval_district'] is not None:
            locality.medieval_district = locality_json['medieval_district']
        if locality_json['medieval_parish'] is not None:
            locality.medieval_parish = locality_json['medieval_parish']
        if locality_json['monastery'] is not None:
            locality.monastery = Monastery.objects.get(id=locality_json['monastery'])
        if locality_json['parish'] is not None:
            locality.parish = Parish.objects.get(id=locality_json['parish'])
        if locality_json['person'] is not None:
            locality.person = Person.objects.get(id=locality_json['person'])
        if locality_json['province'] is not None:
            locality.province = Province.objects.get(id=locality_json['province'])
        if locality_json['start_date'] is not None:
            locality.start_date = locality_json['start_date']
        if locality_json['type'] is not None:
            locality.type = LocalityRelation.objects.get(id=locality_json['type'])
        return locality

    # update locality class method
    def save(self, commit=True):
        web_service = SmpWebService()
        locality_json = web_service.update_locality(self, self.author.token, self.csrftoken)
        locality = Locality.get_locality_from_json(locality_json)
        Locality.all_localities = None
        return locality

    # delete locality class method
    def delete(self):
        web_service = SmpWebService()
        web_service.delete_locality(self, self.author.token, self.csrftoken)
        Locality.all_localities = None
        return self



