from django.core.exceptions import ValidationError
from django.db import models
from ISOF.model.province_smp import Province
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class MonasteryQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Monastery.get_all_monasteries()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        if len(args) > 0:
            args_part = args[0].children[0]
            if (isinstance(args_part, tuple) and isinstance(args_part[0], str) and (args_part[0] == 'province__in')):
                # TODO: Handle the entire list of provinces.
                province = args_part[1][0]
                monasteries = []
                self._fetch_all()
                if (self._result_cache is not None):
                    for monastery in self._result_cache:
                        if (monastery.province_id == province.id):
                            monasteries.append(monastery)
                queryset = MonasteryQuerySet(Monastery)
                queryset._result_cache = monasteries
                queryset.create_dictionary(queryset._result_cache)
                return queryset
            else:
                search_string = args[0].children[0].children[0][1]
                if search_string:
                    monasteries = []
                    self._fetch_all()
                    if self._result_cache is not None:
                        for monastery in self._result_cache:
                            # use casefold () on both strings
                            if search_string.casefold() in monastery.name.casefold():
                                monasteries.append(monastery)
                    queryset = MonasteryQuerySet(Monastery)
                    queryset._result_cache = monasteries
                    queryset.create_dictionary(queryset._result_cache)
                    return queryset
        return self


class monasteryManager(models.Manager):
    def get_queryset(self):
        queryset = MonasteryQuerySet(Monastery)
        return queryset


class Monastery(models.Model):
    all_monasteries = None
    name = models.CharField(max_length=512, null=False, blank=False, verbose_name="Namn på kloster")
    abbreviation = models.CharField(max_length=100, null=False, blank=True, verbose_name="Förkortning")
    province = models.ForeignKey(Province, on_delete=models.CASCADE, null=False, blank=False, verbose_name="Landskap")
    comment = models.CharField(max_length=1000, null=True, blank=True, verbose_name="Kommentar")
    objects = monasteryManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "kloster"
        verbose_name_plural = "kloster"

    def __str__(self):
        return self.name

    @classmethod
    def get_all_monasteries(cls):
        if Monastery.all_monasteries is None:
            smp_service = SmpWebService()
            monasteries_json = smp_service.get_monasteries()
            monasteries = Monastery.get_monasteries_from_json(monasteries_json)
            Monastery.all_monasteries = monasteries
        return Monastery.all_monasteries

    @classmethod
    def get_monasteries_from_json(cls, monasteries_json):
        monasteries = []
        provinces = Province.objects.all()
        for monastery_json in monasteries_json:
            monastery = Monastery()
            monastery.name = monastery_json['name']
            monastery.comment = monastery_json['comment']
            monastery.abbreviation = monastery_json['abbreviation']
            monastery.province = provinces.get(id=monastery_json['province'])
            monastery.id = monastery_json['id']
            monastery.pk = monastery_json['id']
            monastery.abbreviation = monastery_json['abbreviation']
            monasteries.append(monastery)
        return monasteries

    @classmethod
    def get_monastery_from_json(cls, monastery_json):
        monastery = Monastery()
        monastery.name = monastery_json['name']
        monastery.comment = monastery_json['comment']
        monastery.abbreviation = monastery_json['abbreviation']
        monastery.province = Province.objects.get(id=monastery_json['province'])
        monastery.id = monastery_json['id']
        monastery.pk = monastery_json['id']
        monastery.abbreviation = monastery_json['abbreviation']
        return monastery

    #create in smp service
    @classmethod
    def create(cls, monastery):
        smp_service = SmpWebService()
        monastery_json = smp_service.create_monastery(monastery, monastery.author.token, monastery.csrftoken)
        monastery = Monastery.get_monastery_from_json(monastery_json)
        Monastery.all_monasteries = None
        return monastery

    #update in smp service
    def save(self, commit=True):
        smp_service = SmpWebService()
        monastery_json = smp_service.update_monastery(self, self.author.token, self.csrftoken)
        monastery = Monastery.get_monastery_from_json(monastery_json)
        Monastery.all_monasteries = None
        return monastery

    #delete in smp service
    def delete(self):
        smp_service = SmpWebService()
        smp_service.delete_monastery(self, self.author.token, self.csrftoken)
        Monastery.all_monasteries = None
        return self


    # validate unique
    def validate_unique(self, exclude):
        monasteries = Monastery.objects.all()
        for monastery in monasteries:
            if (self.name == monastery.name):
                if (hasattr(self, 'id')) or (self.id is not None):
                    if (self.id != monastery.id and self.name == monastery.name and self.province == monastery.province):
                        raise ValidationError({'monastery_name': 'Kloster med det namnet finns redan i detta landskap.'})
                else:
                    raise ValidationError('Kunde inte spara kloster.')


