from datetime import datetime

from django.db import models

from ISOF.model.archive_post import ArchivePost
from ISOF.model.language_smp import Language
from ISOF.model.lexical_name import LexicalName
from ISOF.model.name_last_element import NameLastElement
from ISOF.model.case import Case
from ISOF.model.person import Person
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class NameOccurrenceQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = NameOccurrence.get_all_name_occurrences()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        if kwargs and 'archive_post' in kwargs:
            archive_post_id = kwargs['archive_post'].id
            smp_service = SmpWebService()
            name_occurrences_json = smp_service.get_name_occurrences_by_archive_post(archive_post_id)
            if name_occurrences_json is None:
                #return empty queryset
                return NameOccurrence.objects.none()

            name_occurrences = NameOccurrence.get_name_occurrences_from_json(name_occurrences_json)
            if name_occurrences:
                for name_occurrence in name_occurrences:
                    smp_service = SmpWebService()
                    article_preview = smp_service.get_name_occurrence_article_text(name_occurrence.id)
                    if article_preview is not None:
                        name_occurrence.article_preview = article_preview['article']

            queryset = NameOccurrenceQuerySet(NameOccurrence)
            queryset._result_cache = name_occurrences
            queryset.create_dictionary(queryset._result_cache)
            return queryset
        else:
            queryset = NameOccurrenceQuerySet(LexicalName)
            queryset.create_dictionary(queryset._result_cache)
            return queryset


class NameOccurrenceManager(models.Manager):
    def get_queryset(self):
        queryset = NameOccurrenceQuerySet(NameOccurrence)
        return queryset


class NameOccurrence(models.Model):
    all_name_occurrences = None
    NAME_TYPE = (
        ('FirstName', 'Förnamn'),
        ('Surname', 'Tillnamn'),
    )

    archive_post = models.ForeignKey(ArchivePost, null=False, blank=False,
                                     on_delete=models.DO_NOTHING)
    case = models.ForeignKey(Case, null=True, blank=True,
                             on_delete=models.DO_NOTHING, verbose_name="Kasus")
    change_date = models.DateTimeField(null=False, blank=False, auto_now=True, verbose_name="Ändringsdatum")
    incorrect = models.BooleanField(default=False, null=False, verbose_name="Felaktig")
    is_in_book = models.BooleanField(default=True, null=False, verbose_name="Finns i boken")
    language = models.ForeignKey(Language, null=False, blank=False, on_delete=models.DO_NOTHING, verbose_name="Språk",
                                 default=96)
    lexical_name = models.ForeignKey(LexicalName, null=False, blank=False,
                                     on_delete=models.DO_NOTHING, verbose_name="Lexikalt namn")
    name_archive_post = models.CharField(null=False, blank=False, max_length=256, verbose_name="Namn belägg")
    name_last_element = models.ForeignKey(NameLastElement, null=True, blank=True,
                                          on_delete=models.DO_NOTHING, verbose_name="Efterled")
    name_lexical = models.CharField(null=True, blank=True, max_length=256, verbose_name="Namn lexikalt")
    person = models.ForeignKey(Person, null=True, blank=True, on_delete=models.DO_NOTHING, editable=True, verbose_name="Person")
    type = models.CharField(null=False, blank=False, choices=NAME_TYPE, max_length=64, verbose_name="Namntyp", default="FirstName")
    #preview article read only field only for presenting text
    article_preview = models.TextField(null=True, blank=True, verbose_name="Förhandsvisning av artikel")
    objects = NameOccurrenceManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "namnförekomst"
        verbose_name_plural = "namnförekomster"

    def __str__(self):
        return "Namnförekomst: " + self.name_archive_post

    @classmethod
    def get_all_name_occurrences(cls):
        if NameOccurrence.all_name_occurrences is None:
            smp_service = SmpWebService()
            name_occurrences_json = smp_service.get_name_occurrences()
            NameOccurrence.all_name_occurrences = NameOccurrence.get_name_occurrences_from_json(name_occurrences_json)
        return NameOccurrence.all_name_occurrences

    @classmethod
    def get_name_occurrences_from_json(cls, name_occurrences_json):
        name_occurrences = []
        for name_occurrence_json in name_occurrences_json:
            name_occurrence = NameOccurrence.get_name_occurrence_from_json(name_occurrence_json)
            name_occurrences.append(name_occurrence)
        return name_occurrences

    @classmethod
    def get_name_occurrence_from_json(cls, name_occurrence_json):
        name_occurrence = NameOccurrence()
        name_occurrence.archive_post = ArchivePost.objects.get(id=name_occurrence_json['archive_post'])
        name_occurrence.language = Language.objects.get(id=name_occurrence_json['language'])
        name_occurrence.change_date = datetime.strptime(name_occurrence_json['change_date'], '%Y-%m-%dT%H:%M:%S.%fZ')
        name_occurrence.incorrect = name_occurrence_json['incorrect']
        name_occurrence.is_in_book = name_occurrence_json['is_in_book']
        name_occurrence.lexical_name = LexicalName.objects.get(id=name_occurrence_json['lexical_name'])
        name_occurrence.name_archive_post = name_occurrence_json['name_archive_post']
        if name_occurrence_json['person'] is not None:
            name_occurrence.person = Person.objects.get(id=name_occurrence_json['person'])
        if name_occurrence_json['name_last_element'] is not None:
            name_occurrence.name_last_element = NameLastElement.objects.get(id=name_occurrence_json['name_last_element'])
        name_occurrence.name_lexical = name_occurrence_json['name_lexical']
        name_occurrence.type = name_occurrence_json['type']
        name_occurrence.id = name_occurrence_json['id']
        name_occurrence.pk = name_occurrence_json['id']
        return name_occurrence

    @classmethod
    def create(cls, name_occurrence):
        smp_service = SmpWebService()
        name_occurrence_json = smp_service.create_name_occurrence(name_occurrence, name_occurrence.author.token,
                                                                  name_occurrence.csrftoken)
        new_name_occurrence = NameOccurrence.get_name_occurrence_from_json(name_occurrence_json)
        NameOccurrence.all_name_occurrences = None
        return new_name_occurrence

    def save(self, commit=True):
        smp_service = SmpWebService()
        name_occurrence_json = smp_service.update_name_occurrence(self, self.author.token, self.csrftoken)
        NameOccurrence.all_name_occurrences = None
        return name_occurrence_json

    def delete(self):
        smp_service = SmpWebService()
        name_occurrence_json = smp_service.delete_name_occurrence(self, self.author.token, self.csrftoken)
        NameOccurrence.all_name_occurrences = None
        return name_occurrence_json
