# import django models
from django.core.exceptions import ValidationError
from django.db import models

from ISOF.model.district_smp import District
from ISOF.model.province_smp import Province
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class ParishQuerySet(SimpleModelQuerySet):
    # fetch all
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Parish.get_all_parishes()
            self.create_dictionary(self._result_cache)



    def filter(self, *args, **kwargs):
        if len(args) > 0:
            search_string = args[0].children[0].children[0][1]
            if search_string:
                parishes = []
                self._fetch_all()
                if self._result_cache is not None:
                    for parish in self._result_cache:
                        if search_string.casefold() in parish.name.casefold():
                            parishes.append(parish)
                queryset = ParishQuerySet(Parish)
                queryset._result_cache = parishes
                queryset.create_dictionary(queryset._result_cache)
                return queryset
        return self


# parish manager
class ParishManager(models.Manager):
    def get_queryset(self):
        queryset = ParishQuerySet(Parish)
        return queryset


class Parish(models.Model):
    all_parishes = None
    name = models.CharField(max_length=512, null=False, blank=False, verbose_name="Namn på socken")
    district = models.ForeignKey(District, on_delete=models.DO_NOTHING, null=True,
                                 blank=True, default=None, verbose_name="Härad (Sverige)")
    parish_id_isof = models.IntegerField(null=True, blank=True, default=None, verbose_name="Socken id Isof")
    province = models.ForeignKey(Province, on_delete=models.DO_NOTHING, null=True, blank=True,
                                 default=None, verbose_name="Landskap (Finland)", limit_choices_to={'country': 'Finland'})
    objects = ParishManager()

    #meta class
    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "socken"
        verbose_name_plural = "socknar"

    def __str__(self):
        return self.name


    #get all parishes and set in Parish.all_parishes
    @classmethod
    def get_all_parishes(cls):
        if Parish.all_parishes is None:
            smp_service = SmpWebService()
            parishes_json = smp_service.get_parishes()
            parishes = Parish.get_parishes_from_json(parishes_json)
            Parish.all_parishes = parishes
        return Parish.all_parishes

    @classmethod
    def get_parish_from_json(cls, parish_json):
        parish = Parish()
        parish.name = parish_json['name']
        if parish_json['district'] is not None:
            parish.district = District.objects.get(id=parish_json['district'])
        parish.parish_id_isof = parish_json['id']
        if parish_json['province'] is not None:
            parish.province = Province.objects.get(id=parish_json['province'])
        parish.id = parish_json['id']
        parish.pk = parish_json['id']
        if parish_json['province'] is not None:
            parish.province = Province.objects.get(id=parish_json['province'])
        return parish

    #get parishes from json class method
    @classmethod
    def get_parishes_from_json(cls, parishes_json):
        parishes = []
        districts = District.objects.all()
        for parish_json in parishes_json:
            parish = Parish()
            parish.name = parish_json['name']
            if parish_json['parish_id_isof'] is not None:
                parish.parish_id_isof = parish_json['id']
            if parish_json['district'] is not None:
                parish.district = districts.get(id=parish_json['district'])
            if parish_json['province'] is not None:
                parish.province = Province.objects.get(id=parish_json['province'])
            parish.id = parish_json['id']
            parish.pk = parish_json['id']
            parishes.append(parish)
        return parishes

    @classmethod
    def create(cls, parish):
        smp_service = SmpWebService()
        parish_json = smp_service.create_parish(parish, parish.author.token, parish.csrftoken)
        Parish.all_parishes = None
        return Parish.get_parish_from_json(parish_json)

    def delete(self):
        smp_service = SmpWebService()
        ok = smp_service.delete_parish(self, self.author.token, self.csrftoken)
        Parish.all_parishes = None
        return ok

    def save(self, commit=True):
        smp_service = SmpWebService()
        parish_json = smp_service.update_parish(self, self.author.token, self.csrftoken)
        Parish.all_parishes = None
        return Parish.get_parish_from_json(parish_json)


    #validate unique parish
    def validate_unique(self, exclude=None):
        parishes = Parish.objects.all()
        for parish in parishes:
            if (self.name == parish.name and self.district == parish.district):
                if (hasattr(self, 'id')) or (self.id is not None):
                    if (self.id != parish.id):
                        raise ValidationError(f'Det finns redan en socken "{self.name}" i "{self.district}.')
                else:
                    raise ValidationError(f'Kunde inte spara socken')

