from django.db import models

from ISOF.model.person import Person
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class PersonRelationQuerySet(SimpleModelQuerySet):
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = PersonRelation.get_all_person_relations()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        #if person in kwargs, filter by
        if kwargs and 'parent' in kwargs:
            parent = kwargs['parent']
            person_relations = []
            self._fetch_all()
            if self._result_cache is not None:
                for person_relation in self._result_cache:
                    if person_relation.parent == parent and person_relation.relation_type == 'ParentChild':
                        person_relations.append(person_relation)
            queryset = PersonRelationQuerySet(PersonRelation)
            queryset._result_cache = person_relations
            queryset.create_dictionary(queryset._result_cache)
            return queryset

        if kwargs and 'child' in kwargs:
            child = kwargs['child']
            person_relations = []
            self._fetch_all()
            if self._result_cache is not None:
                for person_relation in self._result_cache:
                    if person_relation.child == child and person_relation.relation_type == 'ParentChild':
                        person_relations.append(person_relation)
            queryset = PersonRelationQuerySet(PersonRelation)
            queryset._result_cache = person_relations
            queryset.create_dictionary(queryset._result_cache)
            return queryset

        if kwargs and 'person_1' in kwargs:
            person = kwargs['person_1']
            person_relations = []
            self._fetch_all()
            if self._result_cache is not None:
                for person_relation in self._result_cache:
                    if person_relation.person_1 == person and person_relation.relation_type != 'ParentChild':
                        person_relations.append(person_relation)
            queryset = PersonRelationQuerySet(PersonRelation)
            queryset._result_cache = person_relations
            queryset.create_dictionary(queryset._result_cache)
            return queryset

        if len(args) > 0:
            search_string = args[0].children[0].children[0][1]
            if search_string:
                person_relations = []
                self._fetch_all()
                if self._result_cache is not None:
                    for person_relation in self._result_cache:
                        if search_string.casefold() in person_relation.person_1.name.casefold():
                            person_relations.append(person_relation)
                queryset = PersonRelationQuerySet(PersonRelation)
                queryset._result_cache = person_relations
                queryset.create_dictionary(queryset._result_cache)
                return queryset
        return self

    def get_person_relations_by_person(self, person):
        person_relations = []
        self._fetch_all()
        if self._result_cache is not None:
            for person_relation in self._result_cache:
                if person_relation.person_1 == person:
                    person_relations.append(person_relation)
        queryset = PersonRelationQuerySet(PersonRelation)
        queryset._result_cache = person_relations
        return queryset

    def get_person_relations_by_person_id(self, person_id):
        person_relations = []
        self._fetch_all()
        if self._result_cache is not None:
            for person_relation in self._result_cache:
                if person_relation.person_1.id == person_id:
                    person_relations.append(person_relation)
        queryset = PersonRelationQuerySet(PersonRelation)
        queryset._result_cache = person_relations
        return queryset

    def get_person_relations_by_person_2(self, person):
        person_relations = []
        self._fetch_all()
        if self._result_cache is not None:
            for person_relation in self._result_cache:
                if person_relation.person_2 == person:
                    person_relations.append(person_relation)
        queryset = PersonRelationQuerySet(PersonRelation)
        queryset._result_cache = person_relations
        return queryset


class PersonRelationManager(models.Manager):
    def get_queryset(self):
        queryset = PersonRelationQuerySet(PersonRelation)
        return queryset


class PersonRelation(models.Model):
    all_person_relations = None
    RELATION_TYPE_CHOICES = (
        ('Probably', 'Troligen'),
        ('Possibly', 'Möjligen'),
        ('CompareWith', 'Jämför'),
        ('See', 'Se'),
        ('ParentChild', 'Förälder/Barn')
    )
    relation_type = models.CharField(max_length=128, null=False, blank=False, choices=RELATION_TYPE_CHOICES, verbose_name="Relationstyp")
    objects = PersonRelationManager()

    child = models.ForeignKey(Person, on_delete=models.DO_NOTHING, null=True, blank=True, related_name='child', verbose_name="Välj barn")
    parent = models.ForeignKey(Person, on_delete=models.DO_NOTHING, null=True, blank=True, related_name='parent', verbose_name="Välj förälder")
    person_1 = models.ForeignKey(Person, on_delete=models.DO_NOTHING, null=True, blank=True, related_name='person_1', verbose_name="Person 1")
    person_2 = models.ForeignKey(Person, on_delete=models.DO_NOTHING, null=True, blank=True, related_name='person_2', verbose_name="Välj relaterad person")


    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "Personrelation"
        verbose_name_plural = "Personrelationer"
        ordering = ['person_1__name']

    def __str__(self):
        if self.id is not None:
            return self.relation_type + ":" + str(self.id)
        else:
            return self.relation_type

    @classmethod
    def get_all_person_relations(cls):
        if PersonRelation.all_person_relations is None:
            smp_service = SmpWebService()
            person_relations_json = smp_service.get_person_relations()
            person_relations = PersonRelation.get_person_relations_from_json(person_relations_json)
            PersonRelation.all_person_relations = person_relations
        return PersonRelation.all_person_relations

    @classmethod
    def get_person_relations_from_json(cls, person_relations_json):
        person_relations = []
        for person_relation_json in person_relations_json:
            person_relation = PersonRelation.get_person_relation_from_json(person_relation_json)
            person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relation_from_json(cls, person_relation_json):
        person_relation = PersonRelation()
        if person_relation_json['child'] is not None:
            person_relation.child = Person.get_person_by_id(person_relation_json['child'])
        if person_relation_json['parent'] is not None:
            person_relation.parent = Person.get_person_by_id(person_relation_json['parent'])
        if person_relation_json['person_1'] is not None:
            person_relation.person_1 = Person.get_person_by_id(person_relation_json['person_1'])
        if person_relation_json['person_2'] is not None:
            person_relation.person_2 = Person.get_person_by_id(person_relation_json['person_2'])
        person_relation.id = person_relation_json['id']
        person_relation.pk = person_relation_json['id']
        person_relation.relation_type = person_relation_json['relation_type']
        return person_relation

    @classmethod
    def create(cls, person_relation):
        smp_service = SmpWebService()
        person_relation_json = smp_service.create_person_relation(person_relation, person_relation.author.token,
                                                                  person_relation.csrftoken)
        new_person_relation = PersonRelation.get_person_relation_from_json(person_relation_json)
        PersonRelation.all_person_relations = None
        return new_person_relation

    def save(self, commit=True):
        smp_service = SmpWebService()
        person_relation_json = smp_service.update_person_relation(self, self.author.token, self.csrftoken)
        PersonRelation.all_person_relations = None
        return person_relation_json

    def delete(self):
        smp_service = SmpWebService()
        person_relation_json = smp_service.delete_person_relation(self, self.author.token, self.csrftoken)
        PersonRelation.all_person_relations = None
        return person_relation_json

    @classmethod
    def get_person_relation_by_id(cls, person_relation_id):
        person_relations = PersonRelation.get_all_person_relations()
        for person_relation in person_relations:
            if person_relation.id == person_relation_id:
                return person_relation
        return None

    @classmethod
    def get_person_relations_by_person(cls, person):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_1 == person:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_person_id(cls, person_id):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_1.id == person_id:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_person_2(cls, person):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_2 == person:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_person_2_id(cls, person_id):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_2.id == person_id:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_child(cls, person):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.child == person:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_child_id(cls, person_id):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.child.id == person_id:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_parent(cls, person):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.parent == person:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_parent_id(cls, person_id):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.parent.id == person_id:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_person_1(cls, person):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_1 == person:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_person_1_id(cls, person_id):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_1.id == person_id:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_person_2(cls, person):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_2 == person:
                person_relations.append(person_relation)
        return person_relations

    @classmethod
    def get_person_relations_by_person_2_id(cls, person_id):
        person_relations = []
        person_relations_all = PersonRelation.get_all_person_relations()
        for person_relation in person_relations_all:
            if person_relation.person_2.id == person_id:
                person_relations.append(person_relation)
        return person_relations
