from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Exists, Q

from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class ProvinceQuerySet(SimpleModelQuerySet):
    last_country_filtering = None

    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Province.get_all_provinces()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        country_filter = None
        if len(args) > 0 and isinstance(args[0], Exists):
            if ProvinceQuerySet.last_country_filtering:
                provinces = ProvinceQuerySet.last_country_filtering
                ProvinceQuerySet.last_country_filtering = None
                return provinces
            else:
                return self

        if len(args) > 0 and isinstance(args[0], Q):
            if (hasattr(args[0], 'children')):
                country_filter = args[0].children[0][1]
            else:
                search_string = args[0].children[0].children[0][1]
                if search_string:
                    provinces = []
                    self._fetch_all()
                    if self._result_cache is not None:
                        for province in self._result_cache:
                            if (search_string.casefold() in province.name.casefold()):
                                provinces.append(province)
                        queryset = ProvinceQuerySet(Province)
                        queryset._result_cache = provinces
                        queryset.create_dictionary(queryset._result_cache)
                        return queryset

        if country_filter:
            provinces = []
            self._fetch_all()
            if self._result_cache is not None:
                for province in self._result_cache:
                    if province.country == country_filter:
                        provinces.append(province)
            queryset = ProvinceQuerySet(Province)
            queryset._result_cache = provinces
            ProvinceQuerySet.last_country_filtering = queryset
            return queryset

        return self


class ProvinceManager(models.Manager):
    def get_queryset(self):
        queryset = ProvinceQuerySet(Province)
        return queryset

    # Create model for Province


class Province(models.Model):
    # Create fields for Province
    name = models.CharField(max_length=250, null=False, blank=False, unique=True, default="", verbose_name="Namn")
    abbreviation = models.CharField(max_length=250, null=False, blank=False, unique=True, default="", verbose_name="Förkortning")
    country = models.CharField(max_length=250, null=False, blank=False, default="Sverige", verbose_name="Land")
    objects = ProvinceManager()
    all_provinces = None

    # Create meta class for Province
    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "landskap"
        verbose_name_plural = "landskap"

    #class method get all districts
    @classmethod
    def get_all_provinces(cls):
        if Province.all_provinces is None:
            smp_service = SmpWebService()
            provinces_json = smp_service.get_provinces()
            provinces = Province.get_provinces_from_json(provinces_json)
            Province.all_provinces = provinces
        return Province.all_provinces

    # get province from json
    @classmethod
    def get_province_from_json(cls, province_json):
        province = Province()
        province.name = province_json['name']
        province.abbreviation = province_json['abbreviation']
        province.country = province_json['country']
        province.id = province_json['id']
        province.pk = province_json['id']
        return province

    # get provinces from json
    @classmethod
    def get_provinces_from_json(cls, provinces_json):
        provinces = []
        for province_json in provinces_json:
            province = Province()
            province.name = province_json['name']
            province.abbreviation = province_json['abbreviation']
            province.country = province_json['country']
            province.id = province_json['id']
            province.pk = province_json['id']
            provinces.append(province)
        return provinces

    def __str__(self):
        return self.name

    # create province method as in model/abbreviation_smp.py
    @classmethod
    def create(cls, province):
        # Create province in SmpService.
        smp_service = SmpWebService()
        province_json = smp_service.create_province(province, province.author.token, province.csrftoken)
        Province.all_provinces = None
        return Province.get_province_from_json(province_json)

    def delete(self):
        smp_service = SmpWebService()
        ok = smp_service.delete_province(self, self.author.token, self.csrftoken)
        Province.all_provinces = None
        return ok

    def save(self, commit=True):
        # Save province in SmpService.
        smp_service = SmpWebService()
        smp_service.update_province(self, self.author.token, self.csrftoken)
        Province.all_provinces = None
        return self


    def validate_unique(self, exclude):
        """
        Check that unique fields are unique for this data type.
        """
        provinces = Province.objects.all()
        for province in provinces:
            if (self.name == province.name):
                if (hasattr(self, 'id')) or (self.id is None):
                    if (self.id != province.id):
                        raise ValidationError(f'Det finns redan ett landskap "{self.name}".')
                else:
                    raise ValidationError(f'Det finns redan en landskap "{self.name}".')

