from django.core.exceptions import ValidationError
from django.db import models

from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.web_service.smp_web_service import SmpWebService


class SourceQuerySet(SimpleModelQuerySet):
    #fetch all sources from smp service
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = Source.get_all_sources()
            self.create_dictionary(self._result_cache)

    def filter(self, *args, **kwargs):
        if len(args) > 0:
            search_string = args[0].children[0].children[0][1]
            if search_string:
                sources = []
                self._fetch_all()
                if self._result_cache is not None:
                    for source in self._result_cache:
                        if source.abbreviation.casefold().startswith(search_string.casefold()):
                            sources.append(source)
                queryset = SourceQuerySet(Source)
                queryset._result_cache = sources
                queryset.create_dictionary(queryset._result_cache)
                return queryset

        if (len(kwargs) > 0) and isinstance(kwargs, dict) and ('id__in' in kwargs):
            id_set = kwargs['id__in']
            sources = []
            if len(id_set) > 0:
                source_id = int(list(id_set)[0])
                if self.has_dictionary():
                    sources.append(self._result_cache_dictionary[source_id])
                else:
                    for obj in self._result_cache:
                        if obj.id == source_id:
                            sources.append(obj)
                            break

            queryset = SourceQuerySet(Source)
            queryset._result_cache = sources
            return queryset

        return self


class SourceManager(models.Manager):
    def get_queryset(self):
        queryset = SourceQuerySet(Source)
        return queryset


class Source(models.Model):
    all_sources = None
    SOURCE_TYPE = (
        ('HANDWRITTEN', 'Handskriven'),
        ('PRINTED', 'Tryckt'),
    )

    abbreviation = models.CharField(max_length=100, null=False, blank=True, verbose_name="Förkortning")
    source_name = models.TextField(max_length=500, null=False, blank=False, verbose_name="Källnamn")
    shelf_mark = models.CharField(max_length=20, null=True, blank=True, verbose_name="Signum")
    type = models.CharField(max_length=20, null=True, blank=True, choices=SOURCE_TYPE, verbose_name="Typ")
    series_number = models.IntegerField(null=True, blank=True, verbose_name="Serienummer")
    book_number_when_added = models.CharField(max_length=20, null=True, blank=True, verbose_name="Tillagd i bok nummer")
    is_original = models.BooleanField(null=False, verbose_name="Original", default=False)

    several_volumes = models.BooleanField(null=True, blank=True, verbose_name="Flera volymer")
    removed_changed_problem_comment = models.TextField(max_length=4000, null=True, blank=True, verbose_name="Kommentar")
    reference = models.CharField(max_length=2000, null=True, blank=True, verbose_name="Referens")

    objects = SourceManager()

    class Meta:
        base_manager_name = 'objects'
        managed = True
        verbose_name = "källa"
        verbose_name_plural = "källor"

    def __str__(self):
        if self.abbreviation:
            if self.source_name:
                return "(" + self.abbreviation + ")" + " - " + self.source_name
            else:
                return self.abbreviation
        else:
            return self.source_name
    @classmethod
    def get_all_sources(cls):
        if Source.all_sources is None:
            smp_service = SmpWebService()
            sources_json = smp_service.get_sources()
            sources = Source.get_sources_from_json(sources_json)
            Source.all_sources = sources
        return Source.all_sources

    @classmethod
    def get_sources_from_json(cls, sources_json):
        sources = []
        for source_json in sources_json:
            source = Source()
            source.source_name = source_json['source_name']
            source.shelf_mark = source_json['shelf_mark']
            source.type = source_json['type']
            source.series_number = source_json['series_number']
            source.book_number_when_added = source_json['book_number_when_added']
            source.is_original = source_json['is_original']
            source.abbreviation = source_json['abbreviation']
            source.several_volumes = source_json['several_volumes']
            source.removed_changed_problem_comment = source_json['removed_changed_problem_comment']
            source.reference = source_json['reference']
            source.id = source_json['id']
            sources.append(source)
        return sources

    @classmethod
    def get_source_from_json(cls, source_json):
        source = Source()
        source.source_name = source_json['source_name']
        source.shelf_mark = source_json['shelf_mark']
        source.type = source_json['type']
        source.series_number = source_json['series_number']
        source.book_number_when_added = source_json['book_number_when_added']
        source.is_original = source_json['is_original']
        source.abbreviation = source_json['abbreviation']
        source.several_volumes = source_json['several_volumes']
        source.removed_changed_problem_comment = source_json['removed_changed_problem_comment']
        source.reference = source_json['reference']
        source.id = source_json['id']
        return source

    @classmethod
    def create(cls, source):
        smp_service = SmpWebService()
        source_json = smp_service.create_source(source, source.author.token, source.csrftoken)
        Source.all_sources = None
        return Source.get_source_from_json(source_json)

    def save(self, commit=True):
        smp_service = SmpWebService()
        source_json = smp_service.update_source(self, self.author.token, self.csrftoken)
        source = Source.get_source_from_json(source_json)
        Source.all_sources = None
        return source

    def delete(self):
        smp_service = SmpWebService()
        ok = smp_service.delete_source(self, self.author.token, self.csrftoken)
        Source.all_sources = None
        return ok

    #validate unique
    def validate_unique(self, exclude):
        sources = Source.objects.all()
        for source in sources:
            if (hasattr(self, 'id')) or (self.id is not None):
                if source.source_name == self.source_name and source.id != self.id:
                    raise ValidationError(
                        {'source_name': ('Källan med namnet %(value)s finns redan i databasen') % {'value': self.source_name}})

            else:
                raise ValidationError('Kunde inte spara källa.')


