from django import forms
from django.db import models

from ISOF.model.archive_post import ArchivePost
from ISOF.model.person import Person
from ISOF.model.simple_model_query_set import SimpleModelQuerySet
from ISOF.model.source import Source
from ISOF.model.web_service.smp_web_service import SmpWebService


class SourceReferenceQuerySet(SimpleModelQuerySet):
    # fetch all
    def _fetch_all(self):
        if (not hasattr(self, '_result_cache')) or (self._result_cache is None):
            self._result_cache = SourceReference.get_all_source_references()
            self.create_dictionary(self._result_cache)

    # filter
    def filter(self, *args, **kwargs):
        if kwargs and 'archive_post' in kwargs:
            archive_post_id = kwargs['archive_post'].id
            smp_service = SmpWebService()
            source_references = smp_service.get_source_references_by_archive_post(archive_post_id)
            if source_references is None:
                # return empty queryset
                return SourceReference.objects.none()

            name_occurrences = SourceReference.get_source_references_from_json(source_references)
            queryset = SourceReferenceQuerySet(SourceReference)
            queryset._result_cache = name_occurrences
            queryset.create_dictionary(queryset._result_cache)
            return queryset
        else:
            queryset = SourceReferenceQuerySet(SourceReference)
            queryset.create_dictionary(queryset._result_cache)
            return queryset


# get queryset
class SourceReferenceManager(models.Manager):
    def get_queryset(self):
        queryset = SourceReferenceQuerySet(SourceReference)
        return queryset


class SourceReference(models.Model):
    all_source_references = None
    source = models.ForeignKey(Source, null=False, on_delete=models.DO_NOTHING, verbose_name="Källa")
    author_of_source = models.CharField(max_length=512, null=True, blank=True, verbose_name="Författare")
    archive_post = models.ForeignKey(ArchivePost, null=True, on_delete=models.DO_NOTHING, verbose_name="Belägg")
    start_page = models.IntegerField(null=True, blank=True, verbose_name="Startsida")
    has_verse_number = models.BooleanField(default=False, null=False, verbose_name="Versnumrering")
    end_page = models.IntegerField(null=True, blank=True, verbose_name="Slutsida")
    has_roman_page_number = models.BooleanField(default=False, null=False, verbose_name="Romersk sidnummrering")
    is_main_source = models.BooleanField(default=True, null=False, verbose_name="Huvudkälla")
    person = models.ForeignKey(Person, null=True, blank=True, on_delete=models.DO_NOTHING, verbose_name="Person")
    sdhk_number = models.IntegerField(null=True, blank=True, verbose_name="SDHK-nummer")
    serial_number = models.CharField(max_length=100, null=False, blank=True, verbose_name="Löpnummer")

    volume_number = models.IntegerField(null=True, blank=True, verbose_name="Bandnummer")

    objects = SourceReferenceManager()

    # meta
    class Meta:
        base_manager_name = 'objects'
        ordering = ['id']
        verbose_name_plural = "Källhänvisningar"
        verbose_name = "Källhänvisning"

    # string method
    def __str__(self):
        if (hasattr(self, 'id')) and (self.id is not None):
            return f'Källhänvisning: {self.id}'
        else:
            return f'Källhänvisning: ej sparat i databasen'

    @classmethod
    def get_all_source_references(cls):
        if SourceReference.all_source_references is None:
            smp_service = SmpWebService()
            source_references_json = smp_service.get_source_references()
            SourceReference.all_source_references = SourceReference.get_source_references_from_json(
                source_references_json)
        return SourceReference.all_source_references

    @classmethod
    def get_source_references_from_json(cls, source_references_json):
        source_references = []
        for source_reference in source_references_json:
            source_reference = SourceReference.get_source_reference_from_json(source_reference)
            source_references.append(source_reference)
        return source_references

    @classmethod
    def get_source_reference_from_json(cls, source_reference_json):
        source_reference = SourceReference()
        persons = Person.objects.all()
        sources = Source.objects.all()

        source_reference.archive_post = ArchivePost.objects.get(id=source_reference_json['archive_post'])
        source_reference.author_of_source = source_reference_json['author_of_source']
        source_reference.end_page = source_reference_json['end_page']
        source_reference.has_roman_page_number = source_reference_json['has_roman_page_number']
        source_reference.is_main_source = source_reference_json['is_main_source']
        if source_reference_json['person'] is not None:
            source_reference.person = persons.get(id=source_reference_json['person'])
        source_reference.sdhk_number = source_reference_json['sdhk_number']
        source_reference.serial_number = source_reference_json['serial_number']
        source_reference.source = sources.get(id=source_reference_json['source'])
        source_reference.start_page = source_reference_json['start_page']
        source_reference.volume_number = source_reference_json['volume_number']
        source_reference.has_verse_number = source_reference_json['has_verse_number']
        source_reference.id = source_reference_json['id']
        source_reference.pk = source_reference_json['id']
        return source_reference

    # create
    @classmethod
    def create(cls, source_reference):
        smp_service = SmpWebService()
        source_reference_json = smp_service.create_source_reference(source_reference, source_reference.author.token,
                                                                    source_reference.csrftoken)

        new_source_reference = SourceReference.get_source_reference_from_json(source_reference_json)
        SourceReference.all_source_references = None
        return new_source_reference

    # save
    def save(self, commit=True):
        smp_service = SmpWebService()
        source_reference_json = smp_service.update_source_reference(self, self.author.token, self.csrftoken)
        SourceReference.all_source_references = None
        return source_reference_json

    # delete
    def delete(self):
        smp_service = SmpWebService()
        ok = smp_service.delete_source_reference(self, self.author.token, self.csrftoken)
        SourceReference.all_source_references = None
        return ok
