import collections
import itertools

from model.database import Database
from model.import_util import textconvert_annotated_term
from model.kalla_sort_key import KallaSortKey
import model.mongodbcfg
from model.source_file_term_post_field import term_language_priority
from model.term_search_result import TermSearchResult
from model.term_sort_key import TermSortKey
from model.regex_search_string_expression import RegexSearchStringExpression
from model.string_search_string_expression import StringSearchStringExpression
from model.text_prio_key import TextPrioKey
import pymongo
import pymongo.errors
from pymongo import MongoClient
from pymongo.collation import Collation

match_quality_priority = {
    "exact": 0,
    "partial": 1,
}

term_priority = {
    "TE": 0,
    "SYTE": 1,
    "AVTE": 2,
}

text_type_priority = {
    "definition": 0,
    "explanation": 1,
}


class MongoDBManager():
    """ Handles information in MongoDB.

    Attributes:
        database:               MongoDB connection.
        rikstermbanken_public:  Reference to public database.
        rikstermbanken_staging: Reference to staging database.
    """

    SWEDISH_LANGUAGE_CODE = 'sv'

    database = MongoClient(connect=False,
                           username=model.mongodbcfg.MONGODB_USERNAME,
                           password=model.mongodbcfg.MONGODB_PASSWORD,
                           authSource=model.mongodbcfg.MONGODB_DB)
    rikstermbanken_public = database.rikstermbanken
    rikstermbanken_staging = database.rikstermbanken_staging

    @staticmethod
    def _get_database(database: str):
        """ Get a reference to requested MongoDB database.

        Arguments:
            database: Name of the requested database.

        Returns:
            A reference to requested MongoDB database.
        """

        db = None
        if (database == Database.PUBLIC_DATABASE_NAME):
            db = MongoDBManager.rikstermbanken_public
        if (database == Database.STAGING_DATABASE_NAME):
            db = MongoDBManager.rikstermbanken_staging
        assert(db is not None)
        return db

    @staticmethod
    def _get_swedish_collation():
        """ Get Swedish collation.

        Returns:
            Swedish collation.
        """

        return Collation("sv@collation=search", strength=2)

    @staticmethod
    def get_term_post_by_reference(term_post_reference, term_post_source_id, database):
        """ Get term post that matches search criteria.

        Arguments:
            term_post_reference:    Term, language and optional homonym that term post should match.
            term_post_source_id:    Id for the term post source.
            database:               Name of the requested database.

        Returns:
            Term post that matches search criteria.
        """

        db = MongoDBManager._get_database(database)
        term = term_post_reference["term"]
        homonym = term_post_reference.get("homonym")
        language = term_post_reference["lang"]
        if (homonym):
            search_criteria = {"term": term, "homonym": homonym, "lang": language, "status": "TE"}
        else:
            search_criteria = {"term": term, "homonym": {"$exists": False}, "lang": language, "status": "TE"}
        term_post = db.termpost.find_one({"terms": {"$elemMatch": search_criteria}, "kalla.id": term_post_source_id})
        return term_post

    @staticmethod
    def get_term_post_by_slug(slug: str, term_post_source_id: int, database: str):
        """ Get term post that matches search criteria.

        Arguments:
            slug:                   Slug that identifies one term post in specified term post source.
            term_post_source_id:    Id for the term post source.
            database:               Name of the requested database.

        Returns:
            Term post that matches search criteria.
        """

        db = MongoDBManager._get_database(database)
        term_post = db.termpost.find_one({"kalla.id": term_post_source_id, "slugs": slug})

        if (term_post is None):
            # Temporary fix to avoid crash in web application.
            # The problem is that more than one term has been joined together to one slug.
            # TODO: Remove this fix when the reason to the problem has been removed.
            comma_index = slug.find(',')
            if (comma_index != -1):
                slug = slug[0:comma_index]
                term_post = db.termpost.find_one({"kalla.id": term_post_source_id, "slugs": slug})

        return term_post

    @staticmethod
    def get_term_post_count(database):
        """ Get term post count.

        Arguments:
            database:                   Name of the requested database.

        Returns:
            Term post count.
        """

        db = MongoDBManager._get_database(database)
        term_post_count = db.termpost.count_documents({})
        return term_post_count

    @staticmethod
    def _get_term_post_search_criteria(search_string_expressions,
                                       search_type: str,
                                       term_languages: list) -> dict:
        """ Get term post search criteria.

        Arguments:
            search_string_expressions:  Search expression to match.
            search_type:                What data should be search, terms or all texts.
            term_languages:             Limit search to term on these languages.

        Returns:
            Term post search criteria.
        """

        if search_type == "termer":
            search_field = "searchterms"
        else:
            search_field = "fulltextsearch"
        search_criteria = {
            "kalla.status": 2,
            search_field: {"$all": [e.mongodbsearchversion() for e in search_string_expressions]}
        }

        if (len(term_languages) > 0):
            search_criteria["terms.lang"] = {"$all": term_languages}
        return search_criteria

    @staticmethod
    def get_term_post_source(term_post_source_id: int, database: str):
        """ Get term post source with specified id.

        Arguments:
            term_post_source_id:  Term post source id
            database:             Name of the requested database.

        Returns:
            Term post source or None if no term post source has the specified id.
        """

        db = MongoDBManager._get_database(database)
        term_post_source = db.kalla.find_one({"id": term_post_source_id})
        if ((term_post_source is not None) and ('term_post_count' not in term_post_source)):
            term_post_count = db.termpost.count_documents({"kalla.id": term_post_source_id})
            term_post_source['term_post_count'] = term_post_count
        return term_post_source

    @staticmethod
    def get_term_post_source_count(only_published_sources, database):
        """ Get term post source count.

        Arguments:
            only_published_sources:  Indicates if only published sources should be included in the count.
            database:                Name of the requested database.

        Returns:
            Term post source count.
        """

        db = MongoDBManager._get_database(database)
        if (only_published_sources):
            term_post_count = db.kalla.count_documents({"status": {"$eq": 2}})
        else:
            term_post_count = db.kalla.count_documents({})
        return term_post_count

    @staticmethod
    def get_term_post_sources(database: str) -> list:
        db = MongoDBManager._get_database(database)
        return list(db.kalla.find({}))

    @staticmethod
    def get_term_post_sources_by_term_posts(only_published_sources, database):
        """ Get term post sources.

            The difference between this method and method get_term_post_sources() is that this method returns
            a different type of term post source data and that this data is sorted in a specified order.

        Arguments:
            only_published_sources:  Indicates if only published sources should be included in the result.
            database:                Name of the requested database.

        Returns:
            Term post sources.
        """

        # TODO: This method should be removed. Used the same TermPostSource class and sort order in both methods.
        db = MongoDBManager._get_database(database)
        if (only_published_sources):
            sources = db.termpost.aggregate(
                [{"$match": {"kalla.status": 2}}, {"$group": {"_id": "$kalla", "count": {"$sum": 1}}}]
            )
        else:
            sources = db.termpost.aggregate(
                [{"$match": {}}, {"$group": {"_id": "$kalla", "count": {"$sum": 1}}}]
            )
        sources = sorted(sources, key=lambda source: (source["_id"]["publisher"],
                                                      source["_id"]["title"] or "",
                                                      source["_id"]["id"]))
        return sources


    @staticmethod
    def get_term_posts_by_search_criteria(search_string: str,
                                          search_type: str,
                                          search_languages: list,
                                          definition_languages: list,
                                          term_languages: list):
        # TODO: Refactor this functionality.
        search_string_expressions = []
        for word in search_string.split(" "):
            if (("?" in word) or ("*" in word)):
                search_string_expressions.append(RegexSearchStringExpression.create(word))
            else:
                search_string_expressions.append(StringSearchStringExpression(word))

        search_criteria = MongoDBManager._get_term_post_search_criteria(search_string_expressions,
                                                                        search_type,
                                                                        term_languages)
        results = MongoDBManager.rikstermbanken_public.termpost.find(search_criteria,
                                                                     collation=MongoDBManager._get_swedish_collation(),
                                                                     limit=1001)
        results = sorted(results, key=lambda x: MongoDBManager._sortkey(x, search_string_expressions))

        too_many = len(results) > 1000
        if too_many:
            results = results[:500]
        more = len(results) > 500
        results = results[:500]
        lang_matches = collections.Counter()
        term_langs = collections.Counter()
        definition_langs = collections.Counter()
        for result in results:
            matches = MongoDBManager._match(result.get("terms", []), search_string_expressions)
            result["matches"] = MongoDBManager._group_match_lang(matches)
            result["priotext"] = sorted(filter(lambda x: MongoDBManager._text_sortkey(x).type < 100,
                                               result.get("texts", [])),
                                        key=MongoDBManager._text_sortkey)
            result["sortkey"] = MongoDBManager._sortkey(result, search_string_expressions)
            result_lang_matches = set()
            result_term_lang_matches = set()
            result_definition_langs = set()
            for e in matches:
                if e["type"] != "canonical":
                    result_lang_matches.add(e["match"]["lang"])

            result["matches_searchlang"] = True
            if (search_languages is not None):
                for searchlang in search_languages:
                    if (searchlang not in result_lang_matches):
                        result["matches_searchlang"] = False
                        break

            for e in result["terms"]:
                result_term_lang_matches.add(e["lang"])
            for e in result.get("texts", []):
                if e["type"] == "definition":
                    result_definition_langs.add(e["lang"])

            result["matches_deflang"] = True
            for deflang in definition_languages:
                if (deflang not in result_definition_langs):
                    result["matches_deflang"] = False

            if result["matches_searchlang"] and result["matches_deflang"]:
                lang_matches.update(result_lang_matches)
                term_langs.update(result_term_lang_matches)
                definition_langs.update(result_definition_langs)

        results = [result for result in results if result["matches_searchlang"] and result["matches_deflang"]]

        term_search_result = TermSearchResult()
        term_search_result.definition_langs = MongoDBManager._sort_lang_items(definition_langs)
        term_search_result.lang_matches = MongoDBManager._sort_lang_items(lang_matches)
        term_search_result.more = more
        term_search_result.results = results
        term_search_result.search_string_expressions = search_string_expressions
        term_search_result.term_langs = MongoDBManager._sort_lang_items(term_langs)
        term_search_result.search_langs = MongoDBManager._sort_lang_items(lang_matches)
        term_search_result.too_many = too_many
        return term_search_result

    @staticmethod
    def get_term_posts_by_term_post_source(term_post_source_id, only_swedish_term_posts, database):
        """ Get term post that belongs to specified term post source.

        Arguments:
            term_post_source_id:        Term post source id.
            only_swedish_term_posts:    Return only Swedish term post.
            database:                   Name of the requested database.

        Returns:
            Term post that belongs to specified term post source.
        """

        db = MongoDBManager._get_database(database)
        if (only_swedish_term_posts):
            term_posts = list(db.termpost.find({"kalla.id": term_post_source_id},
                                               collation=MongoDBManager._get_swedish_collation(),
                                               limit=100000))
        else:
            term_posts = list(db.termpost.find({"kalla.id": term_post_source_id}, limit=100000))
        return term_posts

    @staticmethod
    def _get_term_text(term):
        if "term" in term:
            return term["term"]
        elif "annotated_term" in term:
            return textconvert_annotated_term(term["annotated_term"])
        else:
            return ""

    @staticmethod
    def _group_match_lang(term_list):
        result = []
        for k, g in itertools.groupby(term_list, key=lambda x: x["match"]["lang"]):
            result.append({"lang": k, "matches": list(g)})
        return result

    @staticmethod
    def _match(terms, search_string_expressions):
        matches = []
        for term in terms:
            term_text = MongoDBManager._get_term_text(term)
            # splitterm = [t.casefold() for t in term_text.split(" ")]
            if all([e.isString() for e in search_string_expressions]) and term_text.casefold() == " ".join(
                    [e.casefold() for e in search_string_expressions]):
                matches.append({"type": "exact", "match": term, "text": term_text})
            else:
                matched = False
                for searchStringExpression in search_string_expressions:
                    if searchStringExpression.matchesAny(term_text.split(" ")):
                        matched = True
                if matched:
                    matches.append({"type": "partial", "match": term, "text": term_text})
        matches = sorted(matches, key=MongoDBManager._match_sort_key)
        matches = MongoDBManager._restrict_matches(matches)
        if not any([match["match"]["status"] == "TE" for match in matches]):
            matches = [{"type": "canonical", "match": term, "text": MongoDBManager._get_term_text(term)} for term in
                       terms if
                       term["status"] == "TE" and term["lang"] == "sv"] + matches
        if not any([match["match"]["lang"] == "sv" for match in matches]):
            matches = [{"type": "canonical", "match": term, "text": MongoDBManager._get_term_text(term)} for term in
                       terms if
                       term["status"] == "TE" and term["lang"] == "sv"] + matches
        if not any([match["match"]["lang"] == "sv" and match["match"]["status"] == "TE" for match in matches]):
            matches = [{"type": "canonical", "match": term, "text": MongoDBManager._get_term_text(term)} for term in
                       terms if
                       term["status"] == "TE" and term["lang"] == "sv"] + matches
        return matches

    @staticmethod
    def _match_sort_key(match):
        match_quality = match_quality_priority.get(match["type"], 100)
        return (match_quality,
                term_language_priority.get(match["match"]["lang"], 100),
                term_priority.get(match["match"]["status"], 100))

    @staticmethod
    def reset_database(database):
        db = MongoDBManager._get_database(database)
        db.termpost.delete_many({})
        db.kalla.delete_many({})
        db.termpost.drop_indexes()
        db.termpost.create_index([("id", pymongo.ASCENDING)])
        db.termpost.create_index([("terms.term", pymongo.ASCENDING), ("kalla.id", pymongo.ASCENDING)])
        db.termpost.create_index([("searchterms", pymongo.ASCENDING)],
                                 collation=MongoDBManager._get_swedish_collation())
        db.termpost.create_index([("kalla.id", pymongo.ASCENDING), ("slugs", pymongo.ASCENDING)], unique=True)

    @staticmethod
    def _restrict_matches(matches):
        if not matches:
            return matches
        grouped = MongoDBManager._group_match_lang(matches)
        if grouped[0]["lang"] == "sv":
            if len(grouped) > 1:
                return grouped[0]["matches"] + [grouped[1]["matches"][0]]
            else:
                return grouped[0]["matches"]
        else:
            return [grouped[0]["matches"][0]]

    @staticmethod
    def _sortkey(term, search_string_expressions):
        #    print(term)
        # Kriterium matchspråk (sv, en, ...)
        # Kriterium matchplats (term, annan plats)
        # Kriterium matchkvalitet (exakt, annan match)
        # Kriterium sorteringsterm (alfabetiskt)
        matches = list(filter(lambda x: x["type"] != "canonical",
                              MongoDBManager._match(term.get("terms", []), search_string_expressions)))
        matchlangprio = min([term_language_priority.get(e["match"]["lang"], 100) for e in matches] + [100])
        matchquality = min([match_quality_priority.get(e["type"], 100) for e in matches] + [100])
        sorteringsterm = [e["text"].lower().replace("-", " ") for e in matches]
        # Kriterium källa (efter år, sedan titel, sedan utgivare)
        if term["kalla"].get("year"):
            try:
                year = -int(term["kalla"]["year"])
            except ValueError:
                year = 0
        else:
            year = 0
        kalla = KallaSortKey(
            year,
            term["kalla"].get("title") or "",
            term["kalla"].get("publisher") or "",
        )
        # Criteria homonym_number
        # XXX should sort homonym number together with term
        homonym_number = [e["match"].get("homonym", 0) for e in matches]
        # Kriterium definition/förklaring (alfabetiskt)
        # Kriterium lankterm (alfabetiskt)
        tiebreak_terms = [MongoDBManager._get_term_text(e) for e in sorted(term.get("terms", []), key=lambda x: (
            term_language_priority.get(x["lang"], 100), term_priority.get(x["status"], 100), x.get("term", "")))]
        return TermSortKey(matchlangprio, matchquality, sorteringsterm, kalla, homonym_number, tiebreak_terms)

    @staticmethod
    def _sort_lang_items(items):
        return sorted(items.items(), key=lambda item: term_language_priority.get(item[0], 100))

    @staticmethod
    def _text_sortkey(text):
        type_prio = text_type_priority.get(text["type"], 100)
        lang_prio = term_language_priority.get(text["lang"], 100)
        return TextPrioKey(type_prio, lang_prio)

    @staticmethod
    def update_term_post_source(term_post_source, term_post_source_id, database):
        """ Update term post source in specified database.

        Arguments:
            term_post_source:       Term post source.
            term_post_source_id:    Term post source id.
            database:               Name of the requested database.
        """

        db = MongoDBManager._get_database(database)
        db.kalla.delete_one({"id": term_post_source_id})
        db.kalla.replace_one({"id": term_post_source_id}, term_post_source, upsert=True)

    @staticmethod
    def update_term_posts(term_posts, term_post_source_id, database):
        """ Update term posts in specified database.

        Arguments:
            term_posts:             Term posts.
            term_post_source_id:    Term post source id.
            database:               Name of the requested database.
        """

        db = MongoDBManager._get_database(database)
        db.termpost.delete_many({"kalla.id": term_post_source_id})
        if term_posts:
            db.termpost.insert_many(term_posts)
