from ISOF.common.list_functions import is_not_empty_list
import json
from model.import_util import unique
from model.source_file_header import SourceFileHeader
from model.source_file_metadata import SourceFileMetadata
from model.source_file_segment import SourceFileSegment
from model.source_file_term_post import SourceFileTermPost
from model.term_post_source_old import TermPostSourceOld
import os
import os.path


class SourceFile():
    """ Holds information about one file in "Bankvalvet".
    """

    def __init__(self, filename):
        self.is_exception_raised = False
        self.filename = filename
        self.basename = os.path.basename(filename)
        self._log = []
        self.metadata = None
        self.source_id = None
        self.term_post_documents = []
        self.term_post_source = None
        self.term_posts = []

        try:
            with open(self.filename, encoding="utf-8") as file:
                self._read_file(file)

                if (self.is_ok()):
                    self.source_id = self.term_post_source.id
                    self.source_status = self.term_post_source.status
                    self._check_references()
                    self._check_definitions()
                    self._get_term_post_documents()
                    self.term_post_source.set_term_post_count(len(self.term_post_documents))
                    self.term_post_source.set_error_log(self._log)
                    print(self.basename + ' ' + str(len(self.term_posts)))

        except UnicodeDecodeError as error:
            self.log("Fel i teckenhantering i filen. ", error)
            self.is_exception_raised = True
        except ValueError as error:
            self.log("Fel på ett värde i filen. ", error)
            self.is_exception_raised = True
        # except Exception as error:
            # self.log("Fel vid inläsning av filen. ", error)
            # self.is_exception_raised = True

        if (self.is_exception_raised):
            # Error handling.
            # Add source with minimum amount of information.
            self.term_post_source_document = {}
            if (self.term_post_source):
                self.term_post_source_document['filename'] = self.term_post_source.filename
                self.term_post_source_document['id'] = self.term_post_source.id
            else:
                filename = os.path.basename(self.filename)
                self.term_post_source_document['filename'] = filename
                self.term_post_source_document['id'] = -1
            self.term_post_source_document["errors"] = self.get_log()

    def best_match_term(self, term_entries, term):
        # print("best_match_term", term_entries)
        matches = [self.get_match(e, term) for e in term_entries]
        if len(term_entries) == 0:
            return None
        elif len(term_entries) == 1:
            return term_entries[0]
        elif len(term_entries) == 2:
            if matches[0]["status"] == "TE" and matches[1]["status"] == "SYTE":
                return term_entries[0]
            elif matches[0]["status"] == "SYTE" and matches[1]["status"] == "TE":
                return term_entries[1]
            else:
                return term_entries[0]
        # print("    ", len(term_entries), matches)
        assert (len(term_entries) == 1)
        term_entry = term_entries[0]
        return term_entry

    def _check_definitions(self):
        """ Check that the same Swedish definition (field DF) does not occur more than once in the same source file.

            This check is only done for recently published term post files.

            Handling (by terminologists) if the same definition occur more than once:
            Term posts should be joined into one term post where extra terms should be added as synonyms.
            Decision by Karin Webjörn 2023-03-09.
        """

        if (self.term_post_source and self.term_post_source.is_recently_published()):
            # Get all definitions in this file.
            definitions = {}
            for term_post in self.term_posts:
                if (term_post.has_definitions()):
                    term_post_definitions = term_post.get_definitions()
                    for term_post_definition in term_post_definitions:
                        term_post_definition = term_post_definition.lower()
                        if (term_post_definition in definitions):
                            term_posts = definitions[term_post_definition]
                            term_posts.append(term_post)
                        else:
                            term_posts = []
                            term_posts.append(term_post)
                            definitions[term_post_definition] = term_posts

            # Check if the same definition occurs more than once.
            for definition in definitions.keys():
                if (len(definitions[definition]) > 1):
                    # More than one term post with the same definition. Handle error.
                    self.log(f'Flera termposter har samma definition. Definition = "{definition}"')
                    for term_post in definitions[definition]:
                        # The term post is ok but the source file needs more work.
                        # This logging is only for information.
                        term_post.log_term_post()

    def _check_references(self):
        if (self.is_ok()):
            # Get references to all terms in source file.
            terms_lookup = {}
            for term_post in self.term_posts:
                if (term_post.is_ok):
                    for term in term_post.document["terms"]:
                        if (term["status"] == "AVTE"):
                            continue
                        language = term["lang"]
                        term_key = (term.get("term", ""), term.get("homonym", 0))
                        terms_lookup.setdefault(language, {})
                        terms_lookup[language].setdefault(term_key, []).append(term_post.document)

            for term_post in self.term_posts:
                if (term_post.is_ok):
                    document = term_post.document

                    if ("seealso" in document):
                        seealso = [self.see_also_convert(term_post, e, terms_lookup) for e in document["seealso"]]
                        seealso = list(filter(lambda x: x is not None, seealso))
                        document["seealso"] = unique(seealso, key=lambda x: json.dumps(x, sort_keys=True))
                        if (not document["seealso"]):
                            del document["seealso"]

                    if ("seeunder" in document):
                        seealso = [self.see_also_convert(term_post, e, terms_lookup) for e in document["seeunder"]]
                        seealso = list(filter(lambda x: x is not None, seealso))
                        document["seeunder"] = unique(seealso, key=lambda x: json.dumps(x, sort_keys=True))
                        if not document["seeunder"]:
                            del document["seeunder"]

                    if ("onlysearch" in document):
                        document["onlysearch"] = unique(document["onlysearch"], key=lambda x: json.dumps(x, sort_keys=True))
                    document["terms"] = unique(document["terms"], key=lambda x: json.dumps(x, sort_keys=True))
                    document["kalla"] = self.term_post_source.get_document_for_term_post()

    def get_log(self):
        return self._log.copy()

    def get_match(self, term_entry, term):
        matches = [e for e in term_entry["terms"] if e["term"] == term]
        # print("get_match", term_entry)
        # print(term)
        # print(matches)
        return matches[0]

    def _get_segments(self, file):
        self.segments = []
        more_segments = True
        while (more_segments):
            segment = SourceFileSegment(self, file)
            more_segments = is_not_empty_list(segment.rows)
            if (more_segments):
                self.segments.append(segment)
        return self.segments

    def _get_term_post_documents(self):
        if (self.is_ok()):
            for term_post in self.term_posts:
                if (term_post.is_ok):
                    entry = term_post.document
                    self.term_post_documents.append(entry)
                # else:
                    # term_post.log_term_post()

    def get_term_post_source_document(self):
        if (self.is_exception_raised):
            return self.term_post_source_document
        else:
            return self.term_post_source.get_document()

    def has_source_id(self):
        return ((self.source_id is not None) and (self.source_id > -1))

    def is_ok(self):
        # Can not import terms without source id.
        # Should not import terms with unknown status.
        return (self.term_post_source and self.term_post_source.has_id() and self.term_post_source.has_status())

    def log(self, *s):
        self._log.append([str(e) for e in s])

    def _read_file(self, file):
        self.header = None

        for segment in self._get_segments(file):
            if (segment.is_header_segment()):
                self.header = SourceFileHeader(segment.rows)
            elif (segment.is_metadata_segment()):
                if (self.term_post_source is not None):
                    # File bankvalvet/Offentlig verksamhet/Statliga myndigheter/Förvaltningsmyndigheter/Arbetsmiljöverket/Övrigt/Fil med flera metadatahuvuden/Arbetsmiljöverket_termer_imkand.txt
                    # has more than one meta data segment. Currently only the first part of the file is imported.
                    self.log("Filen har flera sektioner med metadata.\nBara de termposter som står före andra sektionen med metadata importeras.")
                    return

                # TODO: self.metadata is only used by old code. Remove self.metadata when old code is removed.
                self.metadata = SourceFileMetadata(self, segment.rows)
                self.term_post_source = TermPostSourceOld.get_term_post_source_from_rows(self, segment.rows, self.filename)
                if (not self.term_post_source.has_id()):
                    return
            else:
                self.term_posts.append(SourceFileTermPost(self, segment.rows))

    def see_also_convert(self, term_post, see_also, entries_lookup):
        term = see_also["term"]
        lang = see_also["lang"]
        if lang == "sv":
            return see_also
        homonym = see_also.get("homonym", 0)
        found_terms = entries_lookup.get(lang, {}).get((term, homonym), [])
        log_term_post = False
        if not found_terms:
            self.log("Hittade inte exakt refererad termpost:", term, lang, homonym)
            log_term_post = True
            term_post.is_ok = False
        term_entry = self.best_match_term(found_terms, term)
        if term_entry is None:
            self.log("Även luddigare sökning efter refererad termpost misslyckades:", term, lang, homonym)
            term_post.is_ok = False
            term_post.log_term_post()
            return None
        if (log_term_post):
            term_post.log_term_post()
        terms = term_entry["terms"]
        see_also = {"term": terms[0]["term"], "lang": terms[0]["lang"]}
        if "homonym" in terms[0]:
            see_also["homonym"] = terms[0]["homonym"]

        return see_also
