from model.import_util import string_substitutions


class SourceFileMetadata():
    """ Represents the metadata part of the source file.
    """

    # Name of fields in source file.
    _file_field_name = 'filename'
    _place_field_name = 'Ställe'
    _publish_year_field_name = 'Utgivningsår'
    _publisher_field_name = 'Utgivare'
    _source_id_field_name = 'Källid'
    _source_status_field_name = 'Status'
    _title_field_name = 'Titel_kortform'

    _no_source_id = -1

    def __init__(self, source_file, rows):
        entries_raw = {}
        for row in rows:
            (key, _, value) = row.partition(" ")
            entries_raw[key.rstrip("\r\n").lstrip("\ufeff")] = value.rstrip("\r\n")
        assert (entries_raw)
        self.entries = self._get_metadata(source_file, entries_raw)

        # Get source id.
        if (SourceFileMetadata._source_id_field_name in self.entries):
            self.source_id = int(self.entries[SourceFileMetadata._source_id_field_name])
            del self.entries[SourceFileMetadata._source_id_field_name]
            assert ("id" not in self.entries)
            self.entries["id"] = self.source_id
        else:
            self.source_id = SourceFileMetadata._no_source_id
            SourceFileMetadata._no_source_id = SourceFileMetadata._no_source_id - 1
            self.source_status = 0
            self.entries["id"] = self.source_id
            self.entries[SourceFileMetadata._source_status_field_name.lower()] = 'inteFärdig'
            source_file.log("Id för källa finns inte med i metadata.\nTermposterna i filen har inte importerats.")

        # Get source status.
        if (self.has_source_id()):
            self.source_status = None
            if SourceFileMetadata._source_status_field_name in self.entries:
                status_text = self.entries[SourceFileMetadata._source_status_field_name]
                del self.entries[SourceFileMetadata._source_status_field_name]
                if status_text == "publicerad":
                    self.source_status = 2
                elif status_text in ["avpublicerad", "inaktuell"]:
                    self.source_status = 1
                    # source_file.log(f"Filen har status {status_text}.", )
                elif status_text in ["inteFärdig", "bearbetas"]:
                    self.source_status = 0
                    # source_file.log(f"Filen har status {status_text}.", )
                else:
                    source_file.log("Okänd status i metadata.\nFilen har inte importerats.", status_text)
            else:
                source_file.log("Status finns inte med i metadata.\nFilen har inte importerats.")
            if (self.source_status is not None):
                self.entries[SourceFileMetadata._source_status_field_name.lower()] = self.source_status

        self.entries[SourceFileMetadata._file_field_name] = source_file.basename

    def _get_metadata(self, source_file, rows):
        metapost = {}
        for k, v in rows.items():
            if not v:
                continue
            # print(k, v)
            if k == "metapost":
                continue
            if "." in k:
                components = k.split(".")
                if components[0] == "Utgivare" and len(components) == 4:
                    utgivare = metapost.setdefault("Utgivare", [])
                    self._get_publisher(source_file, utgivare, components, v)
                elif components[0] == "Klassifikationssystem" and len(components) == 3:
                    klassifikationssystem = metapost.setdefault("Klassifikationssystem", [])
                    self._get_classification(klassifikationssystem, components, v)
                else:
                    k = "_".join(components)
                    if k in ["Titel_kortform", "Titel_överordnad"]:
                        v = string_substitutions(v)
                    metapost[k] = v
                continue
            if k in ["Ställe", "Titel"]:
                v = string_substitutions(v)
            assert (k not in metapost)
            metapost[k] = v
        # print(metapost)
        return metapost

    def _get_classification(self, klassifikationssystem, components, value):
        index = int(components[1]) - 1
        while len(klassifikationssystem) < index + 1:
            klassifikationssystem.append({})
        klassifikationssystem[index][components[2]] = string_substitutions(value)

    def get_place(self):
        return self.entries.get(SourceFileMetadata._place_field_name, None)

    def get_publish_year(self):
        return self.entries.get(SourceFileMetadata._publish_year_field_name, None)

    def get_publisher(self):
        return self.entries.get(SourceFileMetadata._publisher_field_name, None)

    def _get_publisher(self, source_file, utgivare, components, value):
        try:
            index = int(components[1]) - 1
        except ValueError:
            # Error is logged in new code.
            # source_file.log("Numreringen av utgivare är inte ett heltal: ", components)
            return
        publisher_type = components[2]
        while len(utgivare) < index + 1:
            utgivare.append({})
        utgivare[index]["type"] = publisher_type
        utgivare[index][components[3]] = value

    def get_title(self):
        return self.entries.get(SourceFileMetadata._title_field_name, None)

    def has_source_id(self):
        return ((self.source_id is not None) and (self.source_id > -1))

    def has_source_status(self):
        return (self.source_status is not None)

    def remove_parsing_data(self):
        """ Remove metadata that should not be stored in database.
        """

        if "Tolka_OKGR-märkning" in self.entries:
            del self.entries["Tolka_OKGR-märkning"]
        if "Klassifikationssystem" in self.entries:
            del self.entries["Klassifikationssystem"]
        if "illustrations" in self.entries:
            del self.entries["illustrations"]
        for key in list(self.entries.keys()):
            if (key.startswith("Tolka_skillnadTermSynonym_")):
                del self.entries[key]

