from django.conf import settings
from ISOF.common.list_functions import is_empty_list, is_not_empty_list
from model.import_util import textconvert_annotated_term
from model.source_file_term_post_field import term_language_priority, SourceFileTermPostField, SourceFileTermPostFieldDf
from model.source_file_term_post_old import SourceFileTermPostOld


slug_replacements = {
    " ": "_",
    "\u00a0": "_",
    "'": "-",
    "’": "-",
    '"': "",
    '”': "",
    "%": "procent",
    "&": "-ampersand-",
    "/": "-",
    ":": "-",
    "§": "-paragraf-",
    "–": "-",
    "—": "-",
    "⁰": "0",
    "¹": "1",
    "²": "2",
    "³": "3",
    "⁴": "4",
    "⁵": "5",
    "⁶": "6",
    "⁷": "7",
    "⁸": "8",
    "⁹": "9",
    "⁺": "+",
    "⁻": "-",
    "⁼": "=",
    "⁽": "(",
    "⁾": ")",
    "ⁿ": "n",
    "₀": "0",
    "₁": "1",
    "₂": "2",
    "₃": "3",
	"₄": "4",
    "₅": "5",
    "₆": "6",
    "₇": "7",
    "₈": "8",
    "₉": "9",
	"₊": "+",
	"₋": "-",
	"₌": "=",
	"₍": "(",
	"₎": ")",
    "ₐ": "a",
	"ₑ": "e",
	"ₒ": "o",
    "ₓ": "x",
}


class TermPostLanguageException(Exception):
    """ This exception is used if a source file contains a language code that Rikstermbanken does not support.
    """
    pass


class SourceFileTermPost():
    """ Holds information about one term post in a file in "Bankvalvet".
    """

    def __init__(self, source_file, rows):
        self.document = {}
        self.is_ok = True
        self.fields = []
        self.rows = rows
        self.source_file = source_file

        self._get_fields()
        self._parse_fields()
        self._get_field_information()
        self._sort_term_post()
        self._get_slugs()
        # self.check_has_swedish_term()
        self._check_has_term()
        if (self.source_file.term_post_source.is_recently_published()):
            self._check_single_fields()
            self._check_has_max_one_swedish_term()
        self._check_is_ok()
        self._compare()


    def _check_has_max_one_swedish_term(self):
        """ Check that term post has max one swedish term field.
        """

        swedish_term_count = 0
        for field in self.fields:
            if ((field.field_name == 'TE') and (field.language == 'sv')):
                swedish_term_count = swedish_term_count + 1
        if (swedish_term_count > 1):
            # Too many swedish terms found. Log error.
            self.source_file.log('Termposten har för många svenska termer.')
            self.is_ok = False

    def _check_has_swedish_term(self):
        """ Check that term post has a swedish term field.

            This check is currently not used. Decision by Karin Webjörn.
        """

        has_swedish_term = False
        for field in self.fields:
            if ((field.field_name == 'TE') and (field.language == 'sv')):
                has_swedish_term = True
                break
        if (not has_swedish_term):
            # No swedish term found. Log error.
            self.source_file.log('Termposten saknar svensk term.')
            self.is_ok = False

    def _check_has_term(self):
        """ Check that at least one field is a term.
        """

        if (not (self.document and ('terms' in self.document) and is_not_empty_list(self.document['terms']))):
            # No term found. Log error.
            self.source_file.log('Termposten saknar term.')
            self.is_ok = False

    def _check_is_ok(self):
        """ Check if term post is ok.
        """

        if (self.is_ok):
            for field in self.fields:
                if (not field.is_ok):
                    self.is_ok = False

        # Log rows if term post is not ok.
        if (not self.is_ok):
            self.log_term_post()

    def _check_single_fields(self):
        """ Check that single fields has max one entry in term post.
        """

        handled_errors = []
        for field in self.fields:
            if (field.is_single_field):
                field_name = field.field_name
                language = field.language
                field_count = 0
                for temp_field in self.fields:
                    if (temp_field.field_name == field_name):
                        if (language is None):
                            field_count = field_count + 1
                        elif (language == temp_field.language):
                            field_count = field_count + 1

                if (language is None):
                    tag = field_name
                else:
                    tag = language + field_name
                if ((field_count > 1) and (tag not in handled_errors)):
                    self.source_file.log(f"Fält av typen {tag} kan bara förekomma en gång i en termpost.")
                    self.is_ok = False
                    handled_errors.append(tag)

    def _compare(self):
        """ Compare output from new and old code.
            Exception will be raised if output is not equal.

            This is done to avoid errors when changing code.
            Sometime in the future this code should probably be removed.
        """

        if (self.is_ok):
            term_post_old = SourceFileTermPostOld(self.source_file, self.rows)
            old_fields = term_post_old.old_fields
            new_fields = self.fields
            parsed_fields = term_post_old.parsed_fields
            assert(len(old_fields) == len(new_fields))
            assert (len(parsed_fields) == len(new_fields))
            field_index = 0
            while (field_index < len(new_fields)):
                # Compare not parsed data.
                field = new_fields[field_index]
                subfields_output = []
                for subfield in field.subfields:
                    subfields_output.append((subfield.field_name, subfield.value))
                new_field = (field.key, field.value, field.field_name, subfields_output, field.language)

                old_field = old_fields[field_index]
                assert (old_field[0] == new_field[0])
                assert (old_field[1] == new_field[1])
                assert (old_field[2] == new_field[2])
                assert (len(old_field[3]) == len(new_field[3]))
                list_index = 0
                old_subordinates = old_field[3]
                new_subordinates = new_field[3]
                while (list_index < len(new_field[3])):
                    old_subordinate = old_subordinates[list_index]
                    new_subordinate = new_subordinates[list_index]
                    list_index = list_index + 1
                    assert (len(old_subordinate) == len(new_subordinate))
                    tuple_index = 0
                    while (tuple_index < len(new_subordinate)):
                        old_tuple_value = old_subordinate[tuple_index]
                        new_tuple_value = new_subordinate[tuple_index]
                        assert (old_tuple_value == new_tuple_value)
                        tuple_index = tuple_index + 1
                assert (old_field[4] == new_field[4])

                # Compare parsed data.
                new_parsed_data = new_fields[field_index].document
                old_parsed_data = parsed_fields[field_index]
                assert (len(new_parsed_data) == len(old_parsed_data))

                parsed_index = 0
                while (parsed_index < len(new_parsed_data)):
                    new_parsed_field = new_parsed_data[parsed_index]
                    old_parsed_field = old_parsed_data[parsed_index]
                    assert (len(new_parsed_field) == len(old_parsed_field))

                    # Match name for parsed data.
                    assert (new_parsed_field[0] == old_parsed_field[0])

                    # Match value for parsed data.
                    new_value = new_parsed_field[1]
                    old_value = old_parsed_field[1]
                    assert (type(new_value) == type(old_value))
                    if (type(new_value) is int):
                        assert (new_value == old_value)
                    elif (type(new_value) is str):
                        assert (new_value == old_value)
                    elif (type(new_value) is dict):
                        if ((len(new_value) != len(old_value)) and (field.field_name == "ILLU")):
                            # Wrong data handling in old code.
                            parsed_index = parsed_index + 1
                            continue
                        assert (len(new_value) == len(old_value))
                        for key in new_value.keys():
                            assert (key in old_value)
                            assert (new_value[key] == old_value[key])
                    elif (type(new_value) is tuple):
                        assert (len(new_value) == len(old_value))
                        assert (len(new_value) == len(old_value))
                        tuple_index = 0
                        while (tuple_index < len(new_value)):
                            new_tuple_value = new_value[tuple_index]
                            old_tuple_value = old_value[tuple_index]
                            assert (new_tuple_value == old_tuple_value)
                            tuple_index = tuple_index + 1

                    parsed_index = parsed_index + 1

                field_index = field_index + 1

    def get_definitions(self) -> list:
        """ Get all swedish definitions in term post.

            Returns: list: All swedish definitions in term post.
        """

        definitions = []
        for field in self.fields:
            if ((field.field_name == SourceFileTermPostFieldDf.get_field_name()) and
                (field.language == settings.SWEDISH_LANGUAGE_CODE)):
                definitions.append(field.value)
        return definitions

    def _get_field_information(self):
        """ Aggregate term post field information in term post.
        """

        for field in self.fields:
            if (field.is_ok):
                for value_key, value in field.document:
                    if (value_key in self.document):
                        value_list = self.document[value_key]
                    else:
                        value_list = []
                        self.document[value_key] = value_list
                    value_list.append(value)

    def _get_fields(self):
        """ Convert text rows to source file term post fields.
        """

        for row in self.rows:
            field = SourceFileTermPostField.get_field(self, row)
            if (field is not None):
                if (field.is_ok):
                    if (field.is_subfield()):
                        if (is_empty_list(self.fields)):
                            self.source_file.log("Kopplat fält måste ha toppnivå- eller språkfält: ",
                                                 field.field_name,
                                                 field.value)
                            self.is_ok = False
                        else:
                            previous_field = self.fields[-1]
                            previous_field.add_subfield(field)
                    else:
                        self.fields.append(field)
                else:
                    self.is_ok = False

    def _get_slugs(self):
        if (self.is_ok):
            terms = self.document['terms']
            if (terms):
                language_names = [term["lang"] for term in terms]
                for termtype in ["TE", "BT"]:
                    for language_name in ["sv", "en", "fi", "la"]:
                        if language_name in language_names:
                            langterms = [term for term in terms if term["status"] == termtype and term["lang"] == language_name and term.get("term", "") not in ["-", "–"]]
                            if not langterms:
                                continue
                            self.document["slugs"] = [self._get_term_term(term) for term in langterms]
                            return

                self.source_file.log("Hittade inte någon slug.")
                self.is_ok = False

    def _get_term_term(self, term):
        if "term" in term:
            s = term["term"]
        elif "annotated_term" in term:
            s = textconvert_annotated_term(term["annotated_term"])
        else:
            assert False, term
        if "\u2029" in s:
            s = s[:s.index("\u2029")]
        for old, new in slug_replacements.items():
            s = s.replace(old, new)
        if "homonym" in term:
            s += "-%d" % (term["homonym"],)
        if term["lang"] != "sv":
            s = term["lang"] + "--" + s
        return s

    def has_definitions(self):
        """ Get information about existence of Swedish definitions (field DF).

            Returns: bool: True if term post has at least one Swedish definition field.
        """

        for field in self.fields:
            if ((field.field_name == SourceFileTermPostFieldDf.get_field_name()) and
                (field.language == settings.SWEDISH_LANGUAGE_CODE)):
                return True
        return False

    def log_term_post(self):
        """ Log rows in term post to the term post source error log.
        """

        self.source_file.log('Termpost:')
        for row in self.rows:
            self.source_file.log(row)
        self.source_file.log('\t')

    def _parse_fields(self):
        """ Parse fields.

            All fields must be created first so that subfields are added to top or language fields.
        """

        for field in self.fields:
            if (field.is_ok):
                field.parse_field()

    @staticmethod
    def sort_terms_key(term):
        """ Get sort key for terms.
        """

        term_status_priority = {
            "TE": 0,
            "SYTE": 1,
            "AVTE": 2,
        }

        language = term["lang"]
        if (language not in term_language_priority):
            raise TermPostLanguageException(f"Språkkod som Rikstermbanken inte har stöd för: {language}")
        priority = term_language_priority.get(language, 100)

        return (
            term_status_priority.get(term["status"], 100),
            priority,
            "term" in term,
            term.get("term", ""),
        )

    def _sort_term_post(self):
        """ Sort term post information parts.

            Duplicate entries are removed from "searchterms" and "fulltextsearch".
        """

        try:
            if ("terms" in self.document):
                self.document["terms"] = sorted(self.document["terms"], key=SourceFileTermPost.sort_terms_key)

            if ("searchterms" in self.document):
                self.document["searchterms"] = sorted(set(self.document["searchterms"]))

            if ("fulltextsearch" in self.document):
                if ("searchterms" in self.document):
                    self.document["fulltextsearch"] = sorted(set(self.document["searchterms"]) |
                                                          set(self.document["fulltextsearch"]))
                else:
                    self.document["fulltextsearch"] = sorted(set(self.document["fulltextsearch"]))

        except TermPostLanguageException as language_exception:
            self.source_file.log(language_exception.args[0])
            self.is_ok = False
