import re
from ISOF.common.string_functions import is_integer_string
from model.import_util import string_substitutions, decode_entities

paragraph_separator = '\u2029'

term_language_priority = {
    "sv": 0,
    "en": 1,
    "de": 2,
    "fr": 3,
    "es": 4,
    "ca": 5,
    "rom": 6,
    "la": 7,
    "da": 8,
    "fo": 9,
    "is": 10,
    "nb": 11,
    "nn": 12,
    "no": 13,
    "hr": 14,
    "pl": 15,
    "ru": 16,
    "gr": 17,
    "fi": 18,
    "fit": 19,
    "se": 20,
    "et": 21,
    "kl": 22,
    "tr": 23,
    "ku": 24,
    "ar": 25,
    "so": 26,
    "ja": 27,
    "nl": 28,
    "it": 29,
    "re": 30,
    "pt": 32,
    "il": 33,
    #    "sy": 34,  # Not used.
    "sq": 35,
    #"rm": 36,
    #"rk": 37,
    #"ra": 38,
    "rmn": 39,
    "rmy": 40,
}

top_level_field_names = [
    "FO",
    "ILLU",
    "INAN",
    "INID",
    "KL",
    "ORG",
    "RIAN",
    "RIFR",
]

language_field_names = [
    "AN",
    "AVTE",
    "BT",
    "DF",
    "EKVI",
    "EX",
    "FK",
    "INNA",
    "INTE",
    "KT",
    "NA",
    "PH",
    "RETE",
    "SA",
    "SPBR",
    "SU",
    "SYPH",
    "SYTE",
    "TE",
    "UPTE",
    "TI",
]

sub_level_field_names = [
    "BNGR",
    "ETYM",
    "FRKT",
    "GE",
    "GR",
    "GNGR",
    "HONR",
    "ILLT",
    "ILTY",
    "OKGR",
    "RF",
    "SA",
    "UT",
]


class SourceFileTermPostField():
    """ Holds information about one term post field in a file in "Bankvalvet".
    """

    # Regular expression matching.
    angle_bracket_tag_valid_re = re.compile(r"^[A-Z]")
    reference_with_homonym_number_re = re.compile(r"^(.+) \((\d+)\)$")
    split_row_re = re.compile(r"^(\S+)\s+(.*)$")

    # field_counter = 0

    def __init__(self, row, source_file_term_post):
        self.accepts_embedded_fields = None
        self.accepts_styled_text = None
        self.document = None
        self.field_value_parts = None
        self.has_embedded_fields = None
        self.information_type = None
        self.is_ok = True
        self.is_single_field = False
        self.log = source_file_term_post.source_file.log
        self.possible_subfields = None
        self.reference_type = None
        self.row = row
        self.source_file = source_file_term_post.source_file
        self.source_file_term_post = source_file_term_post
        self.subfield_names = []
        self.subfields = []
        self.term_post_source = source_file_term_post.source_file.term_post_source

        # Split row into parts.
        (key, field_name, language, value, is_ok) = SourceFileTermPostField.get_field_parts(row, self.log)
        self.field_name = field_name
        self.is_ok = is_ok
        self.key = key
        self.language = language
        self.value = value

        if (self.language):
            if (self.language not in term_language_priority):
                # Not handled language code.
                self.log(f"Språkkod som Rikstermbanken inte har stöd för: {self.language}.")
                self.is_ok = False

            if (self.field_name in top_level_field_names):
                # Top level fields does not have a language code.
                self.log(f"Toppnivåfält {self.field_name}  ska inte ha en språkkod: {self.language}.")
                self.is_ok = False

            if ((self.field_name in sub_level_field_names) and
                (self.field_name not in language_field_names) and
                (self.field_name != 'ETYM')):                           # Do not report error on language code if field name 'ETYM'
                # Sub level fields does not have a language code.
                self.log(f"Kopplat fält {self.field_name} ska inte ha en språkkod: {self.language}.")
                self.is_ok = False

        else:  # No language code
            if ((self.field_name in language_field_names) and
                (self.field_name not in sub_level_field_names)):
                # Language fields must have a language code.
                self.log(f"Språkfält {self.field_name} måste ha en språkkod.")
                self.is_ok = False

        if ((self.field_name in top_level_field_names) or (self.field_name in sub_level_field_names)):
            # Only some language fields has embedded fields.
            self.accepts_embedded_fields = False
        self.accepts_styled_text = (self.field_name in language_field_names)

    def add_embedded_text_tree(self, value_row, tree, start_position, end_position):
        if (tree and (tree[-1][0] == "text")):
            tree[-1] = ("text", tree[-1][1] + value_row[start_position:end_position])
        else:
            tree.append(("text", value_row[start_position:end_position]))

    def add_subfield(self, subfield):
        # Check if it is ok to add subfield to this field.
        if ((self.possible_subfields is not None) and (subfield.field_name in self.possible_subfields)):
            if ((subfield.is_single_field) and (subfield.field_name in self.subfield_names)):
                self.log(f"Kopplat fält av typen {subfield.field_name} får bara förekomma en gång i samma fält.")
                self.is_ok = False
            else:
                # It's ok. Add subfield to this field.
                self.subfield_names.append(subfield.field_name)
                self.subfields.append(subfield)
        else:
            if (self.language is None):
                self.log(f"Toppnivåfält av typen {self.field_name} kan inte ha kopplat fält av typen {subfield.field_name}.")
            else:
                self.log(f"Språkfält av typen {self.field_name} kan inte ha kopplat fält av typen {subfield.field_name}.")
            self.is_ok = False

    def check_tag_position_tree(self, value_row, in_tag_position_tree):
        """ Remove tag positions that can not be interpreted as HTML style tags.

        Returns:
            Tag position tree: Cleaned tag position tree.
        """

        begin_tags = []
        tag_position_tree = []
        for start_position, end_position, subtrees in in_tag_position_tree:
            assert (not subtrees)
            tag_name = value_row[start_position + 1:end_position]
            if (len(tag_name) == 0):
                continue
            elif (tag_name[0] == "/"):
                if (not begin_tags):
                    continue
                while begin_tags:
                    last_tag = begin_tags.pop()
                    if (last_tag[0] == tag_name[1:]):
                        tag_position_tree.append(last_tag[1])
                        tag_position_tree.append((start_position, end_position, subtrees))
                        break
            else:
                begin_tags.append((tag_name, (start_position, end_position, subtrees)))
        return sorted(tag_position_tree)

    def embedded_convert_annotated_term(self, annotated_term):
        """ This method is currently not used, but it may be used in the future if data or demands changes.

            Transform field with embedded tags to original value but with structured annotations.
            Currently, no fields has both embedded tags and structured annotations.
        """

        s = ""
        for e in annotated_term:
            if "text" in e:
                s += e["text"]
            elif "tag" in e:
                s += "<" + e["tag"] + " " + self.embedded_convert_annotated_term(e["content"]) + ">"
            else:
                assert (False)
        return s

    def get_all_positions(self, row, tag):
        """ Get all positions for string "tag" in string "row".

        Arguments:
            row (str): Row.
            tag (str): Tag.

        Returns:
            list[int]: All positions for string "tag" in string "row".
        """

        positions = []
        position = 0
        while True:
            position = row.find(tag, position)
            if (position == -1):
                break
            positions.append(position)
            position += 1
        return positions

    def get_embedded_styled_field(self, in_tree):
        """ Get embedded field where styled text parts are handled.
        """

        tree = []
        for e in in_tree:
            if e[0] == "tag":
                tree.append({"tag": e[1], "content": self.get_embedded_styled_field(e[2])})
            elif e[0] == "text":
                tree.extend(self.get_styled_text(decode_entities(e[1]), "{", "}"))
            else:
                # Unknown type of parsed subfield.
                assert (False)
        return tree

    def get_embedded_tree(self, value_row, in_tree, not_embedded_start_position=0, tree_end_position=None):
        tree = []
        if (tree_end_position is None):
            tree_end_position = len(value_row)
        for start_position, end_position, subtree in in_tree:
            value_start = value_row.find(" ", start_position + 1, end_position)
            if (value_start >= 0):
                key = value_row[start_position + 1:value_start]
                if (not SourceFileTermPostField.angle_bracket_tag_valid_re.match(key)):
                    self.add_embedded_text_tree(value_row, tree, not_embedded_start_position, end_position + 1)
                    not_embedded_start_position = end_position + 1
                    continue
                if (not_embedded_start_position != start_position):
                    self.add_embedded_text_tree(value_row, tree, not_embedded_start_position, start_position)
                tree.append(("tag", key, self.get_embedded_tree(value_row, subtree, value_start + 1, end_position)))
                not_embedded_start_position = end_position + 1
            else:
                self.add_embedded_text_tree(value_row, tree, not_embedded_start_position, end_position + 1)
                not_embedded_start_position = end_position + 1
        if (not_embedded_start_position != tree_end_position):
            self.add_embedded_text_tree(value_row, tree, not_embedded_start_position, tree_end_position)
        return tree

    @staticmethod
    def get_field(source_file_term_post, row: str):
        """ Get SourceFileTermPostField instance that should handle this row.

        Arguments:
            source_file_term_post:  The term post that term post field belongs to.
            row:        One row from a source file.

        Returns:
            SourceFileTermPostField: A SourceFileTermPostField instance.
        """

        field_name, language = SourceFileTermPostField.get_field_name(row, source_file_term_post.source_file.log)
        if (field_name == 'AN'):
            field = SourceFileTermPostFieldAn(row, source_file_term_post)
        elif (field_name == 'AVTE'):
            field = SourceFileTermPostFieldAvte(row, source_file_term_post)
        elif (field_name == 'BNGR'):
            field = SourceFileTermPostFieldBngr(row, source_file_term_post)
        elif (field_name == 'BT'):
            field = SourceFileTermPostFieldBt(row, source_file_term_post)
        elif (field_name == 'DF'):
            field = SourceFileTermPostFieldDf(row, source_file_term_post)
        elif (field_name == 'EKVI'):
            field = SourceFileTermPostFieldEkvi(row, source_file_term_post)
        elif (field_name == 'ETYM'):
            field = SourceFileTermPostFieldEtym(row, source_file_term_post)
        elif (field_name == 'EX'):
            field = SourceFileTermPostFieldEx(row, source_file_term_post)
        elif (field_name == 'FK'):
            field = SourceFileTermPostFieldFk(row, source_file_term_post)
        elif (field_name == 'FO'):
            field = SourceFileTermPostFieldFo(row, source_file_term_post)
        elif (field_name == 'FRKT'):
            field = SourceFileTermPostFieldFrkt(row, source_file_term_post)
        elif (field_name == 'GE'):
            field = SourceFileTermPostFieldGe(row, source_file_term_post)
        elif (field_name == 'GNGR'):
            field = SourceFileTermPostFieldGngr(row, source_file_term_post)
        elif (field_name == 'GR'):
            field = SourceFileTermPostFieldGr(row, source_file_term_post)
        elif (field_name == 'HONR'):
            field = SourceFileTermPostFieldHonr(row, source_file_term_post)
        elif (field_name == 'ILLT'):
            field = SourceFileTermPostFieldIllt(row, source_file_term_post)
        elif (field_name == 'ILLU'):
            field = SourceFileTermPostFieldIllu(row, source_file_term_post)
        elif (field_name == 'ILTY'):
            field = SourceFileTermPostFieldIlty(row, source_file_term_post)
        elif (field_name == 'INAN'):
            field = SourceFileTermPostFieldInan(row, source_file_term_post)
        elif (field_name == 'INID'):
            field = SourceFileTermPostFieldInid(row, source_file_term_post)
        elif (field_name == 'INNA'):
            field = SourceFileTermPostFieldInna(row, source_file_term_post)
        elif (field_name == 'INTE'):
            field = SourceFileTermPostFieldInte(row, source_file_term_post)
        elif (field_name == 'KL'):
            field = SourceFileTermPostFieldKl(row, source_file_term_post)
        elif (field_name == 'ORG'):
            field = SourceFileTermPostFieldOrg(row, source_file_term_post)
        elif (field_name == 'KT'):
            field = SourceFileTermPostFieldKt(row, source_file_term_post)
        elif (field_name == 'NA'):
            field = SourceFileTermPostFieldNa(row, source_file_term_post)
        elif (field_name == 'OKGR'):
            field = SourceFileTermPostFieldOkgr(row, source_file_term_post)
        elif (field_name == 'PH'):
            field = SourceFileTermPostFieldPh(row, source_file_term_post)
        elif (field_name == 'RETE'):
            field = SourceFileTermPostFieldRete(row, source_file_term_post)
        elif (field_name == 'RF'):
            field = SourceFileTermPostFieldRf(row, source_file_term_post)
        elif (field_name == 'RIAN'):
            field = SourceFileTermPostFieldRian(row, source_file_term_post)
        elif (field_name == 'RIFR'):
            field = SourceFileTermPostFieldRifr(row, source_file_term_post)
        elif (field_name == 'SA'):
            if (language):
                field = SourceFileTermPostFieldSaLanguageField(row, source_file_term_post)
            else:
                field = SourceFileTermPostFieldSaSubField(row, source_file_term_post)
        elif (field_name == 'SPBR'):
            field = SourceFileTermPostFieldSpbr(row, source_file_term_post)
        elif (field_name == 'SU'):
            field = SourceFileTermPostFieldSu(row, source_file_term_post)
        elif (field_name == 'SYPH'):
            field = SourceFileTermPostFieldSyph(row, source_file_term_post)
        elif (field_name == 'SYTE'):
            field = SourceFileTermPostFieldSyte(row, source_file_term_post)
        elif (field_name == 'TE'):
            field = SourceFileTermPostFieldTe(row, source_file_term_post)
        elif (field_name == 'TI'):
            field = SourceFileTermPostFieldTi(row, source_file_term_post)
        elif (field_name == 'UPTE'):
            field = SourceFileTermPostFieldUpte(row, source_file_term_post)
        elif (field_name == 'UT'):
            field = SourceFileTermPostFieldUt(row, source_file_term_post)
        else:
            source_file_term_post.source_file.log("Ogiltigt fältnamn: ", field_name)
            field = None
            source_file_term_post.is_ok = False

        return field

    @staticmethod
    def get_field_name(row, log):
        """ Get field name from row.

        Arguments:
            row (str): Field row.
            log: Source file log.

        Returns:
            tuple(str, str): Field name and language or None if field name or language was not found.
        """

        (key, field_name, language, value, is_ok) = SourceFileTermPostField.get_field_parts(row, log, False)
        return (field_name, language)

    def get_field_output(self):
        """ Get output from field in same format as old code uses.

        Returns:
            tuple: Tuple with field output.
        """

        subfields_output = []
        for subfield in self.subfields:
            subfields_output.append((subfield.field_name, subfield.value))
        return (self.key, self.value, self.field_name, subfields_output, self.language)

    @staticmethod
    def get_field_parts(row, log, report_missing_value=True):
        """ Split row into fields parts.

            Last value in returned tuple is a boolean that indicates if operation was successful or not.

        Arguments:
            row (str): Field row.
            log: Source file log.
            report_missing_value (bool): Used to avoid reporting the same error more than once.

        Returns:
            tuple: (key, field_name, language, value, is_ok)
        """

        field_name = None
        is_ok = True
        language = None
        key = None
        value = None

        # Split row into key and value.
        reg_match = SourceFileTermPostField.split_row_re.match(row)
        if (reg_match):
            key = reg_match.group(1)
            value = reg_match.group(2)
        else:
            key = row
            if (report_missing_value):
                log("Fältet saknar värde: ", repr(key))
            is_ok = False

        # Get field name and lang (if applicable).
        if (key[0:2].islower() and key[2:].isupper()):
            # Two character language code.
            field_name = key[2:]
            language = key[0:2]
        elif (key[0:3].islower() and key[3:].isupper()):
            # Three character language code.
            field_name = key[3:]
            language = key[0:3]
        elif (key.isupper()):
            field_name = key
        else:
            log("Ogiltigt fältnamn: ", repr(key))
            is_ok = False

        return (key, field_name, language, value, is_ok)

    def get_field_value_parts(self, value_row):
        tag_position_tree = self.get_tag_position_tree(value_row, "<", ">")
        return self.get_embedded_tree(value_row, tag_position_tree)

    def get_rcon_field(self, embedded_field):
        """ Transform embedded RCON field to term reference.
        """

        field = []
        for embedded_field_row in embedded_field:
            if ('tag' in embedded_field_row):
                if (embedded_field_row['tag'] == "RCON"):
                    embedded_term_reference = embedded_field_row["content"]
                    assert (embedded_term_reference[0].keys() == set(["text"]))
                    term_reference = {"term": embedded_term_reference[0]["text"].rstrip(" ")}
                    if (len(embedded_term_reference) == 1):
                        return term_reference
                    assert (len(embedded_term_reference) == 2 and embedded_term_reference[1].get("tag") == "HONR")
                    assert (len(embedded_term_reference[1]["content"]) == 1)
                    assert (embedded_term_reference[1]["content"][0].keys() == set(["text"]))
                    term_reference["homonym"] = embedded_term_reference[1]["content"][0]["text"]
                    field.append({"term": term_reference})
                else:
                    rcon_field = self.get_rcon_field(embedded_field_row["content"])
                    field.append({"tag": embedded_field_row["tag"], "content": rcon_field})
            else:
                field.append(embedded_field_row)
        return field

    def get_styled_text(self, value_row, tag_start, tag_end):
        tag_position_tree = self.get_tag_position_tree(value_row, tag_start, tag_end)
        if any([subtree for start_position, end_position, subtree in tag_position_tree]):
            self.log("En HTML tag kan inte innehålla HTML taggar: ", value_row)
            self.is_ok = False
            return None

        tag_position_tree = self.check_tag_position_tree(value_row, tag_position_tree)
        return self.get_styled_tree(value_row, tag_position_tree)

    def get_styled_tree(self, value_row, tag_position_tree, content_start=0):
        """ Get a structured representation of the text together with information about styled parts.

            Currently (2023-03-06) the following HTML tags are used: <b>, <i>, <sub> and <sup>.
        """

        tree = []
        while tag_position_tree:
            start_position, end_position, subtrees = tag_position_tree.pop(0)
            assert (not subtrees)
            tag_name = value_row[start_position + 1:end_position]
            if (len(tag_name) == 0):
                # self.log("subtree returns, tag_name empty")
                return (tree + [{"text": "{}"}], None, None, None, content_start)
            elif (tag_name[0] == "/"):
                return (tree, tag_name[1:], start_position, end_position, content_start)
            else:
                styled_tree = self.get_styled_tree(value_row, tag_position_tree, end_position + 1)
                subtree, endtag_name, endtag_start_position, endtag_end_position, inner_tag_end = styled_tree
                if (subtree is None):
                    # self.log("error: subtree returns, None inside: " + value_row)
                    return (tree, None, None, None, None)
                if (tag_name != endtag_name):
                    # self.log("error: subtree returns, tag_name != endtag_name", tag_name, endtag_name)
                    return (tree, None, None, None, None)
                assert (tag_name == endtag_name)
                if (value_row[content_start:start_position]):
                    tree.append({"text": decode_entities(value_row[content_start:start_position])})
                if (value_row[inner_tag_end:endtag_start_position]):
                    subtree.append({"text": decode_entities(value_row[inner_tag_end:endtag_start_position])})
                tree.append({"tag": tag_name, "content": subtree})
                content_start = endtag_end_position + 1

        # This is where recursion ends and tree processing is finished.
        if (tree is None):
            # self.log("error: subtree None: " + value_row)
            return None
        if (tag_position_tree):
            tree.append({"text": "tags:" + repr(tag_position_tree)})
        if (value_row[content_start:]):
            tree = tree + [{"text": decode_entities(value_row[content_start:])}]
        return tree

    def get_subfield_value(self, parent_field):
        """ Get value from subfield and store it in parent field.

            Parent field is expected to contain a property "field"
            that is a dictionary where the subfield values are stored.
            Implementation is in subfield classes.

        Arguments:
            parent_field (SourceFileTermPostField): Parent field.
        """

        pass

    def get_subfields(self):
        """ Add subfield information to output from this field.
        """

        for subfield in self.subfields:
            assert (subfield.field_name in self.possible_subfields)
            subfield.get_subfield_value(self)

    def get_tag_position_subtree(self, positions):
        if (not positions):
            # We have reached end of recursion.
            return []

        start_position, is_start = positions[0]
        if (not is_start):
            # log("end symbol without start symbol")
            return (None, [])

        positions = positions[1:]
        subtrees = []
        if not positions:
            # log("illegal tag, no end symbol")
            return (None, [])

        while positions[0][1]:
            (subtree, positions) = self.get_tag_position_subtree(positions)
            if (not positions):
                # log("illegal tag, no end symbol")
                return (None, [])

            if (subtree is not None):
                subtrees.append(subtree)
        end_position, _ = positions[0]
        positions = positions[1:]
        return ((start_position, end_position, subtrees), positions)

    def get_tag_position_tree(self, value_row, tag_start, tag_end):
        # Get positions for start och end tags.
        start_tag_positions = [(position, True) for position in self.get_all_positions(value_row, tag_start)]
        end_tag_positions = [(position, False) for position in self.get_all_positions(value_row, tag_end)]
        positions = sorted(start_tag_positions + end_tag_positions)

        # Build tree with information about start och end tags.
        tree = []
        while positions:
            (subtree, positions) = self.get_tag_position_subtree(positions)
            if (subtree is not None):
                tree.append(subtree)
        return tree

    def get_text_from_styled_field(self, styled_field):
        """ Get text part of the styled field.

            Returns: str: The text part of the styled field.
        """

        text = ""
        for field_row in styled_field:
            if ("text" in field_row):
                text += field_row["text"]
            elif ("tag" in field_row):
                text += self.get_text_from_styled_field(field_row["content"])
            elif ("term" in field_row):
                text += field_row["term"]["term"]
            else:
                # Unknown type of entry in styled field.
                assert(False)
        return text

    def get_text_from_term(self, annotated_term):
        text = ""
        for field_row in annotated_term:
            if ("text" in field_row):
                text += field_row["text"]
            elif ("tag" in field_row):
                text += self.get_text_from_term(field_row["content"])
            else:
                # Unknown type of entry in term field.
                assert (False)
        return text

    def get_words(self, text):
        """ Split text into words.

            Returns: list: Words in text.
        """
        return [word.strip() for word in text.split(" ") if word]

    def has_only_honr_tags(self, trees):
        """ Test if all tags in trees are of type HONR. This check is not recursive only top level is checked.

            Returns: bool: True if all tags are of type HONR.
        """

        for tree in trees:
            if (("tag" in tree) and (tree["tag"] != "HONR")):
                return False
        return True

    def is_subfield(self):
        # 'SA' is both a top level field and a subfield.
        # Language is not specified in the subfield 'SA'.
        return ((self.field_name in sub_level_field_names) and
                (not ((self.field_name == 'SA') and (self.language is not None))))

    def parse_field(self):
        # Get information about subfields embedded in this fields value.
        value_rows = self.value.split(paragraph_separator)
        self.field_value_parts = [self.get_field_value_parts(value_row) for value_row in value_rows]
        self.has_embedded_fields = not (all([len(field_value_part) == 1 and field_value_part[0][0] == "text"
                                        for field_value_part in self.field_value_parts]))

        # Check if this field type accepts embedded fields.
        assert (self.accepts_embedded_fields is not None)
        if ((not self.accepts_embedded_fields) and self.has_embedded_fields):
            # This field has embedded fields that it should not have.
            self.log(f'Fält av typen {self.field_name} kan inte ha inbäddade fält.')
            self.is_ok = False

    def parse_field_information(self):
        assert (self.field_name in ["DF", "FK", "AN", "EX", "SA", "EKVI", "SPBR"])
        text_rows = []

        # Handle embedded tags and style tags in information.
        for field_value_part in self.field_value_parts:
            value_row = string_substitutions(field_value_part[0][1])
            if (self.has_embedded_fields):
                if ((len(field_value_part) == 1) and (field_value_part[0][0] == "text")):
                    text_rows.append({"text": value_row})
                else:
                    embedded_styled_field = self.get_embedded_styled_field(field_value_part)
                    if (self.has_only_honr_tags(embedded_styled_field)):
                        # text_rows.append({"text": self.embedded_convert_annotated_term(embedded_styled_field)})
                        text_rows.append({"text": self.value})
                    else:
                        text_rows.append({"annotated_text": self.get_rcon_field(embedded_styled_field)})
            else:
                tree = self.get_styled_text(value_row, "{", "}")
                if (tree is None):
                    text_rows.append({"text": value_row})
                elif ((len(tree) == 1) and ("text" in tree[0])):
                    text_rows.append(tree[0])
                else:
                    text_rows.append({"annotated_text": tree})

        # Get full text search words.
        extra = []
        for text_row in text_rows:
            text = None
            if ("annotated_text" in text_row):
                text = self.get_text_from_styled_field(text_row["annotated_text"])
            if ("text" in text_row):
                text = text_row["text"]
            if (text):
                extra.extend([("fulltextsearch", word) for word in self.get_words(text)])

        self.field = {"paragraphs": text_rows}
        self.field["lang"] = self.language
        assert (self.information_type is not None)
        self.field['type'] = self.information_type
        self.get_subfields()
        self.document = [("texts", self.field)] + extra

    def parse_field_reference(self):
        assert (self.field_name in ["RETE", "SU"])
        assert (len(self.field_value_parts) == 1)

        # Get reference structure.
        field_parts = self.field_value_parts[0]
        separator = ','
        reference_parts = []
        for field_part in field_parts:
            if (field_part[0] == "text"):
                value = field_part[1]
                value_parts = value.split(separator)
                # positions = get_all_positions(s, separator)
                # for position in positions:
                # print("parts", value_parts)
                if (reference_parts):
                    previous_reference_part = reference_parts[-1]
                    reference_parts[-1] = (previous_reference_part[0] + value_parts[0], previous_reference_part[1])
                    value_parts = value_parts[1:]
                for value_part in value_parts:
                    if (value_part):
                        reference_parts.append((value_part, []))
            elif (field_part[0] == "tag"):
                embedded_field_name = field_part[1]
                embedded_field = field_part[2]
                assert (len(embedded_field) == 1)
                embedded_field_value = embedded_field[0]
                assert (embedded_field_value[0] == "text")
                if (reference_parts):
                    previous_reference_part = reference_parts[-1]
                    reference_parts[-1] = (previous_reference_part[0],
                                           previous_reference_part[1] + [(embedded_field_name, embedded_field_value[1])])
                else:
                    reference_parts.append("", [(embedded_field_name, embedded_field_value[1])])
            else:
                assert (False)

        # Get references.
        assert (self.reference_type is not None)
        self.document = []
        for reference_part in reference_parts:
            reference_string, embedded_field = reference_part
            reference_string = reference_string.strip()
            homonym_match = SourceFileTermPostField.reference_with_homonym_number_re.match(reference_string)
            if (homonym_match):
                reference = {"term": string_substitutions(homonym_match.group(1)),
                             "homonym": int(homonym_match.group(2))}
            else:
                reference = {"term": string_substitutions(reference_string)}
            for field_name, value in embedded_field:
                if (field_name == "HONR"):
                    reference["homonym"] = int(value)
                else:
                    reference_parts[field_name] = value
            reference["lang"] = self.language
            self.document.append((self.reference_type, reference))

    def parse_field_term(self):
        assert (self.field_name in ["AVTE", "BT", "INTE", "PH", "SYPH", "SYTE", "TE"])
        self.field = {}

        # Get term field structure.
        if (self.has_embedded_fields):
            self.log("Inte bara text i term: ", self.field_value_parts)
            self.is_ok = False
            return
        else:
            if (paragraph_separator in self.value):
                self.log("Nytt stycke i term: ", self.value)
                self.is_ok = False
                return
            tree = self.get_styled_text(self.value, "{", "}")
            if (tree is None):
                self.field = {"term": string_substitutions(self.value)}
            elif ((len(tree) == 1) and (tree[0].keys() == set(["text"]))):
                self.field = {"term": tree[0]["text"]}
            else:
                self.field = {"annotated_term": tree}

        # Check if field is split.
        if (self.field_name in ["TE", "SYTE", "PH", "SYPH"]):
            if (self.term_post_source.is_split(self.language)):
                self.field["split"] = True

        self.get_subfields()
        self.field["status"] = self.field_name
        self.field["lang"] = self.language
        self.document = [("terms", self.field)]

        # Get search terms.
        text = None
        if ("term" in self.field):
            text = self.field["term"]
        if ("annotated_term" in self.field):
            text = self.get_text_from_term(self.field["annotated_term"])
        if (text):
            self.document.extend([("searchterms", word) for word in self.get_words(text)])


class SourceFileTermPostFieldAn(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'AN')
        self.accepts_embedded_fields = True
        self.information_type = 'comment'
        self.possible_subfields = ['ETYM', 'RF']
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        self.parse_field_information()


class SourceFileTermPostFieldAvte(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'AVTE')
        self.accepts_embedded_fields = False
        self.is_single_field = False
        self.possible_subfields = sub_level_field_names

    def parse_field(self):
        super().parse_field()
        self.parse_field_term()


class SourceFileTermPostFieldBngr(SourceFileTermPostField):

    # Regular expression matching.
    eller_defneu_inflection = re.compile(r"^([^, ]+) ([^, ]+) eller ([^, ]+) ([^, ]+)$")
    eller_in_second_clause = re.compile(r"^([^, ]+),? ([^ ]+) eller ([^ ]+)$")
    eller_twice_pl = re.compile(r"^([^, ]+) eller ([^, ]+) pl\.? ([^, ]+) eller ([^, ]+)$")
    koll_pl_only = re.compile(r"^koll\. pl\.$")
    normal_inflection = re.compile(r"^([^, ]+),? (?:pl\.? )?(.+)$")
    normal_inflection_with_koll = re.compile(r"^([^, ]+),? (?:pl\. )?(.+) koll\.? (.+)$")
    pl_och_koll = re.compile(r"^(.+) pl\.? och koll\.? (.+)$")
    pl_och_koll_sep = re.compile(r"^(.+) pl\.? (.+) koll\.? (.+)$")
    pl_only = re.compile(r"^pl\.? (.+)$")

    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'BNGR')
        self.is_single_field = True

    def get_inflection(self, inflection):
        match = SourceFileTermPostFieldBngr.koll_pl_only.match(inflection)
        if (match):
            return []

        match = SourceFileTermPostFieldBngr.pl_och_koll.match(inflection)
        if (match):
            return [{"definite_or_neuter": match.group(1), "plural": match.group(2), "collective": match.group(2)}]

        match = SourceFileTermPostFieldBngr.pl_och_koll_sep.match(inflection)
        if (match):
            return [{"definite_or_neuter": match.group(1), "plural": match.group(2), "collective": match.group(3)}]

        match = SourceFileTermPostFieldBngr.normal_inflection_with_koll.match(inflection)
        if (match):
            return [{"definite_or_neuter": match.group(1), "plural": match.group(2), "collective": match.group(3)}]

        match = SourceFileTermPostFieldBngr.eller_in_second_clause.match(inflection)
        if (match):
            return [{"definite_or_neuter": match.group(1), "plural": match.group(2)},
                    {"definite_or_neuter": match.group(1), "plural": match.group(3)}]

        match = SourceFileTermPostFieldBngr.pl_only.match(inflection)
        if match:
            return [{"plural": match.group(1)}]

        match = SourceFileTermPostFieldBngr.eller_twice_pl.match(inflection)
        if (match):
            return [{"definite_or_neuter": match.group(1), "plural": match.group(3)},
                    {"definite_or_neuter": match.group(2), "plural": match.group(4)}]

        match = SourceFileTermPostFieldBngr.eller_defneu_inflection.match(inflection)
        if (match):
            return [{"definite_or_neuter": match.group(1), "plural": match.group(2)},
                    {"definite_or_neuter": match.group(3), "plural": match.group(4)}]

        match = SourceFileTermPostFieldBngr.normal_inflection.match(inflection)
        if (match):
            return [{"definite_or_neuter": match.group(1), "plural": match.group(2)}]

        definite_or_neuter, _, plural = inflection.partition(" ")
        if (not definite_or_neuter):
            definite_or_neuter = None
        if (not plural):
            plural = None
        return [{"definite_or_neuter": definite_or_neuter, "plural": plural}]

    def get_subfield_value(self, parent_field):
        inflection = self.get_inflection(self.value)
        if (inflection):
            parent_field.field.setdefault("inflection", []).extend(inflection)


class SourceFileTermPostFieldBt(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'BT')
        self.accepts_embedded_fields = False
        self.is_single_field = False

    def parse_field(self):
        super().parse_field()
        # BT (beteckning) is maybe not a term, but it is handled in the same way.
        self.parse_field_term()


class SourceFileTermPostFieldDf(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'DF')
        self.accepts_embedded_fields = True
        self.information_type = 'definition'
        self.is_single_field = True
        self.possible_subfields = ['ETYM', 'RF']

    @staticmethod
    def get_field_name():
        return 'DF'

    def parse_field(self):
        super().parse_field()
        self.parse_field_information()


class SourceFileTermPostFieldEkvi(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'EKVI')
        self.accepts_embedded_fields = True
        self.information_type = 'equivalence'
        self.possible_subfields = ['ETYM', 'RF']
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        self.parse_field_information()


class SourceFileTermPostFieldEtym(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'ETYM')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        if (parent_field.information_type is None):
            parent_field.field["etymology"] = string_substitutions(self.value)
        # else: Ignore ETYM field when under information field.


class SourceFileTermPostFieldEx(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'EX')
        self.accepts_embedded_fields = True
        self.information_type = 'example'
        self.possible_subfields = ['ETYM', 'RF']
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        self.parse_field_information()


class SourceFileTermPostFieldFk(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'FK')
        self.accepts_embedded_fields = True
        self.information_type = 'explanation'
        self.possible_subfields = ['ETYM', 'RF']
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        self.parse_field_information()


class SourceFileTermPostFieldFo(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'FO')
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        self.document = [(self.key, string_substitutions(self.value))]


class SourceFileTermPostFieldFrkt(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'FRKT')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        if (self.value.lower() == "of"):
            parent_field.field["full form"] = True
        elif (self.value.lower() == "f"):
            parent_field.field["abbreviation"] = True
        else:
            self.log("Ogiltig förkortningsstatus: ", self.value)
            self.is_ok = False
            parent_field.is_ok = False


class SourceFileTermPostFieldGe(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'GE')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        parent_field.field["geo"] = self.value


class SourceFileTermPostFieldGngr(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'GNGR')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        parent_field.field["gender"] = " eller ".join([value_part.strip() for value_part in self.value.split(",")])


class SourceFileTermPostFieldGr(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'GR')
        self.is_single_field = False

    def get_subfield_value(self, parent_field):
        parent_field.field["grammar"] = self.value


class SourceFileTermPostFieldHonr(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'HONR')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        parent_field.field["homonym"] = int(self.value)


class SourceFileTermPostFieldIllt(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'ILLT')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        parent_field.field["text"] = string_substitutions(self.value)


class SourceFileTermPostFieldIllu(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'ILLU')
        self.is_single_field = False
        self.possible_subfields = ['ILLT', 'ILTY']

    def parse_field(self):
        super().parse_field()
        assert (self.field_name in ["ILLU"])
        assert (not self.has_embedded_fields)
        filename = self.value
        illustration_entries = {}
        illustration_entries = illustration_entries.setdefault("illustrations", {})
        self.field = illustration_entries.setdefault(filename, {"filename": filename})
        self.get_subfields()
        self.document = [(self.field_name, self.field.copy())] + []


class SourceFileTermPostFieldIlty(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'ILTY')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        if (self.value == "begreppsdiagram"):
            parent_field.field["type"] = "concept diagram"
        else:
            self.log("Oimplementerat illustrationstyp: ", self.value, " i ", self.field_name)
            parent_field.is_ok = False


class SourceFileTermPostFieldInan(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'INAN')
        self.is_single_field = False

    def parse_field(self):
        self.document = [("internal_comment", string_substitutions(self.value))]


class SourceFileTermPostFieldInid(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'INID')
        self.accepts_embedded_fields = False
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        assert (not self.has_embedded_fields)
        if (is_integer_string(self.value)):
            self.document = [(self.key, int(self.value))]
        else:
            self.document = []


class SourceFileTermPostFieldInna(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'INNA')
        self.accepts_embedded_fields = False
        self.is_single_field = True

    def parse_field(self):
        self.document = [(self.key, string_substitutions(self.value))]


class SourceFileTermPostFieldInte(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'INTE')
        self.accepts_embedded_fields = False
        self.possible_subfields = sub_level_field_names
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        self.parse_field_term()


class SourceFileTermPostFieldKl(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'KL')
        self.is_single_field = True

    def parse_field(self):
        self.document = []


class SourceFileTermPostFieldKt(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'KT')
        self.accepts_embedded_fields = False
        self.is_single_field = False
        self.possible_subfields = ['RF']

    def parse_field(self):
        super().parse_field()
        assert (not self.has_embedded_fields)
        text_rows = []
        for field_value_part in self.field_value_parts:
            text = string_substitutions(field_value_part[0][1])
            tree = self.get_styled_text(text, "{", "}")
            if (tree is None):
                text_rows.append({"text": text})
            elif ((len(tree) == 1) and ("text" in tree[0])):
                text_rows.append({"text": tree[0]["text"]})
            else:
                text_rows.append({"annotated_text": tree})

        field = {
            "paragraphs": text_rows,
            "lang": self.language,
            "type": "context",
        }
        self.document = [("texts", field)]



class SourceFileTermPostFieldNa(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'NA')
        self.accepts_embedded_fields = False
        self.is_single_field = True

    def parse_field(self):
        self.document = [(self.key, string_substitutions(self.value))]



class SourceFileTermPostFieldOkgr(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'OKGR')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        if (self.term_post_source.is_position_not_obvious):
            parent_field.field["pos_not_obvious"] = True
            parent_field.field["pos"] = self.value


#
# Waiting on final decision from terminologists
#
# class SourceFileTermPostFieldOrg(SourceFileTermPostField):
#    def __init__(self, row, source_file):
#        super().__init__(row, source_file)
#        assert (self.field_name == 'ORG')
#        self.information_type = 'terminologist_explanation'
#        self.is_single_field = False

#    def parse_field(self):
#        super().parse_field()
#        self.parse_field_information()


class SourceFileTermPostFieldPh(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'PH')
        self.accepts_embedded_fields = False
        self.possible_subfields = sub_level_field_names
        self.is_single_field = False

    def parse_field(self):
        super().parse_field()
        self.parse_field_term()
        self.field["status"] = "TE"
        self.field["phrase"] = True


class SourceFileTermPostFieldRete(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'RETE')
        self.accepts_embedded_fields = True
        self.is_single_field = False
        self.possible_subfields = ['ETYM']
        self.reference_type = "seealso"

    def parse_field(self):
        super().parse_field()
        self.parse_field_reference()


class SourceFileTermPostFieldRf(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'RF')
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        parent_field.field["source"] = string_substitutions(self.value)  # XXX also breaks URLs


class SourceFileTermPostFieldRian(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'RIAN')
        self.is_single_field = False

    def parse_field(self):
        self.document = [("internal_comment", string_substitutions(self.value))]


class SourceFileTermPostFieldRifr(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'RIFR')
        self.is_single_field = False

    def parse_field(self):
        self.document = []


class SourceFileTermPostFieldSaLanguageField(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'SA')
        self.accepts_embedded_fields = True
        self.information_type = 'domain'
        self.possible_subfields = ['ETYM']
        self.is_single_field = True

    def parse_field(self):
        super().parse_field()
        self.parse_field_information()


class SourceFileTermPostFieldSpbr(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        self.accepts_embedded_fields = False
        assert (self.field_name == 'SPBR')
        self.information_type = 'language_usage_explanation'
        self.is_single_field = False

    def parse_field(self):
        super().parse_field()
        self.parse_field_information()


class SourceFileTermPostFieldSaSubField(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'SA')
        self.accepts_embedded_fields = False
        self.is_single_field = True

    def get_subfield_value(self, parent_field):
        parent_field.field["narrowing"] = string_substitutions(self.value)


class SourceFileTermPostFieldSu(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'SU')
        self.accepts_embedded_fields = True
        self.is_single_field = True
        self.reference_type = "seeunder"

    def parse_field(self):
        super().parse_field()
        self.parse_field_reference()


class SourceFileTermPostFieldSyph(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'SYPH')
        self.accepts_embedded_fields = False
        self.possible_subfields = sub_level_field_names
        self.is_single_field = False

    def parse_field(self):
        super().parse_field()
        self.parse_field_term()
        # Change field name to "SYTE". "SYPH has the same result as "SYTE"
        # except for the value 'self.field["phrase"] = True'
        self.field["status"] = "SYTE"
        self.field["phrase"] = True


class SourceFileTermPostFieldSyte(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'SYTE')
        self.accepts_embedded_fields = False
        self.is_single_field = False
        self.possible_subfields = sub_level_field_names

    def parse_field(self):
        super().parse_field()
        self.parse_field_term()


class SourceFileTermPostFieldTe(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'TE')
        self.accepts_embedded_fields = False
        self.possible_subfields = sub_level_field_names
        self.is_single_field = False
        # if (self.value == 'krustläder'):
            # tilt = 42

    def parse_field(self):
        super().parse_field()
        self.parse_field_term()


class SourceFileTermPostFieldTi(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'TI')
        self.accepts_embedded_fields = False
        self.is_single_field = True

    def parse_field(self):
        self.document = []


class SourceFileTermPostFieldUpte(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'UPTE')
        self.accepts_embedded_fields = False
        self.is_single_field = False
        self.possible_subfields = ['ETYM']

    def parse_field(self):
        super().parse_field()
        assert (not self.has_embedded_fields)
        term_references = self.value.split(",")
        fields = [("onlysearch", {"term": string_substitutions(term_reference.strip()), "lang": self.language}) for term_reference in term_references]

        search_terms = []
        for reference_type, term_reference in fields:
            search_terms.extend([("searchterms", word) for word in self.get_words(term_reference["term"])])
        self.document = fields + search_terms


class SourceFileTermPostFieldUt(SourceFileTermPostField):
    def __init__(self, row, source_file):
        super().__init__(row, source_file)
        assert (self.field_name == 'UT')
        self.is_single_field = False

    def get_subfield_value(self, parent_field):
        parent_field.field["pronunciation"] = self.value
