from ISOF.common.list_functions import is_empty_list, is_not_empty_list
from model.import_util import decode_entities, embeddedconvert_annotated_term
from model.import_util import multidict_add, named_tag_tree, sort_terms, string_substitutions
from model.import_util import split_searchterm, tag_tree, textconvert_annotated_text, textconvert_annotated_term
import re


top_level = [
    "FO",
    "ILLU",
    "INAN",
    "INID",
    "KL",
    "ORG",
    "RIAN",
    "RIFR",
]

top_level_lang = [
    "AN",
    "AVTE",
    "BT",
    "DF",
    "EKVI",
    "EX",
    "FK",
    "INNA",
    "INTE",
    "KT",
    "NA",
    "PH",
    "RETE",
    "SA",
    "SPBR",
    "SU",
    "SYPH",
    "SYTE",
    "TE",
    "UPTE",
    "TI",
]

sub_level = [
    "BNGR",
    "ETYM",
    "FRKT",
    "GE",
    "GR",
    "GNGR",
    "HONR",
    "ILLT",
    "ILTY",
    "OKGR",
    "RF",
    "SA",
    "UT",
]

SWEDISH_LANGUAGE_CODE = 'sv'


def field_generator(source_file, rows):
    current_value = None
    current_lang = None
    current_field_name = None
    for row in rows:
        m = source_file.split_row_re.match(row)
        if m:
            key = m.group(1)
            value = m.group(2)
        else:
            key = row
            value = None
        (lang, field_name) = split_key(source_file, key)
        if field_name is None:
            # XXX bug compatibility
            source_file.log("Hittar inget fältnamn, tolkar som fortsättningsrad:", row)
            if not current_value:
                return
            current_value += "\u2029" + row.lstrip(" \t")
            continue
        if current_value is not None:
            yield (current_value, current_lang, current_field_name)
        current_value = value
        current_lang = lang
        current_field_name = field_name
    if current_value is not None:
        yield (current_value, current_lang, current_field_name)


def split_key(source_file, key):
    if (key[0:2].islower() and key[2:].isupper()):
        # Two character language code.
        field_name = key[2:]
        language = key[0:2]
        return (language, field_name)
    elif (key[0:3].islower() and key[3:].isupper()):
        # Three character language code.
        field_name = key[3:]
        language = key[0:3]
        return (language, field_name)
    elif key.isupper():
        return (None, key)
    else:
        source_file.log("Ogiltigt fältnamn:", repr(key))
        # sys.exit(1)
        return (None, None)


class SourceFileTermPostOld():
    """ Holds information about one term post in a file in "Bankvalvet".
    """
    angle_bracket_tag_valid_re = re.compile(r"^[A-Z]")
    reference_with_homonym_number_re = re.compile(r"^(.+) \((\d+)\)$")
    split_row_re = re.compile(r"^(\S+)\s+(.*)$")
    pl_och_koll = re.compile(r"^(.+) pl\.? och koll\.? (.+)$")
    pl_och_koll_sep = re.compile(r"^(.+) pl\.? (.+) koll\.? (.+)$")
    koll_pl_only = re.compile(r"^koll\. pl\.$")
    normal_inflection = re.compile(r"^([^, ]+),? (?:pl\.? )?(.+)$")
    normal_inflection_with_koll = re.compile(r"^([^, ]+),? (?:pl\. )?(.+) koll\.? (.+)$")
    eller_in_second_clause = re.compile(r"^([^, ]+),? ([^ ]+) eller ([^ ]+)$")
    eller_defneu_inflection = re.compile(r"^([^, ]+) ([^, ]+) eller ([^, ]+) ([^, ]+)$")
    eller_twice_pl = re.compile(r"^([^, ]+) eller ([^, ]+) pl\.? ([^, ]+) eller ([^, ]+)$")
    pl_only = re.compile(r"^pl\.? (.+)$")

    def __init__(self, source_file, rows):
        self.entry = {'term_post': self}
        self.is_ok = True
        self.metadata = source_file.metadata
        self.old_fields = []
        self.rows = rows
        self.source_file = source_file

        # Convert text rows to source file term post fields. Old code by Magnus.
        # This code is only used to compare the old and new processing of data.
        self.parsed_fields = []
        current_field = None
        for value, lang, field_name in field_generator(self, rows):
            # print(key, lang, field_name)
            if lang is None:
                key = field_name
            else:
                key = lang + field_name
            if field_name in top_level and not lang:
                if current_field is not None:
                    self.old_fields.append(current_field)
                    parsed_field = self.parse_field(current_field, source_file, self.metadata)
                    self.parsed_fields.append(parsed_field)
                    multidict_add(self.entry, parsed_field)
                current_field = (key, value, field_name, [], lang)
            elif field_name in top_level_lang and lang:
                if current_field is not None:
                    self.old_fields.append(current_field)
                    parsed_field = self.parse_field(current_field, source_file, self.metadata)
                    self.parsed_fields.append(parsed_field)
                    multidict_add(self.entry, parsed_field)
                current_field = (key, value, field_name, [], lang)
            elif field_name in sub_level:
                if current_field is None:
                    # XXX bug compatibility
                    # This error is reported in the new code.
                    # source_file.log("Kopplat fält utan toppnivå- eller språkfält, ignorerar:", field_name, value)
                    continue
                (super_key, super_value, super_field_name, super_subordinate, super_lang) = current_field
                super_subordinate.append((field_name, value))
            else:
                # XXX bug compatibility
                if current_field is None:
                    source_file.log("Ogiltigt fältnamn:", field_name, value, key)
                    continue
                source_file.log("Ogiltigt fältnamn, tolkar som fortsättningsrad:", field_name, value, key)
                (c_key, c_value, c_field_name, c_subordinate, c_lang) = current_field
                if isinstance(c_value, dict) and "value" in c_value:
                    c_value["value"] += "\u2029" + value
                if isinstance(c_value, str):
                    current_field = (c_key, c_value + "\u2029" + key + " " + value, c_field_name, c_subordinate, c_lang)
        if current_field is not None:
            self.old_fields.append(current_field)
            parsed_field = self.parse_field(current_field, source_file, self.metadata)
            self.parsed_fields.append(parsed_field)
            multidict_add(self.entry, parsed_field)

        # Sort term post information parts. Duplicate entries are removed from "searchterms" and "fulltextsearch".
        if ("terms" in self.entry):
            self.entry["terms"] = sort_terms(self.entry["terms"])
        if ("searchterms" in self.entry):
            self.entry["searchterms"] = sorted(set(self.entry["searchterms"]))
        if ("fulltextsearch" in self.entry):
            if ("searchterms" in self.entry):
                self.entry["fulltextsearch"] = sorted(set(self.entry["searchterms"]) |
                                                      set(self.entry["fulltextsearch"]))
            else:
                self.entry["fulltextsearch"] = sorted(set(self.entry["fulltextsearch"]))

        self.entry['is_ok'] = self.is_ok

    def bug_compat_render_term(self, tree):
        s = ""
        for e in tree:
            if e[0] == "tag":
                s += "<%s %s>" % (e[1], self.bug_compat_render_term(e[2]))
            elif e[0] == "text":
                s += e[1]
            else:
                assert (False)
        return s

    def convert_tree(self, s, in_tree, nonembedded_pos=0, limit=None):
        tree = []
        if limit == None:
            limit = len(s)
        for start_pos, end_pos, subtree in in_tree:
            field = s[start_pos + 1:end_pos]
            value_start = s.find(" ", start_pos + 1, end_pos)
            # print("convert_tree", s, in_tree, nonembedded_pos, limit, value_start, start_pos, end_pos)
            if value_start >= 0:
                key = s[start_pos + 1:value_start]
                if not SourceFileTermPostOld.angle_bracket_tag_valid_re.match(key):
                    if tree and tree[-1][0] == "text":
                        tree[-1] = ("text", tree[-1][1] + s[nonembedded_pos:end_pos + 1])
                    else:
                        tree.append(("text", s[nonembedded_pos:end_pos + 1]))
                    nonembedded_pos = end_pos + 1
                    continue
                if nonembedded_pos != start_pos:
                    if tree and tree[-1][0] == "text":
                        tree[-1] = ("text", tree[-1][1] + s[nonembedded_pos:start_pos])
                    else:
                        tree.append(("text", s[nonembedded_pos:start_pos]))
                tree.append(("tag", key, self.convert_tree(s, subtree, value_start + 1, limit=end_pos)))
                nonembedded_pos = end_pos + 1
            else:
                if tree and tree[-1][0] == "text":
                    tree[-1] = ("text", tree[-1][1] + s[nonembedded_pos:end_pos + 1])
                else:
                    tree.append(("text", s[nonembedded_pos:end_pos + 1]))
                nonembedded_pos = end_pos + 1
        if nonembedded_pos != limit:
            if tree and tree[-1][0] == "text":
                tree[-1] = ("text", tree[-1][1] + s[nonembedded_pos:limit])
            else:
                tree.append(("text", s[nonembedded_pos:limit]))
        return tree

    def format_tree(self, in_tree, source_file):
        tree = []
        for e in in_tree:
            if e[0] == "tag":
                tree.append({"tag": e[1], "content": self.format_tree(e[2], source_file)})
            elif e[0] == "text":
                tree.extend(named_tag_tree(decode_entities(e[1]), "{", "}", log=source_file.log))
            else:
                assert (False)
        return tree

    def handle_rcon(self, in_tree):
        tree = []
        for e in in_tree:
            if e.get("tag") == "RCON":
                tree.append({"term": self.handle_rcon_reference(e["content"])})
            elif "tag" in e:
                tree.append({"tag": e["tag"], "content": self.handle_rcon(e["content"])})
            else:
                tree.append(e)
        return tree

    def handle_rcon_reference(self, in_tree):
        assert (in_tree[0].keys() == set(["text"]))
        result = {"term": in_tree[0]["text"].rstrip(" ")}
        if len(in_tree) == 1:
            return result
        assert (len(in_tree) == 2 and in_tree[1].get("tag") == "HONR")
        assert (len(in_tree[1]["content"]) == 1)
        assert (in_tree[1]["content"][0].keys() == set(["text"]))
        result["homonym"] = in_tree[1]["content"][0]["text"]
        return result

    def log_term_post(self):
        # Log rows in term post.
        self.source_file.log('Termpost:')
        for row in self.rows:
            self.source_file.log(row)
        self.source_file.log('\t')

    def only_honr(self, tree):
        tags = set()
        for e in tree:
            if "tag" in e:
                tags.add(e.get("tag"))
        return tags == set(["HONR"])

    def pack_RETE(self, s, embedded, lang, source_file):
        s = s.strip()
        m = SourceFileTermPostOld.reference_with_homonym_number_re.match(s)
        if m:
            result = {"term": string_substitutions(m.group(1)), "homonym": int(m.group(2))}
        else:
            result = {"term": string_substitutions(s)}
        for key, value in embedded:
            if key == "HONR":
                result["homonym"] = int(value)
            else:
                # Error is reported in the new code.
                # source_file.log("Oimplementerat kopplat fält för RETE/SU:", key)
                result[key] = value
        result["lang"] = lang
        return result

    def parse_embedded(self, s, source_file):
        tree = tag_tree(s, "<", ">", log=source_file.log)
        converted = self.convert_tree(s, tree)
        embedded = []
        nonembedded = ""
        nonembedded_pos = 0
        for start_pos, end_pos, subtree in tree:
            if (subtree):
                index = 12
            field = s[start_pos + 1:end_pos]
            (key, _, value) = field.partition(" ")
            embedded.append((key, value, start_pos, end_pos))
            nonembedded += s[nonembedded_pos:start_pos].rstrip()
            nonembedded_pos = end_pos + 1
        nonembedded += s[nonembedded_pos:]
        return (nonembedded, embedded, converted)

    def parse_field(self, field, source_file, metadata):
        (key, rawvalue, fieldname, subordinates, lang) = field
        extra = []
        rows = rawvalue.split("\u2029")
        # (_value, _embedded_fields, converted_tree) = parse_embedded(rawvalue, source_file)
        converted_trees = [self.parse_embedded(row, source_file)[2] for row in rows]
        only_text = all([len(e) == 1 and e[0][0] == "text" for e in converted_trees])
        if only_text:
            only_text_value = "\u2029".join([e[0][1] for e in converted_trees])

        if (subordinates):
            index = 1


        if fieldname in ["TE", "SYTE", "AVTE", "PH", "SYPH", "INTE"]:
            if only_text:
                if "\u2029" in only_text_value:
                    source_file.log("Nytt stycke i term:", only_text_value)
                tree = named_tag_tree(only_text_value, "{", "}", log=source_file.log)
                if tree == None:
                    field = {"term": string_substitutions(only_text_value)}
                elif len(tree) == 1 and tree[0].keys() == set(["text"]):
                    field = {"term": tree[0]["text"]}
                else:
                    field = {"annotated_term": tree}
            else:
                # XXX Bug compatibility
                source_file.log("Inte bara text i term:", converted_trees)
                termstring = "\u2029".join([self.bug_compat_render_term(e) for e in converted_trees])
                field = {"term": string_substitutions(termstring)}

            if fieldname in ["TE", "SYTE", "PH", "SYPH"]:
                if self.splitPreferredAdmitted(metadata, lang):
                    field["split"] = True
            for k, v in subordinates:
                if k == "GE":
                    field["geo"] = v
                elif k == "HONR":
                    field["homonym"] = int(v)
                elif k == "GNGR":
                    field["gender"] = " eller ".join([e.strip() for e in v.split(",")])
                elif k == "GR":
                    field["grammar"] = v
                elif k == "UT":
                    field["pronunciation"] = v
                elif k == "BNGR":
                    definite_or_neuter, _, plural = v.partition(" ")
                    if not definite_or_neuter:
                        definite_or_neuter = None
                    if not plural:
                        plural = None
                    parsed_inflection = self.parse_inflection(v)
                    if parsed_inflection:
                        field.setdefault("inflection", []).extend(parsed_inflection)
                elif k == "OKGR":
                    if metadata.entries.get("Tolka_OKGR-märkning",
                                            "baraOmInteOrdklassSjälvklar") == "baraOmInteOrdklassSjälvklar":
                        field["pos_not_obvious"] = True
                        field["pos"] = v
                elif k == "SA":
                    field["narrowing"] = string_substitutions(v)
                elif k == "RF":
                    field["source"] = string_substitutions(v)  # XXX also breaks URLs
                elif k == "ETYM":
                    field["etymology"] = string_substitutions(v)
                elif k == "FRKT":
                    if v.lower() == "of":
                        field["full form"] = True
                    elif v.lower() == "f":
                        field["abbreviation"] = True
                    else:
                        source_file.log("Ogiltig förkortningsstatus", v)
                else:
                    pass
                    # Error is reported in the new code.
                    # source_file.log("Oimplementerat kopplat fält", k, "under toppnivåfält", fieldname)
            if fieldname == "PH":
                field["status"] = "TE"
                field["phrase"] = True
            elif fieldname == "SYPH":
                field["status"] = "SYTE"
                field["phrase"] = True
            else:
                field["status"] = fieldname
            field["lang"] = lang
            if "term" in field:
                extra.extend([("searchterms", e) for e in split_searchterm(field["term"])])
            if "annotated_term" in field:
                extra.extend(
                    [("searchterms", e) for e in split_searchterm(textconvert_annotated_term(field["annotated_term"]))])
            return [("terms", field)] + extra
        elif fieldname == "UPTE":
            assert (only_text)
            split_value = only_text_value.split(",")
            fields = [("onlysearch", {"term": string_substitutions(s.strip()), "lang": lang}) for s in split_value]

            for k, v in subordinates:
                if k == "OKGR":
                    for field in fields:
                        if metadata.entries.get("Tolka_OKGR-märkning",
                                                "baraOmInteOrdklassSjälvklar") == "baraOmInteOrdklassSjälvklar":
                            field[1]["pos_not_obvious"] = True
                        field[1]["pos"] = v
                elif k == "ETYM":
                    pass
                else:
                    pass
                    # Error is handled in mew code.
                    # source_file.log("Oimplementerat kopplat fält", k, "under toppnivåfält", fieldname)
            for _, fvalue in fields:
                extra.extend([("searchterms", e) for e in split_searchterm(fvalue["term"])])
            return fields + extra
        elif fieldname == "ILLU":
            assert (only_text)
            filename = only_text_value
            illustration_entries = metadata.entries.setdefault("illustrations", {})
            illustration_field = illustration_entries.setdefault(filename, {"filename": filename})
            for k, v in subordinates:
                if k == "ILLT":
                    illustration_field["text"] = string_substitutions(v)
                elif k == "ILTY":
                    if v == "begreppsdiagram":
                        illustration_field["type"] = "concept diagram"
                    else:
                        source_file.log("Oimplementerat illustrationstyp", v, "i", k)
                else:
                    source_file.log("Oimplementerat kopplat fält", k, "under toppnivåfält", fieldname)
            return [(key, illustration_field.copy())] + extra
        elif fieldname == "INID":
            assert (only_text)
            try:
                return [(key, int(only_text_value))]
            except ValueError:
                # self.log("Invalid internal ID:", only_text_value)
                return []
        elif fieldname in ["INAN", "RIAN"]:
            assert (only_text)
            return [("internal_comment", string_substitutions(only_text_value))]
        elif fieldname in ["DF", "FK", "AN", "EX", "SA", "EKVI", "SPBR"]:
            textrows = []
            for converted_tree in converted_trees:
                if not only_text:
                    if len(converted_tree) == 1 and converted_tree[0][0] == "text":
                        textrows.append({"text": string_substitutions(converted_tree[0][1])})
                    else:
                        if self.only_honr(self.format_tree(converted_tree, source_file)):
                            textrows.append({"text": embeddedconvert_annotated_term(self.format_tree(converted_tree, source_file))})
                        else:
                            textrows.append({"annotated_text": self.handle_rcon(self.format_tree(converted_tree, source_file))})
                else:
                    tree = named_tag_tree(string_substitutions(converted_tree[0][1]), "{", "}", log=source_file.log)
                    if tree == None:
                        textrows.append({"text": string_substitutions(converted_tree[0][1])})
                    elif len(tree) == 1 and tree[0].keys() == set(["text"]):
                        textrows.append({"text": tree[0]["text"]})
                    else:
                        textrows.append({"annotated_text": tree})
            field = {"paragraphs": textrows}
            for k, v in subordinates:
                if k == "RF":
                    field["source"] = string_substitutions(v)  # XXX also breaks URLs
                elif k == "ETYM":
                    pass  # ignore ETYM field when under text field
                else:
                    source_file.log("Oimplementerat kopplat fält", k, "under toppnivåfält", fieldname)
            field["lang"] = lang
            text_types = {"DF": "definition", "FK": "explanation", "EX": "example",
                          "AN": "comment", "SPBR": "language_usage_explanation", "SA": "domain", "EKVI": "equivalence"}
            field["type"] = text_types[fieldname]
            # print(field)
            for textrow in textrows:
                if "annotated_text" in textrow:
                    extra.extend([("fulltextsearch", e) for e in
                                  split_searchterm(textconvert_annotated_text(textrow["annotated_text"]))])
                else:
                    extra.extend([("fulltextsearch", e) for e in split_searchterm(textrow["text"])])
            return [("texts", field)] + extra
        elif fieldname == "KT":
            assert (only_text)
            textrows = []
            for converted_tree in converted_trees:
                tree = named_tag_tree(string_substitutions(converted_tree[0][1]), "{", "}", log=source_file.log)
                if tree == None:
                    textrows.append({"text": string_substitutions(converted_tree[0][1])})
                elif len(tree) == 1 and tree[0].keys() == set(["text"]):
                    textrows.append({"text": tree[0]["text"]})
                else:
                    textrows.append({"annotated_text": tree})

            field = {
                "paragraphs": textrows,
                "lang": lang,
                "type": "context",
            }
            return [("texts", field)]
        if fieldname == "RETE":
            assert (len(converted_trees) == 1)
            value = self.split_embedded_converted(converted_trees[0], ",")
            result = [("seealso", self.pack_RETE(*e, lang, source_file)) for e in value]
        elif fieldname == "SU":
            assert (len(converted_trees) == 1)
            value = self.split_embedded_converted(converted_trees[0], ",")
            result = [("seeunder", self.pack_RETE(*e, lang, source_file)) for e in value]
        elif fieldname == "RIFR":
            return []
        elif fieldname == "KL":
            return []
        elif fieldname == "BT":
            assert (only_text)
            if "\u2029" in only_text_value:
                source_file.log("Nytt stycke i beteckning:", only_text_value)
            tree = named_tag_tree(only_text_value, "{", "}", log=source_file.log)
            if tree == None:
                term = {"term": string_substitutions(only_text_value)}
            elif len(tree) == 1 and tree[0].keys() == set(["text"]):
                term = {"term": tree[0]["text"]}
            else:
                term = {"annotated_term": tree}

            term["status"] = "BT"
            term["lang"] = lang
            result = [("terms", term)]
            if "term" in term:
                result.extend([("searchterms", e) for e in split_searchterm(term["term"])])
            if "annotated_term" in term:
                searchterms = split_searchterm(textconvert_annotated_term(term["annotated_term"]))
                result.extend([("searchterms", e) for e in searchterms])
        elif fieldname == "TI":
            result = []
        else:
            if only_text:
                result = [(key, string_substitutions(only_text_value))]
            else:
                # XXX
                result = [
                    (key, string_substitutions("\u2029".join([self.bug_compat_render_term(e) for e in converted_trees])))]
        for k, v in subordinates:
            # assert(only_text) # XXX
            if k == "ETYM":
                continue  # ignore ETYM field
            # Error is handled in mew code.
            # source_file.log("Oimplementerat kopplat fält", k, "under toppnivåfält", fieldname)
        return result

    def splitPreferredAdmitted(self, metadata, lang):
        value = metadata.entries.get("Tolka_skillnadTermSynonym_" + lang,
                                   metadata.entries.get("Tolka_skillnadTermSynonym_förvalt", "troligenUppdelat"))
        if value == "troligenIngenSkillnad":
            return False
        elif value == "troligenUppdelat":
            return True
        else:
            return False  # XXX report error

    def parse_inflection(self, inflection):
        m = SourceFileTermPostOld.koll_pl_only.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "koll_pl_only")
            return []
        m = SourceFileTermPostOld.pl_och_koll.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "pl_och_koll", repr((m.group(1), m.group(2))))
            return [{"definite_or_neuter": m.group(1), "plural": m.group(2), "collective": m.group(2)}]
        m = SourceFileTermPostOld.pl_och_koll_sep.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "pl_och_koll_sep", repr((m.group(1), m.group(2), m.group(3))))
            return [{"definite_or_neuter": m.group(1), "plural": m.group(2), "collective": m.group(3)}]
        m = SourceFileTermPostOld.normal_inflection_with_koll.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "normal_inflection_with_koll", repr((m.group(1), m.group(2), m.group(3))))
            return [{"definite_or_neuter": m.group(1), "plural": m.group(2), "collective": m.group(3)}]
        m = SourceFileTermPostOld.eller_in_second_clause.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "eller_in_second_clause", repr((m.group(1), m.group(2), m.group(3))))
            return [{"definite_or_neuter": m.group(1), "plural": m.group(2)},
                    {"definite_or_neuter": m.group(1), "plural": m.group(3)}]
        m = SourceFileTermPostOld.pl_only.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "pl_only", repr((m.group(1))))
            return [{"plural": m.group(1)}]
        m = SourceFileTermPostOld.eller_twice_pl.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "eller_twice_pl", repr((m.group(1), m.group(2), m.group(3), m.group(4))))
            return [{"definite_or_neuter": m.group(1), "plural": m.group(3)},
                    {"definite_or_neuter": m.group(2), "plural": m.group(4)}]
        m = SourceFileTermPostOld.eller_defneu_inflection.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "eller_defneu_inflection", repr((m.group(1), m.group(2), m.group(3), m.group(4))))
            return [{"definite_or_neuter": m.group(1), "plural": m.group(2)},
                    {"definite_or_neuter": m.group(3), "plural": m.group(4)}]
        m = SourceFileTermPostOld.normal_inflection.match(inflection)
        if m:
            #        self.log("inflection", repr(inflection), "normal_inflection", repr((m.group(1), m.group(2))))
            return [{"definite_or_neuter": m.group(1), "plural": m.group(2)}]

        # self.log("inflection", repr(inflection), "fallthrough")
        definite_or_neuter, _, plural = inflection.partition(" ")
        if not definite_or_neuter:
            definite_or_neuter = None
        if not plural:
            plural = None
        return [{"definite_or_neuter": definite_or_neuter, "plural": plural}]

    def split_embedded_converted(self, tree, separator):
        # print("split_embedded", repr(s), embedded_fields, repr(separator))
        result = []
        for e in tree:
            if e[0] == "text":
                value = e[1]
                parts = value.split(separator)
                # positions = find_all(s, separator)
                # for position in positions:
                # print("parts", parts)
                if result:
                    lastresult = result[-1]
                    result[-1] = (lastresult[0] + parts[0], lastresult[1])
                    parts = parts[1:]
                for part in parts:
                    if part:
                        result.append((part, []))
            elif e[0] == "tag":
                tagname = e[1]
                subtree = e[2]
                assert (len(subtree) == 1)
                node = subtree[0]
                assert (node[0] == "text")
                if result:
                    lastresult = result[-1]
                    result[-1] = (lastresult[0], lastresult[1] + [(tagname, node[1])])
                else:
                    result.append("", [(tagname, node[1])])
            else:
                assert (False)
        return result
