import collections
from copy import deepcopy
from datetime import datetime
from django.conf import settings
import json
from model.comparison import compare_sources
from model.database import Database
from model.mongodb_manager import MongoDBManager
from model.source_file import SourceFile
from model.source_file_manager import SourceFileManager
from model.source_file_term_post_field import SourceFileTermPostField
from model.term_post_source_old import TermPostSourceOld
import os
import sys
import time


def sort_source_changed_file(source):
    if ('last_file_change_seconds' in source):
        return source['last_file_change_seconds']
    else:
        return 0


class TermManager():
    """ Handles information in terms (term posts and term post sources).

    """

    @staticmethod
    def _compare_term_post_information(term_posts_staging, term_post_source_staging):
        term_posts_public = TermManager.get_term_posts_by_term_post_source(term_post_source_staging["id"])
        term_post_source_public = None
        term_post_source_temp = TermManager.get_term_post_source(term_post_source_staging["id"])
        if (term_post_source_temp):
            term_post_source_public = term_post_source_temp.get_document()
        if (not term_post_source_public):
            term_post_source_public = term_post_source_staging

        # Function compare_sources() changes the term posts that it compares.
        # Make a copy of the term posts before function compare_sources() is called.
        # TODO: Change function compare_sources() to not change the data that it compares.
        # TODO: Remove this copying of term posts when function compare_sources() has been changed.
        term_posts_public_temp = collections.defaultdict(list)
        for term_post_public in term_posts_public:
            term_posts_public_temp[term_post_source_staging["id"]].append(deepcopy(term_post_public))

        term_posts_staging_temp = collections.defaultdict(list)
        for term_post_staging in term_posts_staging:
            term_posts_staging_temp[term_post_source_staging["id"]].append(deepcopy(term_post_staging))

        term_post_source_public_temp = TermPostSourceOld.fix_term_post_source_document(term_post_source_public)
        term_post_source_staging_temp = TermPostSourceOld.fix_term_post_source_document(term_post_source_staging)

        difference = compare_sources(term_post_source_public_temp,
                                     term_post_source_staging_temp,
                                     term_posts_staging_temp,
                                     term_posts_public_temp)
        term_post_source_staging['comparison'] = difference

    @staticmethod
    def get_comparison():
        """ Get comparison from staging database.
            The comparison was done during import of files from bankvalvet to staging database.

        Returns:
            (sources_differences, sources_errors): Information about differences and errors.
        """

        information_count = 0
        term_post_sources = TermManager.get_term_post_sources(Database.STAGING_DATABASE_NAME)
        term_post_sources.sort(key=sort_source_changed_file, reverse=True)

        sources_differences = []
        sources_errors = []
        for term_post_source in term_post_sources:
            source_errors = [" ".join(error) for error in term_post_source.get("errors", [])]
            source_differences = term_post_source.get("comparison", [])
            if (source_differences):
                if ("added_metadata" in source_differences):
                    information_count = information_count + len(source_differences["added_metadata"])
                if ("changed_metadata" in source_differences):
                    information_count = information_count + len(source_differences["changed_metadata"])
                if ("removed_metadata" in source_differences):
                    information_count = information_count + len(source_differences["removed_metadata"])
                if ("added_terms" in source_differences):
                    information_count = information_count + len(source_differences["added_terms"])
                if ("changed_terms" in source_differences):
                    information_count = information_count + (len(source_differences["changed_terms"]) * 10)
                if ("removed_terms" in source_differences):
                    information_count = information_count + len(source_differences["removed_terms"])

                sources_differences.append((term_post_source["id"],
                                            term_post_source["filename"],
                                            source_differences,
                                            source_errors))
            elif (source_errors):
                sources_errors.append((term_post_source["id"], term_post_source["filename"], source_errors))

            if (source_errors):
                information_count = information_count + len(source_errors)
            if (information_count > 10000):
                break

        return sources_differences, sources_errors

    @staticmethod
    def get_term_post_by_reference(term_post_reference,
                                   term_post_source_id,
                                   database=Database.PUBLIC_DATABASE_NAME):
        return MongoDBManager.get_term_post_by_reference(term_post_reference, term_post_source_id, database)

    @staticmethod
    def get_term_post_by_slug(slug, term_post_source_id, database=Database.PUBLIC_DATABASE_NAME):
        return MongoDBManager.get_term_post_by_slug(slug, term_post_source_id, database)

    @staticmethod
    def get_term_post_count(database=Database.PUBLIC_DATABASE_NAME):
        return MongoDBManager.get_term_post_count(database)

    @staticmethod
    def get_term_post_source(term_post_source_id, database=Database.PUBLIC_DATABASE_NAME):
        term_post_source_document = MongoDBManager.get_term_post_source(term_post_source_id, database)
        if (term_post_source_document is None):
            term_post_source = None
        else:
            term_post_source = TermPostSourceOld.get_term_post_source_from_document(term_post_source_document)
            if (not hasattr(term_post_source, 'term_post_count')):
                term_post_source.term_post_count = TermManager.get_term_post_source_count(term_post_source_id, database)
        return term_post_source

    @staticmethod
    def get_term_post_source_count(only_published_sources=True, database=Database.PUBLIC_DATABASE_NAME):
        return MongoDBManager.get_term_post_source_count(only_published_sources, database)

    @staticmethod
    def get_term_post_sources(database=Database.PUBLIC_DATABASE_NAME):
        return MongoDBManager.get_term_post_sources(database)

    @staticmethod
    def get_term_post_sources_by_term_posts(only_published_sources=True, database=Database.PUBLIC_DATABASE_NAME):
        return MongoDBManager.get_term_post_sources_by_term_posts(only_published_sources, database)

    @staticmethod
    def get_term_posts_by_search_criteria(search_string: str,
                                          search_type: str,
                                          search_languages: list,
                                          definition_languages: list,
                                          term_languages: list):
        return MongoDBManager.get_term_posts_by_search_criteria(search_string,
                                                                search_type,
                                                                search_languages,
                                                                definition_languages,
                                                                term_languages)

    @staticmethod
    def get_term_posts_by_term_post_source(term_post_source_id,
                                           only_swedish_term_posts=False,
                                           database=Database.PUBLIC_DATABASE_NAME):
        return MongoDBManager.get_term_posts_by_term_post_source(term_post_source_id, only_swedish_term_posts, database)

    @staticmethod
    def get_updated_source_files(database, file_paths):
        term_post_sources = TermManager.get_term_post_sources(database)
        term_post_sources_by_filename = {}
        for term_post_source in term_post_sources:
            term_post_sources_by_filename[term_post_source['filename']] = term_post_source
        file_paths = SourceFileManager.get_updated_files(file_paths, term_post_sources_by_filename)
        return file_paths

    @staticmethod
    def import_bankvalvet(database, import_all_files):
        if (import_all_files):
            MongoDBManager.reset_database(database)

        handled_source_id = set()
        imported_files = []

        # Get file paths.
        file_paths = SourceFileManager.get_text_file_paths()
        if ((database == Database.STAGING_DATABASE_NAME) and (not import_all_files)):   # If we are not importing all files
            file_paths = TermManager.get_updated_source_files(database, file_paths)     # read only the updated files

        # Import text files.
        for file_path in file_paths:
            imported_files.append(SourceFileManager.get_file_name(file_path))
            source_file = SourceFile(file_path)
            term_posts_file = source_file.term_post_documents
            term_post_source_file = source_file.get_term_post_source_document()

            if (source_file.is_ok()):
                # Update all sources in both staging and public databases.
                # Code that is used when needed normally due to changes in datahandling
                # if (import_all_files and (database == Database.STAGING_DATABASE_NAME)):
                    # TermManager.update_term_post_source(term_post_source_file,
                                                        # source_file.source_id,
                                                        # Database.PUBLIC_DATABASE_NAME)

                if (database == Database.STAGING_DATABASE_NAME):
                    TermManager._compare_term_post_information(term_posts_file, term_post_source_file)

                # Save term posts in database.
                TermManager.update_term_posts(term_posts_file, source_file.source_id, database)

            # Save term post source in database.
            # term_post_source_file = source_file.get_term_post_source_document()
            TermManager.update_term_post_source(term_post_source_file, source_file.source_id, database)
            if (source_file.source_id == 457):
                pass
            if (source_file.has_source_id()):
                handled_source_id.add(source_file.source_id)
            TermManager.print_log(file_path, source_file)

            # Give MongoDB time to process the updated term posts.
            # time.sleep(1)

        term_posts_files = []
        meta_data_files = []

        # Import the files on JSON-format:
        # Get file paths.
        file_paths = SourceFileManager.get_json_file_paths()
        if ((database == Database.STAGING_DATABASE_NAME) and (not import_all_files)):
            file_paths = TermManager.get_updated_source_files(database, file_paths)

        # Import JSON files.
        for fileobject in file_paths:
            imported_files.append(os.path.basename(fileobject))
            filename_json = os.path.basename(fileobject)
            assert filename_json[-5:] == ".json"
            source_id = int(filename_json[:-5])
            if source_id in handled_source_id:
                print("error:", source_id, "already handled and exists as 'oklara' JSON")
                sys.exit(1)
            source = json.load(open(fileobject))
            source["kalla"]["filename"] = filename_json
            source["kalla"]["status"] = source["kalla"]["Status"]
            del source["kalla"]["Status"]
            for entry in source["termposts"]:
                entry["gitversion"] = '0.2'
            term_posts_files.extend(source["termposts"])

            meta_data_files.append(source["kalla"])

            meta_data_file = source["kalla"]
            meta_data_file["_id"] = meta_data_file["id"]
            if (database == 'staging'):
                file_information = os.stat(fileobject)
                last_file_change_seconds = file_information.st_mtime
                last_file_change_date_time = datetime.fromtimestamp(last_file_change_seconds,
                                                                    tz=settings.TIME_ZONE_STOCKHOLM)
                meta_data_file['last_file_change_seconds'] = last_file_change_seconds
                meta_data_file['last_file_change_date_time'] = last_file_change_date_time.strftime('%Y-%m-%d %H:%M:%S')

            term_posts_file = source["termposts"]

            # Update all sources in both staging and public databases.
            # Code that is used when needed normally due to changes in datahandling
            # if (import_all_files and (database == Database.STAGING_DATABASE_NAME)):
                # TermManager.update_term_post_source(meta_data_file,
                                                    # meta_data_file["id"],
                                                    # Database.PUBLIC_DATABASE_NAME)

            if (database == 'staging'):
                TermManager._compare_term_post_information(term_posts_file, meta_data_file)

            TermManager.update_term_post_source(meta_data_file, meta_data_file["id"], database)
            TermManager.update_term_posts(term_posts_file, meta_data_file["id"], database)

        # print(f'Antal fält med koden INTE = {SourceFileTermPostField.field_counter}')
        print("Klar med importen")
        return imported_files

    @staticmethod
    def import_source_to_public(term_post_source_id):
        term_post_source = TermManager.get_term_post_source(term_post_source_id, Database.STAGING_DATABASE_NAME)
        if (term_post_source is not None):
            term_posts = TermManager.get_term_posts_by_term_post_source(term_post_source_id,
                                                                        False,
                                                                        Database.STAGING_DATABASE_NAME)
            if (len(term_posts) == 0):
                term_post_source = None
            else:
                # Clear comparison.
                term_post_source.comparison = []
                term_post_source_document = term_post_source.get_document()
                TermManager.update_term_post_source(term_post_source_document,
                                                    term_post_source_id,
                                                    Database.STAGING_DATABASE_NAME)
                TermManager.update_term_post_source(term_post_source_document,
                                                    term_post_source_id,
                                                    Database.PUBLIC_DATABASE_NAME)
                TermManager.update_term_posts(term_posts, term_post_source_id, Database.PUBLIC_DATABASE_NAME)
        return term_post_source

    @staticmethod
    def print_log(file_path, source_file):
        log = source_file.get_log()
        if (log):
            short_path = SourceFileManager.get_short_file_path(file_path)
            print("\n===========================================================")
            if (source_file.has_source_id()):
                print('Id för källa =', source_file.source_id, ': filnamn =', short_path, '\n')
            else:
                print('Inget id för källa : filnamn =', short_path, '\n')
            for row in log:
                print(*row)
            print("===========================================================\n")

    @staticmethod
    def update_term_post_source(term_post_source_document, term_post_source_id, database):
        MongoDBManager.update_term_post_source(term_post_source_document, term_post_source_id, database)

    @staticmethod
    def update_term_posts(term_post_documents, term_post_source_id, database):
        MongoDBManager.update_term_posts(term_post_documents, term_post_source_id, database)
