from django.db import models


class TermPostClassificationSystem(models.Model):
    """ Handles information about one term post classification system.

    Attributes:
        description:    Description of the classification system.
        is_ok:          Indicates if information handling has encountered any problems.
        name:           Name of the classification system.
        short_name:     Short name of the classification system.
        url:            Url to more information about the classification system.
    """

    CLASSIFICATION_SYSTEM_KEY = 'Klassifikationssystem'
    DESCRIPTION_VALUE = 'beskrivning'
    NAME_VALUE = 'namn'
    SHORT_NAME_VALUE = 'kortnamn'
    KEY_SEPARATOR = '.'
    URL_VALUE = 'url'

    description = models.CharField(max_length=2048, null=True, blank=True)
    is_ok = models.BooleanField(null=True, default=True)
    name = models.CharField(max_length=256, null=True, blank=True)
    short_name = models.CharField(max_length=256, null=True, blank=True)
    url = models.CharField(max_length=2048, null=True, blank=True)

    def add_information(self, value_type: str, value: str):
        """ Add classification system information to this instance.

        Arguments:
            value_type: Type of classification system value which must match one of a few predefined values.
            value: The value.
        """

        if (value_type == TermPostClassificationSystem.DESCRIPTION_VALUE):
            self.description = value
        elif (value_type == TermPostClassificationSystem.NAME_VALUE):
            self.name = value
        elif (value_type == TermPostClassificationSystem.SHORT_NAME_VALUE):
            self.short_name = value
        elif (value_type == TermPostClassificationSystem.URL_VALUE):
            self.url = value
        else:
            self.is_ok = False
            raise ValueError(f'Ej hanterad information om klassifikationssystem: {value_type}, {value}')

    @staticmethod
    def is_classification_system_key(key: str) -> bool:
        """ Test if key is related to classification system information.

        Arguments:
            key: The first part of one row in a term post file.

        Returns:
            bool: True if key is related to classification system information.
        """

        if (TermPostClassificationSystem.KEY_SEPARATOR in key):
            components = key.split(TermPostClassificationSystem.KEY_SEPARATOR)
            return (components[0] == TermPostClassificationSystem.CLASSIFICATION_SYSTEM_KEY) and (len(components) == 3)
        else:
            return False
