class TermPostClassificationSystemOld():
    """ Handles information about one term post classification system.

    Attributes:
        description:    Description of the classification system.
        is_ok:          Indicates if information handling has encountered any problems.
        name:           Name of the classification system.
        short_name:     Short name of the classification system.
        url:            Url to more information about the classification system.
    """

    CLASSIFICATION_SYSTEM_KEY = 'Klassifikationssystem'
    DESCRIPTION_VALUE = 'beskrivning'
    NAME_VALUE = 'namn'
    SHORT_NAME_VALUE = 'kortnamn'
    KEY_SEPARATOR = '.'
    URL_VALUE = 'url'

    def __init__(self):
        self.description = None
        self.is_ok = True
        self.name = None
        self.short_name = None
        self.url = None

    def add_information(self, value_type: str, value: str):
        """ Add classification system information to this instance.

        Arguments:
            value_type: Type of classification system value which must match one of a few predefined values.
            value: The value.
        """

        if (value_type == TermPostClassificationSystemOld.DESCRIPTION_VALUE):
            self.description = value
        elif (value_type == TermPostClassificationSystemOld.NAME_VALUE):
            self.name = value
        elif (value_type == TermPostClassificationSystemOld.SHORT_NAME_VALUE):
            self.short_name = value
        elif (value_type == TermPostClassificationSystemOld.URL_VALUE):
            self.url = value
        else:
            self.is_ok = False
            raise ValueError(f'Ej hanterad information om klassifikationssystem: {value_type}, {value}')

    def get_document(self):
        document = {}
        if (self.description):
            document['beskrivning'] = self.description
        if (self.short_name):
            document['kortnamn'] = self.short_name
        if (self.name):
            document['namn'] = self.name
        if (self.url):
            document['url'] = self.url
        return document

    @staticmethod
    def get_term_post_classification_system_from_document(term_post_classification_system_document):
        term_post_classification_system = TermPostClassificationSystemOld()
        if ('beskrivning' in term_post_classification_system_document):
            term_post_classification_system.add_information(TermPostClassificationSystemOld.DESCRIPTION_VALUE,
                                                            term_post_classification_system_document['beskrivning'])
        if ('kortnamn' in term_post_classification_system_document):
            term_post_classification_system.add_information(TermPostClassificationSystemOld.SHORT_NAME_VALUE,
                                                            term_post_classification_system_document['kortnamn'])
        if ('namn' in term_post_classification_system_document):
            term_post_classification_system.add_information(TermPostClassificationSystemOld.NAME_VALUE,
                                                            term_post_classification_system_document['namn'])
        if ('url' in term_post_classification_system_document):
            term_post_classification_system.add_information(TermPostClassificationSystemOld.URL_VALUE,
                                                            term_post_classification_system_document['url'])
        return term_post_classification_system

    @staticmethod
    def is_classification_system_key(key: str) -> bool:
        """ Test if key is related to classification system information.

        Arguments:
            key: The first part of one row in a term post file.

        Returns:
            bool: True if key is related to classification system information.
        """

        if (TermPostClassificationSystemOld.KEY_SEPARATOR in key):
            components = key.split(TermPostClassificationSystemOld.KEY_SEPARATOR)
            return (components[0] == TermPostClassificationSystemOld.CLASSIFICATION_SYSTEM_KEY) and (len(components) == 3)
        else:
            return False
