from django.db import models


class TermPostPublisher(models.Model):
    """ Handles information about one term post publisher.

        A source file can contain term posts from several publishers.

    Attributes:
        author:         Author of term posts.
        contact_person: Contact person.
        email_address:  Email address.
        is_ok:          Indicates if information handling has encountered any problems.
        organisation:   Organisation name
        publisher_type: Publisher type.
    """

    AUTHOR_VALUE = 'författare'
    CONTACT_PERSON_VALUE = 'kontaktperson'
    EMAIL_VALUE = 'e-post'
    KEY_SEPARATOR = '.'
    ORGANIZATION_VALUE = 'organisation'
    PUBLISHER_KEY = 'Utgivare'

    author = models.CharField(max_length=512, null=True, blank=True)
    contact_person = models.CharField(max_length=128, null=True, blank=True)
    email_address = models.CharField(max_length=128, null=True, blank=True)
    is_ok = models.BooleanField(null=True, default=True)
    organisation = models.CharField(max_length=2048, null=True, blank=True)
    publisher_type = models.CharField(max_length=128, null=True, blank=True)

    def __init__(self, publisher_type: str = None):
        super().__init__()
        self.publisher_type = publisher_type

    def add_information(self, publisher_type: str, value_type: str, value: str):
        """ Add publisher information to this publisher.

        Arguments:
            publisher_type: Publisher type which must match the type that was specified when this instance was created.
            value_type: Type of publisher value which must match one of a few predefined values.
            value: The value.
        """

        assert(self.publisher_type == publisher_type)
        if (value_type == TermPostPublisher.AUTHOR_VALUE):
            self.author = value
        elif (value_type == TermPostPublisher.CONTACT_PERSON_VALUE):
            self.contact_person = value
        elif (value_type == TermPostPublisher.EMAIL_VALUE):
            self.email_address = value
        elif (value_type == TermPostPublisher.ORGANIZATION_VALUE):
            self.organisation = value
        else:
            self.is_ok = False
            raise ValueError(f'Ej hanterad information om utgivare: {value_type}, {value}')

    @staticmethod
    def is_publisher_key(key: str) -> bool:
        """ Test if key is related to publisher information.

        Arguments:
            key: The first part of one row in a term post file.

        Returns:
            bool: True if key is related to publisher information.
        """

        if (TermPostPublisher.KEY_SEPARATOR in key):
            components = key.split(TermPostPublisher.KEY_SEPARATOR)
            return (components[0] == TermPostPublisher.PUBLISHER_KEY) and (len(components) == 4)
        else:
            return False
