
class PublisherTypeError(Exception):
    pass


class TermPostPublisherOld():
    """ Handles information about one term post publisher.

        A source file can contain term posts from several publishers.

    Attributes:
        author:         Author of term posts.
        contact_person: Contact person.
        email_address:  Email address.
        is_ok:          Indicates if information handling has encountered any problems.
        organisation:   Organisation name
        publisher_type: Publisher type.
    """

    AUTHOR_VALUE = 'författare'
    CONTACT_PERSON_VALUE = 'kontaktperson'
    EMAIL_VALUE = 'e-post'
    KEY_SEPARATOR = '.'
    ORGANIZATION_VALUE = 'organisation'
    PUBLISHER_KEY = 'Utgivare'

    def __init__(self, publisher_type: str):
        self.author = None
        self.contact_person = None
        self.email_address = None
        self.is_ok = True
        self.organisation = None
        self.publisher_type = publisher_type

    def add_information(self, publisher_type: str, value_type: str, value: str):
        """ Add publisher information to this publisher.

        Arguments:
            publisher_type: Publisher type which must match the type that was specified when this instance was created.
            value_type: Type of publisher value which must match one of a few predefined values.
            value: The value.
        """

        if (self.publisher_type != publisher_type):
            raise PublisherTypeError

        if (value_type == TermPostPublisherOld.AUTHOR_VALUE):
            self.author = value
        elif (value_type == TermPostPublisherOld.CONTACT_PERSON_VALUE):
            self.contact_person = value
        elif (value_type == TermPostPublisherOld.EMAIL_VALUE):
            self.email_address = value
        elif (value_type == TermPostPublisherOld.ORGANIZATION_VALUE):
            self.organisation = value
        else:
            self.is_ok = False
            raise ValueError(f'Ej hanterad information om utgivare: {value_type}, {value}')

    def get_document(self):
        document = {}
        if (self.author):
            document['författare'] = self.author
        if (self.contact_person):
            document['kontaktperson'] = self.contact_person
        if (self.email_address):
            document['e-post'] = self.email_address
        if (self.organisation):
            document['organisation'] = self.organisation
        if (self.publisher_type):
            document['type'] = self.publisher_type
        return document

    @staticmethod
    def get_term_post_publisher_from_document(term_post_publisher_document):
        assert ('type' in term_post_publisher_document)
        publisher_type = term_post_publisher_document['type']
        term_post_publisher = TermPostPublisherOld(publisher_type)
        if ('författare' in term_post_publisher_document):
            term_post_publisher.add_information(publisher_type,
                                                TermPostPublisherOld.AUTHOR_VALUE,
                                                term_post_publisher_document['författare'])
        if ('kontaktperson' in term_post_publisher_document):
            term_post_publisher.add_information(publisher_type,
                                                TermPostPublisherOld.CONTACT_PERSON_VALUE,
                                                term_post_publisher_document['kontaktperson'])
        if ('e-post' in term_post_publisher_document):
            term_post_publisher.add_information(publisher_type,
                                                TermPostPublisherOld.EMAIL_VALUE,
                                                term_post_publisher_document['e-post'])
        if ('organisation' in term_post_publisher_document):
            term_post_publisher.add_information(publisher_type,
                                                TermPostPublisherOld.ORGANIZATION_VALUE,
                                                term_post_publisher_document['organisation'])
        return term_post_publisher

    @staticmethod
    def is_publisher_key(key: str) -> bool:
        """ Test if key is related to publisher information.

        Arguments:
            key: The first part of one row in a term post file.

        Returns:
            bool: True if key is related to publisher information.
        """

        if (TermPostPublisherOld.KEY_SEPARATOR in key):
            components = key.split(TermPostPublisherOld.KEY_SEPARATOR)
            return (components[0] == TermPostPublisherOld.PUBLISHER_KEY) and (len(components) == 4)
        else:
            return False
