from copy import deepcopy
from datetime import datetime
from django.conf import settings
from ISOF.common.string_functions import is_integer_string
from model.import_util import string_substitutions
from model.term_post_classification_system import TermPostClassificationSystem
from model.term_post_publisher import TermPostPublisher
import os


class TermPostSource():
    """ Handles information about ont term post source.

    Attributes:
        classification_systems:     Classification systems for term posts.
        comparison:                 Result from comparing source in staging and public databases.
        error_log                   Errors that was found when the source was imported into the staging database.
        filename:                   Name of file in bankvalvet.
        filepath:                   full path (in bankvalvet) to the file.
        id:                         Integer value that should be unique in Rikstermbanken.
        information:                Dictionary with information about source.
        is_ok:                      Indicates if information handling has encountered any problems.
        is_split_default:           Contains information about how multiple terms
                                    should be displayed in web application.
        is_split_language:          Contains information about how multiple terms
                                    should be displayed in web application.
        last_file_change_date_time: Last time (DateTime as a string) that the source file was changed.
        last_file_change_seconds:   Last time (in seconds) that the source file was changed.
        place:                      Information about where in the published material that the term posts are specified.
        publishers:                 Publishers of term posts.
        publisher_information:      Information about publisher.
        publish_year:               The year that the term information was published.
        term_post_count:            Number of term posts in this source.
        title_short:                The short version of the term post source title.
    """

    IS_SPLIT_DEFAULT_VALUE = 'Tolka.skillnadTermSynonym.förvalt'
    IS_SPLIT_PRE_STRING = 'Tolka.skillnadTermSynonym.'
    IS_SPLIT_VALUES = ['troligenIngenSkillnad', 'troligenUppdelat', r'troligenUppdelat/troligenIngenSkillnad']
    PLACE_VALUE = 'Ställe'
    POSITION_NOT_OBVIOUS_VALUE = 'Tolka.OKGR-märkning'
    POSITION_NOT_OBVIOUS_VALUES = ['annarsOckså',
                                   'AnnarsOckså',
                                   'baraOmInteOrdklassSjälvklar',
                                   r'baraOmInteOrdklassSjälvklar/annarsOckså']
    PUBLISH_YEAR_VALUE = 'Utgivningsår'
    SECRETARIAT_COMMENT = 'Sekretariatskommentar'
    SOURCE_ID_VALUE = 'Källid'
    SOURCE_STATUS_VALUE = 'Status'
    TITLE_DOCUMENT_NUMBER_VALUE = 'Titel.dokumentnr'
    TITLE_SHORT_VALUE = 'Titel.kortform'
    TITLE_SUPERIOR_VALUE = 'Titel.överordnad'
    _no_source_id = -1

    def __init__(self, filepath: str, information: dict, publishers: list, classification_systems: list):
        if (filepath):
            self.comparison = []
            self.error_log = []
            self.filepath = filepath
            self.term_post_count = 0

        self.information = information
        self.classification_systems = classification_systems
        self.publishers = publishers

        # Check if instance is ok so far.
        self.is_ok = True
        if (self.publishers):
            for publisher in self.publishers:
                if (not publisher.is_ok):
                    self.is_ok = False
                    break
        if (self.classification_systems):
            for classification_system in self.classification_systems:
                if (not classification_system.is_ok):
                    self.is_ok = False
                    break

        if (filepath):
            # Get id.
            self.id = None
            if (TermPostSource.SOURCE_ID_VALUE in self.information):
                id_string = self.information[TermPostSource.SOURCE_ID_VALUE]
                if (is_integer_string(id_string)):
                    self.id = int(id_string)
            if (self.id is None):
                self.id = TermPostSource._no_source_id
                TermPostSource._no_source_id = TermPostSource._no_source_id - 1
                self.is_ok = False
                self.status = 0
                self.status_text = 'inteFärdig'
                raise ValueError("Id för källa finns inte med i metadata.\nTermposterna i filen har inte importerats.")

            else:
                # Get status.
                self.status = None
                if (TermPostSource.SOURCE_STATUS_VALUE in self.information):
                    self.status_text = self.information[TermPostSource.SOURCE_STATUS_VALUE]
                    status_text = self.status_text.lower()
                    if (status_text == "publicerad"):
                        self.status = 2
                    elif (status_text in ["avpublicerad", "inaktuell"]):
                        self.status = 1
                    elif (status_text in ["intefärdig", "bearbetas"]):
                        self.status = 0
                    else:
                        self.is_ok = False
                        raise ValueError(f"Okänd status i metadata.\nFilen har inte importerats, {self.status_text}")
                else:
                    self.is_ok = False
                    raise ValueError("Status finns inte med i metadata.\nFilen har inte importerats.")

            # Get file information.
            file_information = os.stat(self.filepath)
            self.last_file_change_seconds = file_information.st_mtime
            self.last_file_change_date_time = datetime.fromtimestamp(self.last_file_change_seconds,
                                                                     tz=settings.TIME_ZONE_STOCKHOLM)
            self.last_file_change_date_time = self.last_file_change_date_time.strftime('%Y-%m-%d %H:%M:%S')
            self.filename = os.path.basename(self.filepath)
            bankvalvet_index = self.filepath.find('bankvalvet')
            if (bankvalvet_index != -1):
                self.filepath = self.filepath[bankvalvet_index:]

        if (filepath is None):
            assert('id' in self.information)
            self.id = self.information['id']
            assert('status' in self.information)
            self.status = self.information['status']
            if ('comparison' in self.information):
                self.comparison = self.information['comparison']
            if ('errors' in self.information):
                self.error_log = self.information['errors']
            if ('filename' in self.information):
                self.filename = self.information['filename']
            if ('filepath' in self.information):
                self.filepath = self.information['filepath']
            if ('last_file_change_seconds' in self.information):
                self.last_file_change_seconds = self.information['last_file_change_seconds']
            if ('last_file_change_date_time' in self.information):
                self.last_file_change_date_time = self.information['last_file_change_date_time']
            if ('rows' in self.information):
                self.rows = self.information['rows']
            if ('status_text' in self.information):
                self.status_text = self.information['status_text']
            if ('term_post_count' in self.information):
                self.term_post_count = self.information['term_post_count']

        # Get publisher information.
        self.publisher_information = ''
        for publisher in self.publishers:
            #  TODO: Move next line out of loop if all publishers should be included.
            publisher_list = []
            if (publisher.organisation is not None):
                publisher_list.append(publisher.organisation)
            if (publisher.author is not None):
                publisher_list.append(publisher.author)
            #  TODO: Move next two lines out of loop if all publishers should be included.
            if (publisher_list):
                self.publisher_information = ", ".join(publisher_list)

        # Get other specific values.
        if (TermPostSource.TITLE_DOCUMENT_NUMBER_VALUE in self.information):
            self.title_document_number = self.information[TermPostSource.TITLE_DOCUMENT_NUMBER_VALUE]
        else:
            self.title_document_number = None
        if (TermPostSource.TITLE_SHORT_VALUE in self.information):
            self.title_short = self.information[TermPostSource.TITLE_SHORT_VALUE]
        else:
            self.title_short = None
        if (TermPostSource.TITLE_SUPERIOR_VALUE in self.information):
            self.title_superior = self.information[TermPostSource.TITLE_SUPERIOR_VALUE]
        else:
            self.title_superior = None
        if (TermPostSource.PUBLISH_YEAR_VALUE in self.information):
            self.publish_year = self.information[TermPostSource.PUBLISH_YEAR_VALUE]
        else:
            self.publish_year = None
        if (TermPostSource.PLACE_VALUE in self.information):
            self.place = self.information[TermPostSource.PLACE_VALUE]
        else:
            self.place = None

        self.is_position_not_obvious = True
        if (TermPostSource.POSITION_NOT_OBVIOUS_VALUE in self.information):
            is_position_not_obvious = self.information[TermPostSource.POSITION_NOT_OBVIOUS_VALUE]
            if (is_position_not_obvious not in TermPostSource.POSITION_NOT_OBVIOUS_VALUES):
                raise ValueError(f"Okänt värde för Tolka.OKGR-märkning {is_position_not_obvious}")
            self.is_position_not_obvious = (is_position_not_obvious == "baraOmInteOrdklassSjälvklar")

        # Get information about if term is split.
        self.is_split_default = 'troligenUppdelat'
        if (TermPostSource.IS_SPLIT_DEFAULT_VALUE in self.information):
            self.is_split_default = self.information[TermPostSource.IS_SPLIT_DEFAULT_VALUE]
            if (self.is_split_default not in TermPostSource.IS_SPLIT_VALUES):
                raise ValueError(f"Okänt värde för Tolka.skillnadTermSynonym: {self.is_split_default}")
        self.is_split_language = {}
        for key in self.information.keys():
            if (key.startswith(TermPostSource.IS_SPLIT_PRE_STRING) and (key != TermPostSource.IS_SPLIT_DEFAULT_VALUE)):
                length_key = len(TermPostSource.IS_SPLIT_PRE_STRING)
                language = key[length_key: length_key + 3]
                # language = key[20: 3].strip()
                is_split_value = self.information[key]
                if (is_split_value not in TermPostSource.IS_SPLIT_VALUES):
                    raise ValueError(f"Okänt värde för Tolka.skillnadTermSynonym: {is_split_value}")
                self.is_split_language[language] = is_split_value

        # Change styling.
        if ((TermPostSource.SECRETARIAT_COMMENT in self.information) and
            (self.information[TermPostSource.SECRETARIAT_COMMENT] is not None)):
            self.information[TermPostSource.SECRETARIAT_COMMENT] = self.information[TermPostSource.SECRETARIAT_COMMENT].replace(r'{b}', r'<b>')
            self.information[TermPostSource.SECRETARIAT_COMMENT] = self.information[TermPostSource.SECRETARIAT_COMMENT].replace(r'{/b}', r'</b>')

    @staticmethod
    def fix_term_post_source_document(term_post_source_document):
        term_post_source_document = deepcopy(term_post_source_document)

        # Change '_' to '.'.
        if ('Titel_dokumentnr' in term_post_source_document):
            term_post_source_document['Titel.dokumentnr'] = term_post_source_document['Titel_dokumentnr']
            del term_post_source_document['Titel_dokumentnr']
        if ('Titel_kortform' in term_post_source_document):
            term_post_source_document['Titel.kortform'] = term_post_source_document['Titel_kortform']
            del term_post_source_document['Titel_kortform']
        if ('Titel_överordnad' in term_post_source_document):
            term_post_source_document['Titel.överordnad'] = term_post_source_document['Titel_överordnad']
            del term_post_source_document['Titel_överordnad']
        if ('Tolka_OKGR-märkning' in term_post_source_document):
            term_post_source_document['Tolka.OKGR-märkning'] = term_post_source_document['Tolka_OKGR-märkning']
            del term_post_source_document['Tolka_OKGR-märkning']
        for key in term_post_source_document.keys():
            if (key.startswith('Tolka_skillnadTermSynonym_')):
                new_key = key.replace('_', '.')
                term_post_source_document[new_key] = term_post_source_document[key]
                del term_post_source_document[key]

        return term_post_source_document

    def get_document(self):
        """ Get document for term posts source that is stored in the database.

        Returns:
            dict: Information about term post source.
        """

        document = {}
        for key in self.information.keys():
            if (key in ['Källid', 'Status']):
                continue
            document[key] = self.information[key]
        if (hasattr(self, 'comparison')):
            document['comparison'] = self.comparison
        if (hasattr(self, 'error_log')):
            document['errors'] = self.error_log
        if (hasattr(self, 'filename')):
            document['filename'] = self.filename
        if (hasattr(self, 'filepath')):
            document['filepath'] = self.filepath
        document['id'] = self.id
        document['_id'] = self.id
        if (hasattr(self, 'last_file_change_seconds')):
            document['last_file_change_seconds'] = self.last_file_change_seconds
        if (hasattr(self, 'last_file_change_date_time')):
            document['last_file_change_date_time'] = self.last_file_change_date_time
        document['status'] = self.status
        if (hasattr(self, 'status_text')):
            document['status_text'] = self.status_text
        document['term_post_count'] = self.term_post_count
        if (self.classification_systems):
            classification_systems = []
            for classification_system in self.classification_systems:
                classification_systems.append(classification_system.get_document())
            if (classification_systems):
                document['Klassifikationssystem'] = classification_systems
        if (self.publishers):
            publishers = []
            for publisher in self.publishers:
                publishers.append(publisher.get_document())
            if (publishers):
                document['Utgivare'] = publishers
        return document

    def get_document_for_term_post(self):
        """ Get information about term post source that is included in the document for term posts in the database.

        Returns:
            dict: Information about term post source.
        """

        return {
            "id": self.id,
            "year": self.publish_year,
            "title": self.title_short,
            "publisher": self.publisher_information,
            "place": self.place,
            "status": self.status,
        }

    @staticmethod
    def get_term_post_classification_system(source_file, classification_systems, key, value):
        value = string_substitutions(value)
        classification_system_components = key.split(TermPostClassificationSystem.KEY_SEPARATOR)
        classification_system_index = classification_system_components[1]
        if (is_integer_string(classification_system_index)):
            classification_system_index = int(classification_system_index)
            classification_system_value_type = classification_system_components[2]

            if (classification_system_index in classification_systems):
                classification_system = classification_systems[classification_system_index]
            else:
                classification_system = TermPostClassificationSystem()
                classification_systems[classification_system_index] = classification_system
            classification_system.add_information(classification_system_value_type, value)

        else:
            source_file.log(f"Numreringen av klassifikationssystem är inte ett heltal: {classification_system_components}")

    @staticmethod
    def get_term_post_publisher(source_file, publishers, key, value):
        publisher_components = key.split(TermPostPublisher.KEY_SEPARATOR)
        publisher_index = publisher_components[1]
        if (is_integer_string(publisher_index)):
            publisher_index = int(publisher_index)
            publisher_type = publisher_components[2]
            publisher_value_type = publisher_components[3]

            if (publisher_index in publishers):
                publisher = publishers[publisher_index]
            else:
                publisher = TermPostPublisher(publisher_type)
                publishers[publisher_index] = publisher
            publisher.add_information(publisher_type, publisher_value_type, value)

        else:
            source_file.log(f"Numreringen av utgivare är inte ett heltal: {publisher_components}")

    @staticmethod
    def get_term_post_source_from_document(term_post_source_document):
        term_post_source_document = TermPostSource.fix_term_post_source_document(term_post_source_document)

        if ('Utgivare' in term_post_source_document):
            publishers = []
            for publisher_document in term_post_source_document['Utgivare']:
                publisher = TermPostPublisher.get_term_post_publisher_from_document(publisher_document)
                publishers.append(publisher)
            del term_post_source_document['Utgivare']
        else:
            publishers = None

        if ('Klassifikationssystem' in term_post_source_document):
            classification_systems = []
            for classification_system_document in term_post_source_document['Klassifikationssystem']:
                classification_system = TermPostClassificationSystem.get_term_post_classification_system_from_document(classification_system_document)
                classification_systems.append(classification_system)
            del term_post_source_document['Klassifikationssystem']
        else:
            classification_systems = None

        term_post_source = TermPostSource(None, term_post_source_document, publishers, classification_systems)
        return term_post_source

    @staticmethod
    def get_term_post_source_from_rows(source_file, rows, filename):
        classification_systems = {}
        publishers = {}
        information = {}

        for row in rows:
            # Split row into key and value.
            (key, _, value) = row.partition(" ")
            key = key.rstrip("\r\n").lstrip("\ufeff")
            value = value.rstrip("\r\n")

            if (not value):
                continue

            # Parse row.
            if (TermPostPublisher.is_publisher_key(key)):
                TermPostSource.get_term_post_publisher(source_file, publishers, key, value)
            elif (TermPostClassificationSystem.is_classification_system_key(key)):
                TermPostSource.get_term_post_classification_system(source_file, classification_systems, key, value)
            else:
                information[key] = value

        information['rows'] = rows
        term_post_source = TermPostSource(filename,
                                          information,
                                          publishers,
                                          classification_systems,
                                          rows)

        return term_post_source

    def has_id(self):
        return ((self.id is not None) and (self.id > -1))

    def has_status(self):
        return (self.status is not None)

    def is_published(self):
        return (self.status == 2)

    def is_split(self, language):
        if (language in self.is_split_language):
            return (self.is_split_language[language] == 'troligenUppdelat')
        else:
            return (self.is_split_default == 'troligenUppdelat')

    def set_comparison(self, comparison: list):
        self.comparison = comparison.copy()

    def set_error_log(self, error_log: list):
        self.error_log = error_log.copy()

    def set_term_post_count(self, term_post_count: int):
        self.term_post_count = term_post_count

